//
// cybimg.cpp - Convert a text image to a TIFF image for TRT2 on the Cyber
// =======================================================================
//
// The TIFF image is uncompressed 3 channel (RGB interleaved), 
// 1 byte per sample.
//
// A "text image" is a text file as produced by TRT2 for Cyber mainframes.
// It contains an RGB image as follows:
// First line: pixels lines (2 integer numbers)
// Then: each line of the image is encoded as a set of text lines. Each
// text line (except possibly the last) contains 24 RGB triplets, each
// triplet is a 3 digit decimal number in the range 0:255, no spaces between
// them, leading space filled. Each text line also begins with a space,
// which should be just skipped.
//
// This is a self contained program containing hacked portions of a much
// larger image processing library and is *not* pretty.
//
// Nick Glazzard
//

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#undef DEBUG_PRINT

#ifdef WIN32
#define STRICMP stricmp
#define LOCAL_TIFF_ENDIAN 0x4949
#endif
#ifdef UNIX
#define STRICMP strcasecmp
#ifdef LINUX
#define LOCAL_TIFF_ENDIAN 0x4949
#else
#define LOCAL_TIFF_ENDIAN 0x4d4d
#endif
#endif

#ifndef TRUE
#define TRUE   1
#define FALSE  0
#endif

// Largest allowed image size:
#define IMG_MAX_WIDTH  65535
#define IMG_MAX_HEIGHT 65535

// Image data formats:
#define IMG_FMT_BYTE   0    // 8 bits/channel
#define IMG_FMT_WORD   1    // 16 bits/channel
#define IMG_FMT_WORD32 2    // 32 bits/channel
#define IMG_FMT_FLOAT  3    // 32 bit floating point/channel

// Image file formats:
#define IMG_FMT_UNKNOWN 0
#define IMG_FMT_TIFF    1

// Image data pointer type:
typedef union
{
  float          *f;
  unsigned int   *ui;
  unsigned short *us;
  unsigned char  *uc;
  void           *v;
} image_ptr;

// Image data type:
// Images are ALWAYS channel packed pixels. E.g. [<r><g><b>][<r><g><b>] ... NEVER planar. E.g. <r><r><r> ... 
// If you need planar organisation, use multiple 1 channel images as necessary.
typedef struct
{
  unsigned short w, h, c, t;  // Width, Height, Channels, Type of data
  unsigned short top_down;    // TRUE if data is top-down else bottom-up
  unsigned short owns_data;   // "data" was allocated for this structure (rather than shared).
  double vmin, vmax;          // Min and max value in image across all channels.
  int vminmaxset;             // TRUE if vmin and vmax have been set (not necessarily valid, though).
  image_ptr data;
} image_data;

// Basic image data type functions:
image_data* image_data_create( int w, int h, int c, int t, int owns_data );
int image_data_delete( image_data* img );
int image_sizeof( int w, int h, int c, int t );
int image_stride( int w, int h, int c, int t );
int image_stride_1( image_data* img );

// Image file i/o utilities:
int image_write_file( char* filename, image_data* image );

// TIFF file format data structures.
struct TiffHead
{
	unsigned short endian;
	unsigned short magic;
	unsigned int ifd_offset;
};

struct TiffField
{
	unsigned short tag;
	unsigned short field;
	unsigned int count;
	unsigned int offval;
};

struct TiffInfo
{
	FILE *fp;
	struct TiffHead head;
	unsigned short entry_count;
	struct TiffField* ifd;
	unsigned int width;
	unsigned int height;
	unsigned int samples;
	unsigned int bits;
	unsigned int dataoffset;
	unsigned int bytes;
	unsigned int rows_per_strip;
	unsigned int* strip_offsets;
	unsigned int* strip_counts;
	int strips;
	unsigned int orientation;
};

void image_tiff_free_info( void** info );
int image_tiff_write_header( FILE* fin, int w, int h, int c, int t, void** info );
int image_tiff_write_data( image_data* data, void** info );

//==============
// Main program
//==============

int main( int argc, char** argv )
//-------------------------------
// Read a text image file and write an equivalent TIFF image file.
{
  FILE* fin;
  int nx, ny, y, trec, trem, r, i, j, ostride;
  char line[80], cval[4];
  char *ip, *op;
  image_data* outimage;

  if( argc != 3 ){
    fprintf( stderr, "Usage: cybimg in.txt out.tif\n" );
    return( 1 );
  }

  fin = fopen( argv[1], "r" );
  if( fin == NULL ){
    fprintf( stderr, "Failed to open input text file: %s\n", argv[1] );
    return( 2 );
  }
  
  if( fgets( line, 80, fin ) == NULL ){
    fprintf( stderr, "Unexpected end of file.\n" );
    return( 15 );
  }
  if( sscanf( line, "%d %d", &nx, &ny ) != 2 ){
    fprintf( stderr, "Failed to read image dimensions.\n", nx, ny );
    return( 3 );
  }

  printf( "Image size: %d x %d\n", nx, ny );
  trec = nx / 8;
  trem = nx - ( 8 * trec );

  outimage = image_data_create( nx, ny, 3, IMG_FMT_BYTE, TRUE );
  ostride = image_stride_1( outimage );

  // Step over image lines
  for( y=0; y<ny; y++ ){
    unsigned char *iop = &(outimage->data.uc[y*ostride]);

    // All the fully used text lines ...
    for( r=0; r<trec; r++ ){
      if( fgets( line, 80, fin ) == NULL ){
        fprintf( stderr, "Unexpected end of file.\n" );
        return( 5 );
      }
      ip = line;
      ++ip;
      for( i=0; i<24; i++ ){
        for( j=0,op=cval; j<3; j++ ){
          if( isdigit(*ip) ){
            *op++ = *ip;
          }
          ip++;
        }
        *op = 0;
        *iop++ = atoi( cval );
      }
    }

    // Any partially filled last line...
    if( trem > 0 ){
      if( fgets( line, 80, fin ) == NULL ){
        fprintf( stderr, "Unexpected end of file.\n" );
        return( 5 );
      }
      ip = line;
      ++ip;
      for( i=0; i<(3*trem); i++ ){
        for( j=0,op=cval; j<3; j++ ){
          if( isdigit(*ip) ){
            *op++ = *ip;
          }
          ip++;
        }
        *op = 0;
        *iop++ = atoi( cval );
      }
    }
  }

  // Write the image file
  if( ! image_write_file( argv[2], outimage ) ){
    fprintf( stderr, "Failed to write output image: %s\n", argv[2] );
    return( 6 );
  }

  // Clean up
  image_data_delete( outimage );
  return( 0 );
}

//======================
// TIFF format routines
//======================

int image_sizeof( int w, int h, int c, int t )
//--------------------------------------------
// Calculate number of bytes in an image of supplied Width, Height, Channels and data Type.
// Return -1 if type of image data not recognised.
{
	int t_size = -1;
	
	t_size = w * h * c;
	
	switch( t ){
	case IMG_FMT_BYTE:  break;
	case IMG_FMT_WORD:  t_size <<= 1; break;
	case IMG_FMT_WORD32: t_size <<= 2; break;
	case IMG_FMT_FLOAT: t_size *= sizeof(float); break;
	default: t_size = -1; break;
	}
    return( t_size );
}

int image_stride( int w, int h, int c, int t )
//--------------------------------------------
// Calculate line stride : Number of bytes to move to next line of image.
{
	int t_stride = -1;
	
	t_stride = w * c;
	
	switch( t ){
	case IMG_FMT_BYTE:  break;
	case IMG_FMT_WORD:  t_stride <<= 1; break;
	case IMG_FMT_WORD32: t_stride <<= 2; break;
	case IMG_FMT_FLOAT: t_stride *= sizeof(float); break;
	default: t_stride = -1; break;
	}
	
	return( t_stride );
}

int image_stride_1( image_data* img )
//-----------------------------------
// Calculate line stride. 
{
    return( image_stride( img->w, img->h, img->c, img->t ) );
}

image_data* image_data_create( int w, int h, int c, int t, int owns_data )
//------------------------------------------------------------------------
// Create an image data structure for given Width, Height, Channels, Type.
// If [owns_data] allocate a buffer for image data. Otherwise don't.
// Return NULL on error.
{
  image_data* img = NULL;
  int dsize;
  void* data = NULL;
  
  if( w <= 0 || w > IMG_MAX_WIDTH || h <= 0 || h > IMG_MAX_HEIGHT ){
    return( NULL );
  }
  
  dsize = image_sizeof( w, h, c, t );
  
  if( dsize > 0 ){
    if( owns_data )
      data = malloc( dsize );
    
    if( data || ( owns_data == FALSE ) ){
      img = (image_data *)malloc( sizeof(image_data) );
      if( img ){
        img->data.v = data;
        img->w = w;
        img->h = h;
        img->c = c;
        img->t = t;
        img->top_down = FALSE;
        img->owns_data = owns_data;
        img->vmin = 0.0;
        img->vmax = 0.0;
        img->vminmaxset = FALSE; // Well, we could say TRUE, but more useful not to.
      }
      else if( data ) // This is CORRECT for owns_data FALSE too, because then data == NULL. Phew!
        free( data );
    }
  } 
  
  return( img );
}

int image_data_delete( image_data* img )
//--------------------------------------
// Delete an image, Return -1 on error else 0.
{
   if( img ){
       if( img->data.v && img->owns_data )
		   free( img->data.v );
       free( img );
       return( 0 );
   }
   return( -1 );
}

static int filename_to_img_fmt( char* filename )
//----------------------------------------------
// Return an image format code based on the extension part of filename.
{
	char* extp = strrchr( filename, '.' );
	if( extp == NULL )
		return( IMG_FMT_UNKNOWN );
	else{
		if( ! STRICMP( extp, ".tif" ) )
			return( IMG_FMT_TIFF );
		else
			return( IMG_FMT_UNKNOWN );
	}
}

int image_write_file( char* filename, image_data* image )
//-------------------------------------------------------
// Write an image file from an image.
// Image file format is determined solely from the supplied filename extension.
{
	FILE* fout;
	int status = FALSE;
	void* info;

	int img_fmt = filename_to_img_fmt( filename );
	switch( img_fmt ){
	case IMG_FMT_UNKNOWN:
		return( status );

	case IMG_FMT_TIFF:
		fout = fopen( filename, "wb+" );
		if( fout ){
			if( image_tiff_write_header( fout, image->w, image->h, image->c, image->t, &info ) ){
				if( image_tiff_write_data( image, &info ) )
					status = TRUE;
				image_tiff_free_info( &info );
			}
			fclose( fout );
			return( status );
		}
		else
			return( status );
		break;

	default:
		return( status );
	}

	return( status );
}


#define SF(f,t,d,c,v) \
	f.tag = t; \
	f.field = d; \
	f.count = c; \
	f.offval = v;

int image_tiff_write_header( FILE* fin, int w, int h, int c, int t, void** info )
//-------------------------------------------------------------------------------
// Write the vital things to a TIFF file header.
// Keep it as simple as possible:
// - Uncompressed
// - RGB or RGBA or Monochrome (Photometric Interpretation = 2 or 1)
// - Single plane (Planar Configuration = 1 )
// - Entire image in a single strip
// - Only one image in the file (single IFD).
// This is a really *horrible* file format ...
// Return FALSE on error.
{
#define N_TIFF_FIELDS 14
#define TIFF_SHORT    3
#define TIFF_LONG     4
#define TIFF_RATIONAL 5
	struct TiffHead head;
	struct TiffField field[N_TIFF_FIELDS];  // Various bits of shit we must fill in...

	unsigned short bits_per_sample[4];
	unsigned int xres[2], yres[2];
	unsigned int nullifd=0;
	int i, shs;

	unsigned int bits_per_sample_offset = sizeof(head)+sizeof(short)+N_TIFF_FIELDS*sizeof(struct TiffField)+sizeof(int);
	unsigned int xres_offset = bits_per_sample_offset + c*sizeof(short);
	unsigned int yres_offset = xres_offset + 2 * sizeof(int);
	unsigned int data_offset = yres_offset + 2 * sizeof(int);
	unsigned int total_bytes = image_sizeof( w, h, c, t );

	bits_per_sample[0] = 8 * ( total_bytes / ( w * h * c ) );

	xres[0] = 300; xres[1] = 1;
	yres[0] = 300; yres[1] = 1;

	if( c > 4 )
		c = 4;

	head.endian = LOCAL_TIFF_ENDIAN;
	shs = ( head.endian == 0x4d4d ) ? 16 : 0 ;
	head.magic = 42;
	head.ifd_offset = sizeof(head);

#ifdef DEBUG_PRINT
	printf( "w:%d h:%d c:%d t:%d sizeof(TiffField)=%d\n", w, h, c, t, sizeof(struct TiffField) );
#endif

	// Fill in the field structures to reflect our image.
	// Note - these buggers must be in ascending tag order.
	SF( field[0], 256, TIFF_SHORT, 1, (w<<shs) ); //--------------------- Width
#ifdef DEBUG_PRINT
	printf( "field[0].offval=%d\n", field[0].offval );
#endif
	SF( field[1], 257, TIFF_SHORT, 1, (h<<shs) ); //--------------------- Height
	if( c == 1 ){
	  SF( field[2], 258, TIFF_SHORT, c, (bits_per_sample[0]<<shs) ); //-- Bits per channel - 1 channel, so scalar
	}
	else{
		SF( field[2], 258, TIFF_SHORT, c, bits_per_sample_offset );  //-- Bits per channel - Multiple channels, so vector (sigh)
	}
	SF( field[3], 259, TIFF_SHORT, 1, (1<<shs) ); //--------------------- Compression = NONE
	if( c == 1 ){
	  SF( field[4], 262, TIFF_SHORT, 1, (1<<shs) ); //------------------- Mono: 0 = black, bigger => brighter (sigh)
	}
	else{
	  SF( field[4], 262, TIFF_SHORT, 1, (2<<shs) ); //------------------- RGB
	}
	SF( field[5],  273, TIFF_LONG,  1, data_offset ); //----------------- Where the image data starts
	SF( field[6],  274, TIFF_SHORT, 1, (4<<shs) ); //-------------------- First pixel is at BOTTOM LEFT!!!!!!
	SF( field[7],  277, TIFF_SHORT, 1, (c<<shs) ); //-------------------- Number of channels
	SF( field[8],  278, TIFF_SHORT, 1, (h<<shs) ); //-------------------- All lines ("rows") in one "strip" (sigh)
	SF( field[9],  279, TIFF_LONG,  1, image_sizeof( w, h, c, t ) ); //-- Bytes in the image data
	SF( field[10], 282, TIFF_RATIONAL, 1, xres_offset ); //-------------- X resolution in pixels per resolution unit.
	SF( field[11], 283, TIFF_RATIONAL, 1, yres_offset ); //-------------- Y resolution in pixels per resolution unit.
	SF( field[12], 284, TIFF_SHORT, 1, (1<<shs) ); //-------------------- Channels are interleaved
	SF( field[13], 296, TIFF_SHORT, 1, (1<<shs) ); //-------------------- Resolution unit is the inch.

	struct TiffInfo* ti = (struct TiffInfo*) malloc( sizeof( struct TiffInfo ) );
	if( ti ){
		memset( ti, 0, sizeof( struct TiffInfo ) );
		ti->fp = fin;
		*info = (void*)ti;

		if( fwrite( &head, sizeof(head), 1, ti->fp ) != 1 )
			goto error;
		ti->entry_count = N_TIFF_FIELDS;
		if( fwrite( &ti->entry_count, sizeof(unsigned short), 1, ti->fp ) != 1 )
			goto error;
		if( fwrite( &field, sizeof(struct TiffField), N_TIFF_FIELDS, ti->fp ) != N_TIFF_FIELDS )
			goto error;
		if( fwrite( &nullifd, sizeof(unsigned int), 1, ti->fp ) != 1 )
			goto error;
		ti->bytes = total_bytes;
		for( i=1; i<c; i++ )
			bits_per_sample[i] = bits_per_sample[0];
		if( (int)fwrite( bits_per_sample, sizeof(unsigned short), c, ti->fp ) != c )
			goto error;
		if( (int)fwrite( xres, sizeof(unsigned int), 2, ti->fp ) != 2 )
			goto error;
		if( (int)fwrite( yres, sizeof(unsigned int), 2, ti->fp ) != 2 )
			goto error;
		ti->dataoffset = data_offset;
		return( TRUE );
	}

error:
	return FALSE;
}

int image_tiff_write_data( image_data* data, void** info )
//-------------------------------------------------------
// Write data to an open TIFF format file.
{
	struct TiffInfo* ti = (struct TiffInfo*)(*info);
	
	if( fseek( ti->fp, ti->dataoffset, SEEK_SET ) != 0 )
		return( FALSE );
	if( fwrite( data->data.uc, sizeof(unsigned char), ti->bytes, ti->fp ) != ti->bytes )
		return( FALSE );
	return( TRUE );
}

void image_tiff_free_info( void** info )
//--------------------------------------
// Free up TIFF specific data
{
	struct TiffInfo* ti = (struct TiffInfo*)(*info);
	if( ti ){
		if( ti->ifd )
			free( ti->ifd );
		if( ti->strip_offsets )
			free( ti->strip_offsets );
		if( ti->strip_counts )
			free( ti->strip_counts );
		free( ti );
	}
}

