//
// TRT2 - Toy Ray Tracer Mark 2.
// =============================
// C++ IMPLEMENTATION
// ==================
//
// WARNING! If you are looking for a good raytracer - this isn't it!
// You will be better off looking at POV-Ray, et. al. (Not that there
// is anything actually *wrong* with this one that I know of - but it
// was written for a very specific and peculiar purpose.)
//
// TRT2 is a very simple ray tracing renderer coded in a very odd way.
// This is intended to make it easy to recode it in FORTRAN-77 (!).
// And maybe other (more) obscure languages.
//
// It is also an experiment in how simple and small a "useful" renderer can
// be - in terms of lines of code, anyway.
//
// The goals were:
//  1) Portability (it is self contained - no other libraries, etc.)
//  2) Ease of recoding in different languages
//  3) Small memory footprint
//  4) Still capable of making pretty pictures
//  5) Easy extensibility:
//     a) New geometric primitive types should be straightforward to add
//     b) Texture mapping, etc., would be straightforward to add on
//        machines with enough memory to make it feasible.
//
// Current features:
//  1) Sphere primitives
//  2) Triangle primitives with optional vertex normals (these will be correctly interpolated)
//  3) Stochastic anti-aliasing (basic distributed ray tracing)
//  4) Optional shadows, optionally appropriately "soft edged"
//  5) Specularly reflective materials (correct specular reflection)
//  6) Transparent materials (correct refraction)
//  7) Simple surface roiughness factor for materials.
//  8) Accelerated ray intersection tests using a space subdivision (voxel) scheme
//  9) Images output as text files (well - it makes for portability!).
// 10) Directional lights (raised cosine and barn door types)
//
// Having space subdivision acceleration means it isn't all that simple!
// However, this can make a really big difference with lots of primitives.
// Well - assuming they can be "indexed" (see code). To save memory, primitives that
// are too big cannot be "accelerated".
//
// Because this code must be translatable to FORTRAN, recursion is
// explicitly coded using a ray stack without making any actual
// recursive calls. Also, all storage must be statically allocated at
// compile time. Another oddity - all arrays are dimensioned to use 1 based
// array indices (again for direct, default, FORTRAN-77 compatibility).
//
// As always with floating point based code, the precision with which
// calculations are done is important. Tests for approximate equivalence
// and "nudging" things to avoid degenerate cases is unavoidable. Parts of
// the code which may suffer from "precision" issues are highlighted with
// the comment: PRECISION! The parameters EPS and MINEPS can tune some of
// this. Note that 32bit floating point is *not* good enough for the grid
// traversal code. It works perfectly with 64bit (well - pretty exhaustive
// testing has failed to make it go wrong) but *not* with 32 bit.
//
// The primary goal of the FORTRAN version of this code is to run on emulated
// CDC mainframes. Hence the accounting of the sizes of the main data structures
// below in "words". The total size of the program must be kept (well) below
// 131000 words. (60 bit words, that is!).
//
// Author: Nick Glazzard (in occasional time wasting mode) 2003,2004,2005 
//                       (Original TRT - 1984)
//

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#ifndef TRUE            // The usual ...
#define TRUE   1
#define FALSE  0
#endif

//==============================//
// BUILT-IN TIFF OUTPUT SUPPORT //
//==============================//

#undef DEBUG_PRINT

#ifdef WIN32
#define STRICMP stricmp
#define LOCAL_TIFF_ENDIAN 0x4949
#endif
#ifdef UNIX
#define STRICMP strcasecmp
#ifdef LINUX
#define LOCAL_TIFF_ENDIAN 0x4949
#else
#define LOCAL_TIFF_ENDIAN 0x4d4d
#endif
#endif

// Largest allowed image size:
#define IMG_MAX_WIDTH  65535
#define IMG_MAX_HEIGHT 65535

// Image data formats:
#define IMG_FMT_BYTE   0    // 8 bits/channel
#define IMG_FMT_WORD   1    // 16 bits/channel
#define IMG_FMT_WORD32 2    // 32 bits/channel
#define IMG_FMT_FLOAT  3    // 32 bit floating point/channel

// Image file formats:
#define IMG_FMT_UNKNOWN 0
#define IMG_FMT_TIFF    1

// Image data pointer type:
typedef union
{
  float          *f;
  unsigned int   *ui;
  unsigned short *us;
  unsigned char  *uc;
  void           *v;
} image_ptr;

// Image data type:
// Images are ALWAYS channel packed pixels. E.g. [<r><g><b>][<r><g><b>] ... NEVER planar. E.g. <r><r><r> ... 
// If you need planar organisation, use multiple 1 channel images as necessary.
typedef struct
{
  unsigned short w, h, c, t;  // Width, Height, Channels, Type of data
  unsigned short top_down;    // TRUE if data is top-down else bottom-up
  unsigned short owns_data;   // "data" was allocated for this structure (rather than shared).
  double vmin, vmax;          // Min and max value in image across all channels.
  int vminmaxset;             // TRUE if vmin and vmax have been set (not necessarily valid, though).
  image_ptr data;
} image_data;

// Basic image data type functions:
image_data* image_data_create( int w, int h, int c, int t, int owns_data=TRUE );
int image_data_delete( image_data* img );
int image_sizeof( int w, int h, int c, int t );
int image_stride( int w, int h, int c, int t );
int image_stride( image_data* img );

// Image file i/o utilities:
int image_write_file( char* filename, image_data* image );

// TIFF file format data structures.
struct TiffHead
{
	unsigned short endian;
	unsigned short magic;
	unsigned int ifd_offset;
};

struct TiffField
{
	unsigned short tag;
	unsigned short field;
	unsigned int count;
	unsigned int offval;
};

struct TiffInfo
{
	FILE *fp;
	struct TiffHead head;
	unsigned short entry_count;
	struct TiffField* ifd;
	unsigned int width;
	unsigned int height;
	unsigned int samples;
	unsigned int bits;
	unsigned int dataoffset;
	unsigned int bytes;
	unsigned int rows_per_strip;
	unsigned int* strip_offsets;
	unsigned int* strip_counts;
	int strips;
	unsigned int orientation;
};

void image_tiff_free_info( void** info );
int image_tiff_write_header( FILE* fin, int w, int h, int c, int t, void** info );
int image_tiff_write_data( image_data* data, void** info );

//=====================================//
// END OF BUILT-IN TIFF OUTPUT SUPPORT //
//=====================================//

// Sundry parameters

#define REAL double     // Floating point type for grid traversal. double is needed on 32bit machines.

#define GTHUGE 1e7      // Outside world value for grid traversal algorithm
#define MINEPS (1e-5)   // Precision factor for some floating point comparisons.
#define EPS 1e-7        // Precision factor for other floating point things (E.g. grid traversal algorithm)
#define MINFLT (-1e20)  // Very small (very big negative) number
#define MAXFLT (1e20)   // Very large (very big positive) number
#define TRTPI 3.1415926 // PI - avoiding conflicts
#define DEGRAD (TRTPI/180.0) // Degrees to radians constant.

#define EYERAY 1        // Ray is an eye ray
#define RFLRAY 2        // Ray is a reflected ray
#define TRNRAY 3        // Ray is a transmitted ray

#define ENTER  1        // Transmitted ray is entering primitive
#define LEAVE  2        // Transmitted ray is leaving primitive

#define NOACL  0        // No intersection acceleration
#define SUBACL 1        // Space subdivision acceleration

#define BKCNST 0        // Constant background colour (initial)
#define BKVERT 1        // Vertical grade
#define BKHORZ 2        // Horizontal grade

#define LGTCON 0        // Light not directional
#define LGTCOS 1        // Light raised cosine directionality
#define LGTBRN 2        // Light "barn door" directionality

#define SPHPRM 1        // Sphere primitive id
#define TRIPRM 2        // Triangle primitive id

#define DAXX   1        // Triangle normal X dominant case
#define DAXY   2        // Triangle normal Y dominant case
#define DAXZ   3        // Triangle normal Z dominant case

// In-memory and globally accessible data structures.

// Ray stack "common block"
// Size: DIMSTK + 2 words = 515
#define MAXSTK 512
#define DIMSTK (MAXSTK+1)

struct
{
	int nray;           // Number of rays on stack
	int ray[DIMSTK];    // Indices of rays on stack.
	int cray;           // Current ray index.
} raystk;

// Ray list "common block"
// Size: 13 * DIMRAY + 2 words = 6671
#define MAXRAY  512
#define DIMRAY  (MAXRAY+1)

struct
{
	int nray;           // Number of rays
	REAL ox[DIMRAY];    // Origin (eye space)
	REAL oy[DIMRAY];
	REAL oz[DIMRAY];
	REAL dx[DIMRAY];    // Direction (unit vector)
	REAL dy[DIMRAY];
	REAL dz[DIMRAY];
	int isect[DIMRAY];  // Index of intersection in intersection list.
	int type[DIMRAY];   // Type of ray.
	REAL wgtr[DIMRAY];  // Ray weight (accumulated as tree is built)
	REAL wgtg[DIMRAY];
	REAL wgtb[DIMRAY];
	int dep[DIMRAY];    // Ray depth in tree.
	int rayid[DIMRAY];  // Ray ID number (acceleration variable)
	int raynum;         // Number used for Ray ID (increments for each ray created).
} raylst;

// Intersection list "common block"
// Size: 13 * DIMINS + 1 words = 6696
#define MAXINS 512
#define DIMINS (MAXINS+3)
#define NEWINS (MAXINS+1)   // Where a new intersection is temporarily saved.
#define SHDINS (MAXINS+2)   // Where a shadow test intersection is saved.

struct
{
	int nins;           // Number of intersections
	int prim[DIMINS];   // Index of primitive it intersects with.
	int rray[DIMINS];   // Index of reflected ray
	int tray[DIMINS];   // Index of transmitted ray
	REAL px[DIMINS];    // Position of intersection (eye space)
	REAL py[DIMINS];
	REAL pz[DIMINS];
	REAL nx[DIMINS];    // Normal at intersection (unit vector)
	REAL ny[DIMINS];
	REAL nz[DIMINS];
	int ttyp[DIMINS];   // Transmitted ray entering or leaving?
	REAL colr[DIMINS];  // Colour at intersection.
	REAL colg[DIMINS];
	REAL colb[DIMINS];
} inslst;

// Primitive list "common block"
// Size: 5 * DIMPRM + 1 words = 5126
#define MAXPRM 1024
#define DIMPRM (MAXPRM+1)

struct
{
	int nprm;           // Number of primitives.
	int matter[DIMPRM]; // Index of material it is made from.
	int accel[DIMPRM];  // 1 if referenced in acceleration structure.
	int rayid[DIMPRM];  // Ray ID number of last ray that intersected.
	int prmtyp[DIMPRM]; // Primitive type: Sphere, Triangle, whatever.
	int prmidx[DIMPRM]; // Index of definition in primitive specific table.
} prmlst;

// Sphere list "common block"
// Size: 4 * DIMSPH + 1 words = 4101
#define MAXSPH 1024
#define DIMSPH (MAXSPH+1)

struct
{
	int nsph;           // Number of spheres
	REAL ox[DIMSPH];    // Origin (eye space)
	REAL oy[DIMSPH];
	REAL oz[DIMSPH];
	REAL rad[DIMSPH];   // Radius
} sphlst;

// Triangle list "common block"
// Size: 23 * DIMTRI + 1 words = 5192 
#define MAXTRI 256
#define DIMTRI (MAXTRI+1)

struct
{
	int ntri;           // Number of triangles
	REAL x1[DIMTRI];    // 1st. vertex position
	REAL y1[DIMTRI];
	REAL z1[DIMTRI];
	REAL x2[DIMTRI];    // 2nd. vertex position
	REAL y2[DIMTRI];
	REAL z2[DIMTRI];
	REAL x3[DIMTRI];    // 3rd. vertex position
	REAL y3[DIMTRI];
	REAL z3[DIMTRI];
	REAL nvx1[DIMTRI];  // 1st. vertex normal
	REAL nvy1[DIMTRI];
	REAL nvz1[DIMTRI];
	REAL nvx2[DIMTRI];  // 2nd. vertex normal
	REAL nvy2[DIMTRI];
	REAL nvz2[DIMTRI];
	REAL nvx3[DIMTRI];  // 3rd. vertex normal
	REAL nvy3[DIMTRI];
	REAL nvz3[DIMTRI];
	REAL nx[DIMTRI];    // Plane equation contants / surface normal
	REAL ny[DIMTRI];
	REAL nz[DIMTRI];
	REAL nd[DIMTRI];
	int dax[DIMTRI];    // Dominant axis number
} trilst;

// Material list "common block"
// Size: 14 * DIMMTR + 1 words = 463
#define MAXMTR 32
#define DIMMTR (MAXMTR+1)

struct
{
	int nmtr;           // Number of materials.
	REAL kr[DIMMTR];    // Coefficient of reflection.
	REAL kt[DIMMTR];    // Coefficient of transmission.
	REAL car[DIMMTR];   // Ambient colour/intensity
	REAL cag[DIMMTR];
	REAL cab[DIMMTR];
	REAL cdr[DIMMTR];   // Diffuse colour/intensity
	REAL cdg[DIMMTR];
	REAL cdb[DIMMTR];
	REAL csr[DIMMTR];   // Specular colour/intensity
	REAL csg[DIMMTR];
	REAL csb[DIMMTR];
	REAL eta[DIMMTR];   // Refractive index.
	REAL gls[DIMMTR];   // Specular glossiness
	REAL rgh[DIMMTR];   // Random roughness.
} mtrlst;

// Light list "common block"
// Size: 13 * DIMLGT + 4 words = 433
#define MAXLGT 32
#define DIMLGT (MAXLGT+1)

struct
{
	int nlgt;           // Number of lights.
	REAL car;           // An ambient light colour.
	REAL cag;
	REAL cab;
	REAL ox[DIMLGT];    // Location (eye space)
	REAL oy[DIMLGT];
	REAL oz[DIMLGT];
	REAL clr[DIMLGT];   // Light colour/intensity
	REAL clg[DIMLGT];
	REAL clb[DIMLGT];
	int dir[DIMLGT];    // Light directionality type
	REAL dx[DIMLGT];    // Light pointing vector
	REAL dy[DIMLGT];
	REAL dz[DIMLGT];
	REAL tgt[DIMLGT];   // Directionality tightness / main angle
	REAL tg2[DIMLGT];   // Directionality falloff angle
	REAL rad[DIMLGT];   // Light radius (treat as sphere of light)
} lgtlst;

// Screen "common block"
// Size: 3 * DIMPXL + 15 words = 12318
#define MAXPXL 4096
#define DIMPXL (MAXPXL+1)

struct
{
	int nx, ny;         // Resolution at which to compute.
	REAL fpd;           // Focal plane distance from lens center.
	REAL aper;          // Aperture on image plane (in X direction).
	REAL focal;         // Focal length of lens.
	REAL fstop;         // F-Stop setting of lens.
	REAL liner[DIMPXL]; // Result image line
	REAL lineg[DIMPXL];
	REAL lineb[DIMPXL];
	int maxdep;         // Maximum desired ray tree depth
	REAL minwgt;        // Minimum ray weight to bother with.
	REAL colr;          // Colour computed for current ray by trace()
	REAL colg;
	REAL colb;
	int os;             // Oversampling factor
	int accel;          // Acceleration scheme.
	int shadow;         // Cast shadows if > 0, with this many feeler rays
	REAL backr;         // Background colour
	REAL backg;         
	REAL backb;
	REAL bkri, bkrf, bkrd;
	REAL bkgi, bkgf, bkgd;
	REAL bkbi, bkbf, bkbd;
	int bkt;            // Background type
	int curx, cury;     // Current screen x,y (1 based)
} screen;

// Statistics "common block"
// Size: 8 words.
struct
{
	int maxdep;         // Maximum ray depth traced.
	int maxray;         // Maximum rays in a tree.
	int maxins;         // Maximum intersections used.
	int totray;         // Total number of rays spawned.
	int totfel;         // Total number of feeler rays.
	int nacc;           // Number of primitives indexed in acceleration structure.
	int noacc;          // Number of primitives not indexed.
	REAL nsects;        // Total intersection tests (int could easily overflow!)
} stats;

// Acceleration structure "common block"
// For typical scenes, the "max cells" would be best around 20 to 40.
// (Maybe - my tests say 10 is good). But the memory cost is quite high.
// Likewise, it would be good to allow largish primitives to be indexed,
// but the maximum possible list size goes up as the cube of the largest
// number of cells allowed in any direction for the primitive to be accepted
// for indexing. So the memory cost is (potentially) very high. It would be
// possible to use the same list for more than one grid cell if it is identical
// to that for another cell. We don't try to detect that, though.
// If the ratio of the biggest dimension range to the smallest exceeds
// MXDMRT, adjust the ranges for calculating the grid parameters. This seems
// to be always a good idea (hence set to 1.0).
//
// Size: CELDIM + 2 * LSTDIM + 13 words = 1331 + 2 * 8001 + 13 = 17346 words.

#define MXCELX 10       // Max cells in X direction. (10)
#define MXCELY 10       // Max cells in Y direction. (10)
#define MXCELZ 10       // Max cells in Z direction. (10)
#define DMCELX (MXCELX+1)
#define DMCELY (MXCELY+1)
#define DMCELZ (MXCELZ+1)
#define CELDIM (DMCELX*DMCELY*DMCELZ)
#define MXCELS (MXCELX*MXCELY*MXCELZ)
#define MXPCVR 2        // Max cells in a direction a primitive can cover if it is to be indexed. (2)
#define MAXLST (MXPCVR*MXPCVR*MXPCVR*MXCELS)
#define LSTDIM (MAXLST+1)
#define MXDMRT 1.0     // Max / Min dimension ratio to start adjusting scales at. 

struct
{
	int anx, any, anz;  // Number of cells in each direction.
	REAL axl, axh;      // Eye space scene bounds in X.
	REAL ayl, ayh;      // Eye space scene bounds in Y.
	REAL azl, azh;      // Eye space scene bounds in Z.
	REAL dxs, dys, dzs; // Eye space scene ranges in X, Y & Z.
	int cells[CELDIM];  // Space subdivision cells.
	int lprim[LSTDIM];  // Primitives in cell list.
	int lnext[LSTDIM];  // Next in list.
	int nlist;          // Used list element count.
} accel;


// Grid traversal info "common block"
// Size: 10 words.
struct
{
	REAL ox, oy, oz;    // Ray origin for grid traversal.
	REAL dx, dy, dz;    // Ray direction unit vector for grid traversal.
	int ninst;          // Index of next free intersection list element.
	int asect;          // Index of nearest intersected primitive (zero if none).
	REAL tmin;          // Parameter of nearest intersection from grid traversal.
	int urayid;			// Use ray id based acceleration if TRUE.
} vscacc;

// Total "common block" data size:
//   515 ray stack
//  6671 ray list
//  6696 intersections
//  5126 primitives
//  4101 spheres
//  5192 triangles
//   463 materials
//   433 lights
// 12318 screen
//     8 stats
// 17346 acceleration
//    10 traversal
// =====
// 59599 words (of 131008 max available.)


// Prototypes

// Database
void initdb( void );
void initbk( void );
void initlg( void );
void initri( int i );
int readdb( char* scene );
// Tracing
void initln( void );
void trcln( int y );
void outln( int y, image_data* image );
void trace( void );
void cpyins( int i, int j );
// Intersection test/find
void gensct( REAL ox, REAL oy, REAL oz, REAL dx, REAL dy, REAL dz, int ninst, int& asect, REAL& tval, int urayid );
void celsct( REAL ox, REAL oy, REAL oz, REAL dx, REAL dy, REAL dz, int ninst, int& asect, REAL& tval, int urayid );
void sphins( REAL ox, REAL oy, REAL oz, REAL dx, REAL dy, REAL dz, int i, int& sect, REAL& t, int iprim );
void triins( REAL ox, REAL oy, REAL oz, REAL dx, REAL dy, REAL dz, int i, int& sect, REAL& t, int iprim );
// Shading
void shade( int i, int matl );
void ltfeel( int srfidx, REAL px, REAL py, REAL pz, REAL lx, REAL ly, REAL lz, REAL& lattr, REAL& lattg, REAL& lattb );
void calcbk( void );
// Statistics
void initst( void );
// Acceleration (voxel grid)
void stacel( void );
void addlst( int i, int j, int k, int p );
void grdtrv( REAL ux, REAL uy, REAL uz, REAL vx, REAL vy, REAL vz, int nx, int ny, int nz );
void viscel( int x, int y, int z, int& quit );
void boxins( REAL ox, REAL oy, REAL oz, REAL dx, REAL dy, REAL dz, int& sect, REAL& xi, REAL& yi, REAL& zi );
void sphbnd( int i, REAL& pxl, REAL& pxh, REAL& pyl, REAL& pyh, REAL& pzl, REAL& pzh );
void tribnd( int i, REAL& pxl, REAL& pxh, REAL& pyl, REAL& pyh, REAL& pzl, REAL& pzh );
// Utility functions
REAL ranhlf( void );
void nrmvec( REAL& dx, REAL& dy, REAL& dz );

//****************************//
// MAIN LINE AND IMAGE OUTPUT //
//****************************//

int main( int argc, char** argv )
{
	char* scene; // Name of scene file
	char* image; // Name of image file
	image_data* out_image;
	int y;

	printf( "\n" );
	printf( " Toy Ray Tracer Mark 2 - Version 0.1 - C++ Implementation\n" );
	printf( " ========================================================\n" );
	printf( "\n" );

	if( argc != 3 ){
		fprintf( stderr, "Usage: trt2 scenefile imagefile\n" );
		return( 1 );
	}

	scene = argv[1];
	image = argv[2];

	// Initialise the scene database.
	initdb();

	// Initialise statistics.
	initst();

	// Read the scene description and add to the scene database.
	if( readdb( scene ) != 0 ){
		fprintf( stderr, "Error: Failed to read scene: %s\n", scene );
		return( 1 );
	}

	// Print information
	printf( "\nScreen specification:\n" );
	if( screen.accel == NOACL )
		printf( "Acceleration scheme     = NONE\n"  );
	else
		printf( "Acceleration scheme     = VOXEL GRID SPACE SUBDIVISION\n" );
	if( screen.shadow == 0 )
		printf( "Casting shadows         = NO\n" );
	else
		printf( "Casting shadows         = YES  with %d feelers.\n", screen.shadow );
	printf( "Output pixels           = %d / %d\n", screen.nx, MAXPXL );
	printf( "Output lines            = %d / %d\n", screen.ny, MAXPXL );
	printf( "Oversample rate         = %d x %d = %d\n", screen.os, screen.os, screen.os*screen.os );
	printf( "Aperture                = %f\n", screen.aper );
	printf( "Focal plane distance    = %f\n", screen.fpd );
	printf( "Focal length of lens    = %f\n", screen.focal );
	printf( "F-stop of lens          = %f\n", screen.fstop );
	if( screen.bkt == BKCNST ){
		printf( "Background type         = CONSTANT\n"  );
		printf( "Background colour       = %f,%f,%f\n", screen.bkri, screen.bkgi, screen.bkbi );
	}
	else if( screen.bkt == BKVERT ){
		printf( "Background type         = VERTICAL GRADE\n"  );
		printf( "Background start colour = %f,%f,%f\n", screen.bkri, screen.bkgi, screen.bkbi );
		printf( "Background end colour   = %f,%f,%f\n", screen.bkrf, screen.bkgf, screen.bkbf );
	}
	else if( screen.bkt == BKHORZ ){
		printf( "Background type         = HORIZONTAL GRADE\n"  );
		printf( "Background start colour = %f,%f,%f\n", screen.bkri, screen.bkgi, screen.bkbi );
		printf( "Background end colour   = %f,%f,%f\n", screen.bkrf, screen.bkgf, screen.bkbf );
	}

	printf( "\nConstraints:\n" );
	printf( "Maximum ray tree depth  = %d\n", screen.maxdep );
	printf( "Min significant weight  = %f\n", screen.minwgt );

	printf( "\nDatabase statistics:\n" );
	printf( "Number of primitives    = %d / %d\n", prmlst.nprm, MAXPRM );
	printf( "Number of spheres       = %d / %d\n", sphlst.nsph, MAXSPH );
	printf( "Number of triangles     = %d / %d\n", trilst.ntri, MAXTRI );
	printf( "Number of materials     = %d / %d\n", mtrlst.nmtr, MAXMTR );
	printf( "Number of lights        = %d / %d\n", lgtlst.nlgt, MAXLGT );

	// Create an acceleration structure - if accelerating
	if( screen.accel == SUBACL )
		stacel();

	// Create an output image to write to a file
	out_image = image_data_create( screen.nx, screen.ny, 3, IMG_FMT_BYTE );
	if( ! out_image ){
		printf( "image_data_create() FAILED!\n" );
		return 1;
	}

	// Step over the lines of the output image. Trace all rays needed to generate the line.
	// Shade the line.
	// Output the line.
	printf( "\nTracing...\n" );
	for( y=1; y<=screen.ny; y++ ){
	  printf( "... line %d\r", y ); fflush( stdout );
		screen.cury = y;
		initln();
		trcln( y );
		outln( y, out_image );
	}
	printf( "\n" );

	// Write the output image to an image file.
	if( ! image_write_file( image, out_image ) ){
		printf( "image_write_file() FAILED!\n" );
		return 1;
	}

	// Print statistics.
	printf( "\nExecution statistics:\n" );
	printf( "Maximum rays in a tree  = %d / %d\n", stats.maxray, MAXRAY );
	printf( "Maximum intersections   = %d / %d (per ray)\n", stats.maxins, MAXINS );
	printf( "Total intersect tests   = %.0f\n", stats.nsects );
	printf( "Total scene rays traced = %d\n", stats.totray );
	printf( "Total light feelers     = %d\n", stats.totfel );

	return( 0 );
}

void outln( int y, image_data* image )
//------------------------------------
// Output current screen line to a file of some kind.
{
	int i, x;
	unsigned char* lp;
	unsigned char vr, vg, vb;
	REAL fvr, fvg, fvb;

	lp = image->data.uc + ( screen.ny - y ) * image_stride( image );
	for( i=1,x=0; i<=screen.nx; i++ ){
		fvr = screen.liner[i];
		if( fvr < 0.0 )fvr = 0.0; else if( fvr > 1.0f )fvr = 1.0f;
		vr = (unsigned char)( 255 * fvr );
		fvg = screen.lineg[i];
		if( fvg < 0.0 )fvg = 0.0; else if( fvg > 1.0f )fvg = 1.0f;
		vg = (unsigned char)( 255 * fvg );
		fvb = screen.lineb[i];
		if( fvb < 0.0 )fvb = 0.0; else if( fvb > 1.0f )fvb = 1.0f;
		vb = (unsigned char)( 255 * fvb );
		*lp++ = vr;
		*lp++ = vg;
		*lp++ = vb;
		++x;
	}
}

//*************//
// RAY TRACING //
//*************//

void initln( void )
//-----------------
// Initialize the output line.
{
	int i;

	for( i=1; i<=screen.nx; i++ ){
		screen.liner[i] = 0.0;
		screen.lineg[i] = 0.0;
		screen.lineb[i] = 0.0;
	}
}

void initst( void )
//-----------------
// Initialise statistics info.
{
	stats.maxdep = 0;
	stats.maxray = 0;
	stats.maxins = 0;
	stats.totray = 0;
	stats.totfel = 0;
	stats.nacc = 0;
	stats.noacc = 0;
	stats.nsects = 0;
}

void trcln( int y )
//-----------------
// Trace all rays needed to create line y.
{
	REAL left, right, bottom, top, dp, dpx, dpy, rns, rmag;
	REAL fy, fx, ds, fys, fysj, fxs, fxsj, dsx, dsy, lrad;
	REAL etheta, erad, ex, ey;
	int x, subx, suby, ns, os2, uos;

	// Calculate the pixel center coordinates at the edges of the aperture
	dp = screen.aper / (REAL)screen.nx;
	left = ( -screen.aper + dp ) * 0.5f;
	right = ( screen.aper - dp ) * 0.5f;
	bottom = left * (REAL)screen.ny / (REAL)screen.nx;
	top = right * (REAL)screen.ny / (REAL)screen.nx;

	// Calculate the reciprocal of the magnification of the lens
	if( screen.fpd < ( screen.focal + 0.5 ) )
		screen.fpd = screen.focal + 0.5;
	rmag = ( screen.fpd - screen.focal ) / screen.focal ;

	// Scale aperture edge coordinates to focal plane edge coordinates
	left *= rmag;
	right *= rmag;
	bottom *= rmag;
	top *= rmag;

	// Calculate the effective lens radius
	lrad = 0.5 * screen.focal / screen.fstop;

	// Calculate constants for subsampling
	if( screen.os == 0 ){
		ds = 0.0;
		ns = 1;
		rns = 1.0f;
		os2 = 0;
		uos = 1;
	}
	else{
		ds = 1.0f / screen.os;
		ns = screen.os * screen.os;
		rns = 1.0f / ns;
		os2 = screen.os / 2;
		uos = screen.os;
	}

	// Calculate the X and Y steps on the focal plane between pixel centers.
	dpx = ( right - left ) / (REAL)( screen.nx - 1 );
	dpy = ( top - bottom ) / (REAL)( screen.ny - 1 );
	dsx = ds * dpx;
	dsy = ds * dpy;

	// Calculate Y coordinate of current line center
	fy = bottom + ( y - 1 ) * dpy;

	// Step over the pixels in the line.
	for( x=1; x<=screen.nx; x++ ){
		screen.curx = x;
		fx = left + ( x - 1 ) * dpx;

		// Step over Y subsamples - and jitter them
		for( suby=1; suby<=uos; suby++ ){
			fys = ( suby - os2 - 1 ) * dsy + fy;
			fysj = fys + ranhlf() * dsy;
			
			// Step over X subsamples - and jitter them.
			for( subx=1; subx<=uos; subx++ ){
				fxs = ( subx - os2 - 1 ) * dsx + fx;
				fxsj = fxs + ranhlf() * dsx;
				
				// Begin tracing rays. There are initially no rays. And no intersections.
				raylst.nray = 0;
        raystk.nray = 0;
				inslst.nins = 0;
				screen.colr = 0.0;
				screen.colg = 0.0;
				screen.colb = 0.0;
				
				// Generate a screen ray from a randomly chosen point on the lens (ex,ey,0) 
				// through the focal plane position (fx,fy,screen.fpd)
				++raylst.nray;
				++raylst.raynum;
				etheta = ranhlf() * TRTPI;
				erad = ( ranhlf() + 0.5 ) * lrad;
				ex = erad * cos( etheta );
				ey = erad * sin( etheta );
				raylst.rayid[raylst.nray] = raylst.raynum;
				raylst.type[raylst.nray] = EYERAY;
				raylst.isect[raylst.nray] = 0;
				raylst.wgtr[raylst.nray] = 1.0;
				raylst.wgtg[raylst.nray] = 1.0;
				raylst.wgtb[raylst.nray] = 1.0;
				raylst.dep[raylst.nray] = 1;
				raylst.ox[raylst.nray] = ex;
				raylst.oy[raylst.nray] = ey;
				raylst.oz[raylst.nray] = 0.0;
				raylst.dx[raylst.nray] = fxsj - ex;
				raylst.dy[raylst.nray] = fysj - ey;
				raylst.dz[raylst.nray] = screen.fpd;
				nrmvec( raylst.dx[raylst.nray], raylst.dy[raylst.nray], raylst.dz[raylst.nray] );
				++stats.totray;
				
				// Push the ray onto the ray stack.
				++raystk.nray;
				raystk.ray[raystk.nray] = raylst.nray;
				if( raylst.nray > stats.maxray )
					stats.maxray = raylst.nray;
				
				// While there are rays on the ray stack, pop a ray off
				while( raystk.nray > 0 ){
					raystk.cray = raystk.ray[raystk.nray];
					--raystk.nray;
					
					// Find intersections of this current ray with primitives.
					// When there is an intersection, calculate a shade at the intersection point
					// and add this contribution to the pixel colour.
					// Add new reflected and transmitted rays to the ray stack if appropriate.
					trace();
				}

				screen.liner[x] += screen.colr;
				screen.lineg[x] += screen.colg;
				screen.lineb[x] += screen.colb;
			}
		}

		screen.liner[x] *= rns;
		screen.lineg[x] *= rns;
		screen.lineb[x] *= rns;
	}
}

void trace( void )
//----------------
// Intersect the "current ray" (raystk.cray) with all the primitives in the scene.
// If there is an intersection record the intersection in the intersection list.
// Set the current ray intersection index to the new intersection.
// Find a shade at the intersection and add that contribution to the sample colour.
// Spawn new reflected and transmitted rays if appropriate. Push them on the ray stack.
{
	REAL ox, oy, oz;
	REAL dx, dy, dz;
	REAL ndoti, eta, reta, cos1, cos22, k2, swgt, tval;
	REAL rwgt, rwgtr, rwgtg, rwgtb;
	REAL twgt, twgtr, twgtg, twgtb;
	REAL cwgtr, cwgtg, cwgtb;
	int ninst, prim, matl, asect, dep, ttyp;

	// Get useful things about the current ray.
	ox = raylst.ox[raystk.cray];
	oy = raylst.oy[raystk.cray];
	oz = raylst.oz[raystk.cray];
	dx = raylst.dx[raystk.cray];
	dy = raylst.dy[raystk.cray];
	dz = raylst.dz[raystk.cray];
	cwgtr = raylst.wgtr[raystk.cray];
	cwgtg = raylst.wgtg[raystk.cray];
	cwgtb = raylst.wgtb[raystk.cray];
	dep = raylst.dep[raystk.cray];
	if( dep > stats.maxdep )
		stats.maxdep = dep;

	// Test for intersections. If there are any, we need the nearest.
	// If there was an intersection, its position and surface normal will be in
	// inslst.XXX[inslst.nins+1] after this.
	ninst = inslst.nins + 1;
	if( screen.accel == NOACL )
		gensct( ox, oy, oz, dx, dy, dz, ninst, asect, tval, TRUE );
	else
		celsct( ox, oy, oz, dx, dy, dz, ninst, asect, tval, TRUE );

	// If there actually was an intersection, update the intersection count.
	// Create new reflected and transmitted rays if appropriate.
	if( asect != 0 ){
		inslst.nins = ninst;              // Update intersection count
		if( ninst > stats.maxins )
			stats.maxins = ninst;
		prim = inslst.prim[ninst];        // Primitive index we have intersected with
		matl = prmlst.matter[prim];       // Material index primitive is made from
		rwgtr = mtrlst.kr[matl] * mtrlst.cdr[matl] * cwgtr;  // Weight of a reflected ray at this stage
		rwgtg = mtrlst.kr[matl] * mtrlst.cdg[matl] * cwgtg;
		rwgtb = mtrlst.kr[matl] * mtrlst.cdb[matl] * cwgtb;
		twgtr = mtrlst.kt[matl] * mtrlst.cdr[matl] * cwgtr;  // Weight of a transmitted ray at this stage
		twgtg = mtrlst.kt[matl] * mtrlst.cdg[matl] * cwgtg;
		twgtb = mtrlst.kt[matl] * mtrlst.cdb[matl] * cwgtb;
		ttyp = inslst.ttyp[ninst];        // Transmitted ray entering or leaving primitive?
		eta = mtrlst.eta[matl];           // Refractive index of material transmitted ray is entering or leaving.

		// Use maximum of colour component weights to decide if to terminate
		rwgt = rwgtr;
		if( rwgtg > rwgt )
			rwgt = rwgtg;
		if( rwgtb > rwgt )
			rwgt = rwgtb;
		twgt = twgtr;
		if( twgtg > twgt )
			twgt = twgtg;
		if( twgtb > twgt )
			twgt = twgtb;

		// Flip the normal?
		if( ( dx * inslst.nx[ninst] + dy * inslst.ny[ninst] + dz * inslst.nz[ninst] ) > 0.0 ){
			inslst.nx[ninst] = -inslst.nx[ninst];
			inslst.ny[ninst] = -inslst.ny[ninst];
			inslst.nz[ninst] = -inslst.nz[ninst];
		}

		// Calculate the colour at the intersection.
		shade( ninst, matl );
		swgt = 1.0f - mtrlst.kr[matl] - mtrlst.kt[matl];
		if( swgt < 0.0 )
			swgt = 0.0;
		screen.colr += inslst.colr[ninst] * swgt * cwgtr;
		screen.colg += inslst.colg[ninst] * swgt * cwgtg;
		screen.colb += inslst.colb[ninst] * swgt * cwgtb;

		// Spawn new rays (recurse, effectively) if we are not too deep.
		if( dep < screen.maxdep ){
			++dep;
			
			// New reflected ray
			if( rwgt > screen.minwgt ){
				ndoti = 2 * ( inslst.nx[ninst] * dx + inslst.ny[ninst] * dy + inslst.nz[ninst] * dz );
				++raylst.nray;
				++raylst.raynum;
				raylst.rayid[raylst.nray] = raylst.raynum;
				raylst.type[raylst.nray] = RFLRAY;
				raylst.isect[raylst.nray] = 0;
				raylst.wgtr[raylst.nray] = rwgtr;
				raylst.wgtg[raylst.nray] = rwgtg;
				raylst.wgtb[raylst.nray] = rwgtb;
				raylst.dep[raylst.nray] = dep;
				raylst.ox[raylst.nray] = inslst.px[ninst];
				raylst.oy[raylst.nray] = inslst.py[ninst];
				raylst.oz[raylst.nray] = inslst.pz[ninst];
				raylst.dx[raylst.nray] = dx - ndoti * inslst.nx[ninst];
				raylst.dy[raylst.nray] = dy - ndoti * inslst.ny[ninst];
				raylst.dz[raylst.nray] = dz - ndoti * inslst.nz[ninst];
				nrmvec( raylst.dx[raylst.nray], raylst.dy[raylst.nray], raylst.dz[raylst.nray] );
				inslst.rray[ninst] = raylst.nray;
				++raystk.nray;
				raystk.ray[raystk.nray] = raylst.nray;
				++stats.totray;
				if( raylst.nray > stats.maxray )
					stats.maxray = raylst.nray;
			}
			
			// New transmitted ray
			if( twgt > screen.minwgt ){
				if( ttyp == ENTER )
					reta = eta;
				else
					reta = 1.0f / eta;
				cos1 = -( inslst.nx[ninst] * dx + inslst.ny[ninst] * dy + inslst.nz[ninst] * dz );
				cos22 = 1.0f - reta * reta * ( 1.0f - cos1 * cos1 );
				if( cos22 > 0.0 ){ // Otherwise, we have total internal reflection...
					k2 = reta * cos1 - sqrt( cos22 );
					++raylst.nray;
					++raylst.raynum;
					raylst.rayid[raylst.nray] = raylst.raynum;
					raylst.type[raylst.nray] = TRNRAY;
					raylst.isect[raylst.nray] = 0;
					raylst.wgtr[raylst.nray] = twgtr;
					raylst.wgtg[raylst.nray] = twgtg;
					raylst.wgtb[raylst.nray] = twgtb;
					raylst.dep[raylst.nray] = dep;
					raylst.ox[raylst.nray] = inslst.px[ninst];
					raylst.oy[raylst.nray] = inslst.py[ninst];
					raylst.oz[raylst.nray] = inslst.pz[ninst];
					raylst.dx[raylst.nray] = reta * dx + k2 * inslst.nx[ninst];
					raylst.dy[raylst.nray] = reta * dy + k2 * inslst.ny[ninst];
					raylst.dz[raylst.nray] = reta * dz + k2 * inslst.nz[ninst];
					nrmvec( raylst.dx[raylst.nray], raylst.dy[raylst.nray], raylst.dz[raylst.nray] );
					inslst.tray[ninst] = raylst.nray;
					++raystk.nray;
					raystk.ray[raystk.nray] = raylst.nray;
					++stats.totray;
					if( raylst.nray > stats.maxray )
						stats.maxray = raylst.nray;
				}
			}
		}
	}

	// If this ray doesn't hit anything, add in the background colour.
	else{
		calcbk();
		screen.colr += screen.backr * cwgtr;
		screen.colg += screen.backg * cwgtg;
		screen.colb += screen.backb * cwgtb;
	}
}

void calcbk( void )
//-----------------
// Calculate the background colour for graded backgrounds
{
	int cx, cy;

	cx = screen.curx - 1;
	cy = screen.cury - 1;
	if( screen.bkt == BKCNST ){
		screen.backr = screen.bkri;
		screen.backg = screen.bkgi;
		screen.backb = screen.bkbi;
	}
	else if( screen.bkt == BKVERT ){
		screen.backr = screen.bkri + screen.bkrd * cy;
		screen.backg = screen.bkgi + screen.bkgd * cy;
		screen.backb = screen.bkbi + screen.bkbd * cy;
	}
	else if( screen.bkt == BKHORZ ){
		screen.backr = screen.bkri + screen.bkrd * cx;
		screen.backg = screen.bkgi + screen.bkgd * cx;
		screen.backb = screen.bkbi + screen.bkbd * cx;
	}
	if( screen.backr < 0.0 )screen.backr = 0.0; else if( screen.backr > 1.0 )screen.backr = 1.0;
	if( screen.backg < 0.0 )screen.backg = 0.0; else if( screen.backg > 1.0 )screen.backg = 1.0;
	if( screen.backb < 0.0 )screen.backb = 0.0; else if( screen.backb > 1.0 )screen.backb = 1.0;
}

//*********//
// SHADING //
//*********//

void shade( int i, int matl )
//---------------------------
// Find the colour at intersection i where material matl is to be found.
{
	int l, f;
	REAL dx, dy, dz, d, s, vx, vy, vz, hx, hy, hz, cr, cg, cb, lattr, lattg, lattb, p;
	REAL nx, ny, nz, a, lxf, lyf, lzf, lattrf, lattgf, lattbf;
	
	// Initialize intersection colour to ambient term.
	inslst.colr[i] = mtrlst.car[matl] * lgtlst.car;
	inslst.colg[i] = mtrlst.cag[matl] * lgtlst.cag;
	inslst.colb[i] = mtrlst.cab[matl] * lgtlst.cab;

	// Step over lights. Find contribution to intersection colour for each light.
	for( l=1; l<=lgtlst.nlgt; l++ ){
		
		// See if this light is partially shadowed.
		if( screen.shadow == 0 ){       // Don't do shadows.
			lattr = 1.0f;
			lattg = 1.0f;
			lattb = 1.0f;
		}
		else if( screen.shadow == 1 || lgtlst.rad[l] < EPS ){  // 1 feeler ray = point light source.
			ltfeel( inslst.prim[i], inslst.px[i], inslst.py[i], inslst.pz[i], lgtlst.ox[l], lgtlst.oy[l], lgtlst.oz[l], lattr, lattg, lattb );
		}
		else{                           // Multiple feeler rays = uniform sphere of light.
			lattr = 0.0;
			lattg = 0.0;
			lattb = 0.0;
			for( f=1; f<=screen.shadow; f++ ){
				lxf = lgtlst.ox[l] + ranhlf() * lgtlst.rad[l];
				lyf = lgtlst.oy[l] + ranhlf() * lgtlst.rad[l];
				lzf = lgtlst.oz[l] + ranhlf() * lgtlst.rad[l];
				ltfeel( inslst.prim[i], inslst.px[i], inslst.py[i], inslst.pz[i], lxf, lyf, lzf, lattrf, lattgf, lattbf );
				lattr += lattrf;
				lattg += lattgf;
				lattb += lattbf;
			}
			lattr /= screen.shadow;
			lattg /= screen.shadow;
			lattb /= screen.shadow;
		}
		
		// Find unit vector from surface intersection to light.
		dx = lgtlst.ox[l] - inslst.px[i];
		dy = lgtlst.oy[l] - inslst.py[i];
		dz = lgtlst.oz[l] - inslst.pz[i];
		nrmvec( dx, dy, dz );

		// If the light is directional, modify light attenuation based on angle btw light-surface vector and light direction.
		if( lgtlst.dir[l] != LGTCON ){
			p =  dx * lgtlst.dx[l] + dy * lgtlst.dy[l] + dz * lgtlst.dz[l];
			if( p < 0.0 )
				p = 0.0;
			else{
				if( lgtlst.dir[l] == LGTCOS ){          // Raised cosine sort ...
					p = pow( p, lgtlst.tgt[l] );
				}
				else if( lgtlst.dir[l] == LGTBRN ){     // Barn door sort ...
					a = acos( p );
					if( a < lgtlst.tgt[l] )
						p = 1.0;
					else if( a > lgtlst.tg2[l] )
						p = 0.0;
					else{
						p = ( ( a - lgtlst.tgt[l] ) / ( lgtlst.tg2[l] - lgtlst.tgt[l] ) );
						p = 1.0 - ( p * p * ( 3 - 2 * p ) ); // Nice smooth Hermite ...
					}
				}
			}
			lattr *= p;
			lattg *= p;
			lattb *= p;
		}

		// Get the normal and perturb it if there is some roughness.
		nx = inslst.nx[i];
		ny = inslst.ny[i];
		nz = inslst.nz[i];
		if( mtrlst.rgh[matl] > 0.0 ){
			nx += ranhlf() * mtrlst.rgh[matl];
			ny += ranhlf() * mtrlst.rgh[matl];
			nz += ranhlf() * mtrlst.rgh[matl];
			nrmvec( nx, ny, nz );
		}
		
		// Diffuse term.
		d = dx * nx + dy * ny + dz * nz;
		if( d < 0.0 )
			d = 0.0;
		
		// Specular term.
		vx = - inslst.px[i];
		vy = - inslst.py[i];
		vz = - inslst.pz[i];
		nrmvec( vx, vy, vz );
		
		hx = 0.5f * ( dx + vx );
		hy = 0.5f * ( dy + vy );
		hz = 0.5f * ( dz + vz );
		
		s = hx * nx + hy * ny + hz * nz;
		if( s < 0.0 )
			s = 0.0;
		else
			s = pow( s, mtrlst.gls[matl] );
		
		// Find colour due to this light.
		cr = lgtlst.clr[l] * lattr * ( d * mtrlst.cdr[matl] + s * mtrlst.csr[matl] );
		cg = lgtlst.clg[l] * lattg * ( d * mtrlst.cdg[matl] + s * mtrlst.csg[matl] );
		cb = lgtlst.clb[l] * lattb * ( d * mtrlst.cdb[matl] + s * mtrlst.csb[matl] );

		// Add to intersection colour.
		inslst.colr[i] += cr;
		inslst.colg[i] += cg;
		inslst.colb[i] += cb;
	}
}

void ltfeel( int srfidx, REAL px, REAL py, REAL pz, REAL lx, REAL ly, REAL lz, REAL& lattr, REAL& lattg, REAL& lattb )
//--------------------------------------------------------------------------------------------------------------------
// Generate a light feeler ray from surface point (px,py,pz) to light (lx,ly,lz).
// See what this intersects and generate an appropriate (ish) light attenuation factor.
// This tries to deal with transparent objects as well as opaque ones. It does so 
// incorrectly, as refraction cannot be taken into account.
{
	REAL dx, dy, dz, ds, acatr, acatg, acatb, mincat;
	int prim, asect, matl;
	REAL tval, pox, poy, poz;

	// Find distance between surface point & light and a normalized vector
	// pointing from the surface to the light. If point is very close to light, there is
	// no attenuation.
	lattr = 1.0f;
	lattg = 1.0f;
	lattb = 1.0f;
	++stats.totfel;

	dx = lx - px;
	dy = ly - py;
	dz = lz - pz;
	ds = dx * dx + dy * dy + dz * dz;
	if( ds < MINEPS ){
		lattr = 1.0f;
		lattg = 1.0f;
		lattb = 1.0f;
		return;
	}
	ds = sqrt( ds );
	dx /= ds;
	dy /= ds;
	dz /= ds;

	// Find all intersecting primitives. If a primitive is made of fully
	// opaque material, or the accumulated attenuation is very great, abandon search and return latt=0.
	acatr = 1.0f;
	acatg = 1.0f;
	acatb = 1.0f;

	pox = px;
	poy = py;
	poz = pz;

	for(;;){
		if( screen.accel == NOACL )
			gensct( pox, poy, poz, dx, dy, dz, SHDINS, asect, tval, FALSE );
		else
			celsct( pox, poy, poz, dx, dy, dz, SHDINS, asect, tval, FALSE );

		// See if there is an intersection between the surface point being shaded and the light (not beyond it).
		if( asect == 0 || tval < 0.0 || tval > ds )
			break;
		if( asect != 0 ){
			// Check intersection is not with primitive we are finding shadows for.
			// (Primitives cannot shade themselves - difficult numerical issues if they could!).
			if( asect != srfidx ){
				prim = inslst.prim[SHDINS];       // Primitive index we have intersected with
				matl = prmlst.matter[prim];       // Material index primitive is made from
				if( mtrlst.kt[matl] < screen.minwgt ){
					lattr = 0.0;
					lattg = 0.0;
					lattb = 0.0;
					return;
				}
							
				acatr *= mtrlst.kt[matl] * mtrlst.cdr[matl];
				acatg *= mtrlst.kt[matl] * mtrlst.cdg[matl];
				acatb *= mtrlst.kt[matl] * mtrlst.cdb[matl];

				mincat = acatr;
				if( acatg < mincat )
					mincat = acatg;
				if( acatb < mincat )
					mincat = acatb;
				if( mincat < screen.minwgt ){
					lattr = 0.0;
					lattg = 0.0;
					lattb = 0.0;
					return;
				}
			}

			// Found an intersection (possibly with self).
			// Move a tad further along the feeler and see if we intersect anything else.
			// PRECISION!
			pox = inslst.px[SHDINS] + EPS * dx;
			poy = inslst.py[SHDINS] + EPS * dy;
			poz = inslst.pz[SHDINS] + EPS * dz;
		}
	}

	lattr = acatr;
	lattg = acatg;
	lattb = acatb;
}

//**********************//
// INTERSECTION TESTING //
//**********************//

void gensct( REAL ox, REAL oy, REAL oz, REAL dx, REAL dy, REAL dz, int ninst, int& asect, REAL& tval, int urayid )
//----------------------------------------------------------------------------------------------------------------
// General intersection finder. Given ray with origin (ox,oy,oz) and unit direction vector
// (dx,dy,dz), the index of the next free element in the intersection list [ninst], find the
// intersection closest to (ox,oy,oz) for all primitives. If there is an intersection, return
// [asect] set to the index of the primitive we hit else set it to 0. 
// In the case of an intersection, its position and the surface
// normal at that position will be in inslst.XXX[ninst].
// If there is an intersection, return the ray parameter of the intersection in [tval].
// If [urayid], use ray id based intersection acceleration. This can't be done for shadow feelers.
{
	int i, sect, pidx;
	REAL t, tmin;

	asect = 0;
	raylst.isect[raystk.cray] = 0;
	for( i=1; i<=prmlst.nprm; i++ ){
		if( prmlst.rayid[i] != raylst.rayid[raystk.cray] || urayid == FALSE ){
			pidx = prmlst.prmidx[i];
			if( prmlst.prmtyp[i] == SPHPRM )
				sphins( ox, oy, oz, dx, dy, dz, pidx, sect, t, i );
			else if( prmlst.prmtyp[i] == TRIPRM )
				triins( ox, oy, oz, dx, dy, dz, pidx, sect, t, i );
			else
				fprintf( stderr, "INTERNAL ERROR: Got unknown primitive type in gensct().\n" );
			if( sect != 0 ){
				if( asect == 0 ){
					tmin = t;
					cpyins( ninst, NEWINS );
					raylst.isect[raystk.cray] = ninst;
					asect = i;
					prmlst.rayid[i] = raylst.rayid[raystk.cray];
				}
				else{
					if( t < tmin ){
						tmin = t;
						cpyins( ninst, NEWINS );
						raylst.isect[raystk.cray] = ninst;
						asect = i;
						prmlst.rayid[i] = raylst.rayid[raystk.cray];
					}
				}
			}
		}
	}
	if( asect != 0 )
		tval = tmin;
	else
		tval = 0.0;
}

void celsct( REAL ox, REAL oy, REAL oz, REAL dx, REAL dy, REAL dz, int ninst, int& asect, REAL& tval, int urayid )
//----------------------------------------------------------------------------------------------------------------
// Find an intersection using the acceleration structure.
// The arguments are the same as for gensct().
{
	REAL xcell, ycell, zcell;
	int bsect;
	REAL xi, yi, zi;
	int i, sect, pidx;
	REAL t, tmin;

	// Calculate the origin coordinates in the cell coordinate system.
	xcell = ( ( ox - accel.axl ) / accel.dxs ) * (REAL)accel.anx;
	ycell = ( ( oy - accel.ayl ) / accel.dys ) * (REAL)accel.any;
	zcell = ( ( oz - accel.azl ) / accel.dzs ) * (REAL)accel.anz;

	// Check we are inside the grid. Except for eye rays, we should always be.
	// If not, then see if the ray intersects the bounding box of the scene geometry.
	if( xcell < 0 || xcell >= MXCELX || ycell < 0 || ycell >= MXCELY || zcell < 0 || zcell >= MXCELZ ){
		boxins( ox, oy, oz, dx, dy, dz, bsect, xi, yi, zi );

		// The ray doesn't hit the bounding box. No intersection.
		if( bsect == 0 ){
			asect = 0;
			return;
		}

		// Use the intersection point to find the initial grid cell to traverse from.
		xcell = ( ( xi - accel.axl ) / accel.dxs ) * (REAL)accel.anx;
		ycell = ( ( yi - accel.ayl ) / accel.dys ) * (REAL)accel.any;
		zcell = ( ( zi - accel.azl ) / accel.dzs ) * (REAL)accel.anz;
	}

	// Save ray data for use by viscel()
	vscacc.asect = 0;
	vscacc.dx = dx;
	vscacc.dy = dy;
	vscacc.dz = dz;
	vscacc.ox = ox;
	vscacc.oy = oy;
	vscacc.oz = oz;
	vscacc.ninst = ninst;
	vscacc.urayid = urayid;

	// Traverse the cell grid. Each cell visited is handled by viscel()
	// Note correction to the direction vector for grid space!
	grdtrv( xcell, ycell, zcell, dx/accel.dxs, dy/accel.dys, dz/accel.dzs, accel.anx, accel.any, accel.anz );

	// Grid traversal ends either when we have fallen off the sides of the grid,
	// or viscel() has found an intersection. Return [asect] in either case.
	asect = vscacc.asect;

	// Now check all the primitives that were too large to be indexed in the
	// acceleration structure. One of these may have a closer intersection.
	if( asect != 0 )
		tmin = vscacc.tmin;

	for( i=1; i<=prmlst.nprm; i++ ){
		if( prmlst.accel[i] == 0 ){
			if( prmlst.rayid[i] != raylst.rayid[raystk.cray] || urayid == FALSE ){
				pidx = prmlst.prmidx[i];
				if( prmlst.prmtyp[i] == SPHPRM )
					sphins( ox, oy, oz, dx, dy, dz, pidx, sect, t, i );
				else if( prmlst.prmtyp[i] == TRIPRM )
					triins( ox, oy, oz, dx, dy, dz, pidx, sect, t, i );
				else
					fprintf( stderr, "INTERNAL ERROR: Got unknown primitive type in celsct().\n" );
				if( sect != 0 ){
					if( asect == 0 ){
						tmin = t;
						cpyins( ninst, NEWINS );
						raylst.isect[raystk.cray] = ninst;
						asect = i;
						prmlst.rayid[i] = raylst.rayid[raystk.cray];
					}
					else{
						if( t < tmin ){
							tmin = t;
							cpyins( ninst, NEWINS );
							raylst.isect[raystk.cray] = ninst;
							asect = i;
							prmlst.rayid[i] = raylst.rayid[raystk.cray];
						}
					}
				}
			}
		}
	}

	if( asect != 0 )
		tval = tmin;
	else
		tval = 0.0;
}

void viscel( int x, int y, int z, int& quit )
//-------------------------------------------
// Process grid cell (x,y,z). If there are any primitives in this cell, check them
// for intersection with the current ray (in vscacc.XXX). If an intersection is found,
// leave the primitive index in vscacc.asect with the position and normal stored in
// inslst.XXX[vscacc.ninst] in the usual way. And terminate grid traversal.
//(If we want to stop grid traversal at this cell, return [quit] non-zero, else 0.)
{
	int ci, li, lp, pidx;
	int sect, lasect, cx, cy, cz;
	REAL xcell, ycell, zcell, t, tmin;

	// See if there is a list of primitives for this cell. If not continue traversal.
  ci = x + y * accel.anx + z * (accel.anx*accel.any) + 1;
	li = accel.cells[ci];
	if( li == 0 ){
		vscacc.asect = 0;
		quit = 0;
		return;
	}

	// There is a list of primitives starting at list index li. Move over this list
	// looking for intersections with each primitive on the list in turn.
	lasect = 0;
	for(;;){
		lp = accel.lprim[li];
		if( prmlst.rayid[lp] != raylst.rayid[raystk.cray] || vscacc.urayid == FALSE ){
			pidx = prmlst.prmidx[lp];
			if( prmlst.prmtyp[lp] == SPHPRM )
				sphins( vscacc.ox, vscacc.oy, vscacc.oz, vscacc.dx, vscacc.dy, vscacc.dz, pidx, sect, t, lp );
			else if( prmlst.prmtyp[lp] == TRIPRM )
				triins( vscacc.ox, vscacc.oy, vscacc.oz, vscacc.dx, vscacc.dy, vscacc.dz, pidx, sect, t, lp );
			else
				fprintf( stderr, "INTERNAL ERROR: Got unknown primitive type in viscel().\n" );

			if( sect != 0 ){

				// Check that this intersection lies inside this grid cell.
				xcell = ( ( inslst.px[NEWINS] - accel.axl ) / accel.dxs ) * accel.anx;
				ycell = ( ( inslst.py[NEWINS] - accel.ayl ) / accel.dys ) * accel.any;
				zcell = ( ( inslst.pz[NEWINS] - accel.azl ) / accel.dzs ) * accel.anz;
				cx = (int)xcell;
				cy = (int)ycell;
				cz = (int)zcell;

				if( x == cx && y == cy && z == cz ){

					// Yes - a valid intersection. Is it the closest in this cell?
					if( lasect == 0 ){
						tmin = t;
						cpyins( vscacc.ninst, NEWINS );
						raylst.isect[raystk.cray] = vscacc.ninst;
						lasect = lp;
						prmlst.rayid[lp] = raylst.rayid[raystk.cray];
					}
					else{
						if( t < tmin ){
							tmin = t;
							cpyins( vscacc.ninst, NEWINS );
							raylst.isect[raystk.cray] = vscacc.ninst;
							lasect = lp;
							prmlst.rayid[lp] = raylst.rayid[raystk.cray];
						}
					}
				}
			}
		}

		li = accel.lnext[li];
		if( li == 0 )
			break;
	}

	// No intersection with primitives in this cell. Continue traversal.
	if( lasect == 0 ){
		vscacc.asect = 0;
		quit = 0;
		return;
	}

	// Intersection found. Terminate traversal.
	else{
		vscacc.asect = lasect;
		vscacc.tmin = tmin;
		quit = 1;
		return;
	}
}

void sphins( REAL ox, REAL oy, REAL oz, REAL dx, REAL dy, REAL dz, int i, int& sect, REAL& t, int iprim )
//------------------------------------------------------------------------------------------------------
// See if the ray from (ox,oy,oz) direction (dx,dy,dz) intersects sphere i. The primitive index is iprim.
// If so, set sect non-zero and set t to the intersection parameter.
// Otherwise set sect to zero.
// If there is an intersection save the position and normal in inslst.XXX[NEWINS].
{
	REAL ocx, ocy, ocz, rs;
	REAL b, c, d, t0, t1, tmin;
	
	// Init
	++stats.nsects; // Bump intersection test count.
	sect = 0;
	t = 0.0;
	
	// Solve implicit quadratic equation
	ocx = ox - sphlst.ox[i];
	ocy = oy - sphlst.oy[i];
	ocz = oz - sphlst.oz[i];
	rs = sphlst.rad[i];
	b = -2 * ( dx * ocx + dy * ocy + dz * ocz );
	c = ocx * ocx + ocy * ocy + ocz * ocz - rs * rs;
	d = b * b - 4 * c;
	
	// If discriminant -ve, no intersection.
	if( d >= 0.0 ){
		d = sqrt( d );
		
		// OK - Two possible intersections to be found - t0 and t1.
		// PRECISION!
		t1 = 0.5f * ( b + d ); // t1 must be > t0 ... t1 would be exiting the sphere if t0 also valid.
		if( t1 > MINEPS ){     // t1 is ahead, so we have an intersection. So t0 might be ahead and closer ... 
			tmin = t1;
			sect = 1;
			inslst.ttyp[NEWINS] = LEAVE; // This would be leaving the sphere.
			
			t0 = 0.5f * ( b - d );
			if( t0 > MINEPS ){ // t0 is ahead, we have a closer intersection ... 
				tmin = t0;
				inslst.ttyp[NEWINS] = ENTER; // This would be entering the sphere.
			}
			
			inslst.prim[NEWINS] = iprim;
			inslst.rray[NEWINS] = 0;
			inslst.tray[NEWINS] = 0;
			inslst.colr[NEWINS] = 0.0;
			inslst.colg[NEWINS] = 0.0;
			inslst.colb[NEWINS] = 0.0;
			inslst.px[NEWINS] = ox + tmin * dx;
			inslst.py[NEWINS] = oy + tmin * dy;
			inslst.pz[NEWINS] = oz + tmin * dz;
			inslst.nx[NEWINS] = ( inslst.px[NEWINS] - sphlst.ox[i] ) / rs;
			inslst.ny[NEWINS] = ( inslst.py[NEWINS] - sphlst.oy[i] ) / rs;
			inslst.nz[NEWINS] = ( inslst.pz[NEWINS] - sphlst.oz[i] ) / rs;

			t = tmin;
		}
	}
}

void triins( REAL ox, REAL oy, REAL oz, REAL dx, REAL dy, REAL dz, int i, int& sect, REAL& t, int iprim )
//-------------------------------------------------------------------------------------------------------
// See if the ray from (ox,oy,oz) direction (dx,dy,dz) intersects triangle i. The primitive index is iprim.
// If so, set sect non-zero and set t to the intersection parameter.
// Otherwise set sect to zero.
// If there is an intersection save the position and normal in inslst.XXX[NEWINS].
{
	REAL ndotd, ndoto, px, py, pz, u0, u1, u2, v0, v1, v2;
	REAL alpha, beta, gamma, npx, npy, npz;

	// Check ray is not parallel to plane embedding triangle.
	// PRECISION!
	ndotd = trilst.nx[i] * dx + trilst.ny[i] * dy + trilst.nz[i] * dz;
	if( fabs( ndotd ) > EPS ){

		// Find intersection parameter with plane.
		ndoto = trilst.nx[i] * ox + trilst.ny[i] * oy + trilst.nz[i] * oz;
		t = -( ndoto + trilst.nd[i] ) / ndotd;

		// Find intersection point with embedding plane.
		px = ox + t * dx;
		py = oy + t * dy;
		pz = oz + t * dz;

		// Find constants for locating intersection point in terms of parametric
		// coordinates in a coordinate system using two edges of the triangle as
		// its basis vectors. Use the "biggest normal component" precomputed as dax
		// to select the plane normal to the least degenerate possible parametrization.
		if( trilst.dax[i] == DAXX ){
			u0 = py - trilst.y1[i];
			u1 = trilst.y2[i] - trilst.y1[i];
			u2 = trilst.y3[i] - trilst.y1[i];
			v0 = pz - trilst.z1[i];
			v1 = trilst.z2[i] - trilst.z1[i];
			v2 = trilst.z3[i] - trilst.z1[i];
		}
		else if( trilst.dax[i] == DAXY ){
			u0 = px - trilst.x1[i];
			u1 = trilst.x2[i] - trilst.x1[i];
			u2 = trilst.x3[i] - trilst.x1[i];
			v0 = pz - trilst.z1[i];
			v1 = trilst.z2[i] - trilst.z1[i];
			v2 = trilst.z3[i] - trilst.z1[i];
		}
		else if( trilst.dax[i] == DAXZ ){
			u0 = px - trilst.x1[i];
			u1 = trilst.x2[i] - trilst.x1[i];
			u2 = trilst.x3[i] - trilst.x1[i];
			v0 = py - trilst.y1[i];
			v1 = trilst.y2[i] - trilst.y1[i];
			v2 = trilst.y3[i] - trilst.y1[i];
		}
		else{
			fprintf( stderr, "INTERNAL ERROR: Invalid DAX code in triins().\n" );
			exit( 1 );
		}

		// Find the (alpha,beta) coordinates of the intersection point.
		// PRECISION!
		if( fabs( u1 ) < MINEPS ){
			beta = u0 / u2;
			if( beta >= 0.0 && beta <= 1.0 )
				alpha = ( v0 - beta * v2 ) / v1;
			else{
				t = 0.0;
				sect = 0;
				return;
			}
		}
		else{
			beta = ( v0 * u1 - u0 * v1 ) / ( v2 * u1 - u2 * v1 );
			if( beta >= 0.0 && beta <= 1.0 )
				alpha = ( u0 - beta * u2 ) / u1;
			else{
				t = 0.0;
				sect = 0;
				return;
			}
		}

		// Find the gamma coordinate.
		gamma = alpha + beta;
		if( alpha < 0.0 || gamma > 1.0 ){
				t = 0.0;
				sect = 0;
				return;
		}

		// We are inside the triangle.
		// Use (alpha,beta,gamma) to interpolate the vertex properties (currently the normal).
		npx = gamma * trilst.nvx1[i] + alpha * trilst.nvx2[i] + beta * trilst.nvx3[i];
		npy = gamma * trilst.nvy1[i] + alpha * trilst.nvy2[i] + beta * trilst.nvy3[i];
		npz = gamma * trilst.nvz1[i] + alpha * trilst.nvz2[i] + beta * trilst.nvz3[i];
		nrmvec( npx, npy, npz );

		// Set the intersection data
		inslst.prim[NEWINS] = iprim;
		inslst.rray[NEWINS] = 0;
		inslst.tray[NEWINS] = 0;
		inslst.colr[NEWINS] = 0.0;
		inslst.colg[NEWINS] = 0.0;
		inslst.colb[NEWINS] = 0.0;
		inslst.px[NEWINS] = px;
		inslst.py[NEWINS] = py;
		inslst.pz[NEWINS] = pz;
		inslst.nx[NEWINS] = npx;
		inslst.ny[NEWINS] = npy;
		inslst.nz[NEWINS] = npz;

		// Guess whether we are entering or leaving.
		ndotd = npx * dx + npy * dy + npz * dz;
		if( ndotd < 0.0 )
			inslst.ttyp[NEWINS] = ENTER;
		else
			inslst.ttyp[NEWINS] = LEAVE;

		sect = 1;
		return;
	}

	// Ray is parallel to embedding plane. No intersection possible.
	else{
		t = 0.0;
		sect = 0;
		return;
	}
}

void cpyins( int i, int j )
//-------------------------
// Move a newly found intersection from inslst.XXX[j] to inslst.XXX[i]
{
	inslst.prim[i] = inslst.prim[j];
	inslst.rray[i] = inslst.rray[j];
	inslst.tray[i] = inslst.tray[j];
	inslst.colr[i] = inslst.colr[j];
	inslst.colg[i] = inslst.colg[j];
	inslst.colb[i] = inslst.colb[j];
	inslst.px[i] = inslst.px[j];
	inslst.py[i] = inslst.py[j];
	inslst.pz[i] = inslst.pz[j];
	inslst.nx[i] = inslst.nx[j];
	inslst.ny[i] = inslst.ny[j];
	inslst.nz[i] = inslst.nz[j];
	inslst.ttyp[i] = inslst.ttyp[j];
}

//**********//
// DATABASE //
//**********//

void initdb( void )
//-----------------
// Initialize the scene database
{
	prmlst.nprm = 0;
	sphlst.nsph = 0;
	mtrlst.nmtr = 0;
	lgtlst.nlgt = 0;
	raylst.raynum = 0;
}

void initbk( void )
//-----------------
// Initialise things for graded backgrounds
{
	REAL fn;

	if( screen.bkt == BKVERT )
		fn = screen.ny - 1;
	else if( screen.bkt == BKHORZ )
		fn = screen.nx - 1;
	else
		fn = 1.0;
	screen.bkrd = ( screen.bkrf - screen.bkri ) / fn;
	screen.bkgd = ( screen.bkgf - screen.bkgi ) / fn;
	screen.bkbd = ( screen.bkbf - screen.bkbi ) / fn;
}

void initlg( void )
//-----------------
// Normalize and invert the light pointing vectors for directional lights.
{
	int l;

	for( l=1; l<=lgtlst.nlgt; l++ ){
		if( lgtlst.dir[l] != LGTCON ){
			nrmvec( lgtlst.dx[l], lgtlst.dy[l], lgtlst.dz[l] );
			lgtlst.dx[l] = -lgtlst.dx[l];
			lgtlst.dy[l] = -lgtlst.dy[l];
			lgtlst.dz[l] = -lgtlst.dz[l];
		}
	}
}

void initri( int i )
//------------------
// Precompute plane equation and dominant axis info for triangle [i]
{
	REAL dxa, dya, dza, dxb, dyb, dzb;
	REAL fnx, fny, fnz, len, nvx, nvy, nvz;

	dxa = trilst.x2[i] - trilst.x1[i];
	dya = trilst.y2[i] - trilst.y1[i];
	dza = trilst.z2[i] - trilst.z1[i];
	dxb = trilst.x3[i] - trilst.x1[i];
	dyb = trilst.y3[i] - trilst.y1[i];
	dzb = trilst.z3[i] - trilst.z1[i];
	trilst.nx[i] = dya * dzb - dza * dyb;
	trilst.ny[i] = dza * dxb - dxa * dzb;
	trilst.nz[i] = dxa * dyb - dya * dxb;
	trilst.nd[i] = - ( trilst.x1[i] * trilst.nx[i] + trilst.y1[i] * trilst.ny[i] + trilst.z1[i] * trilst.nz[i] );
	if( trilst.nx[i] > trilst.ny[i] && trilst.nx[i] > trilst.nz[i] )
		trilst.dax[i] = DAXX;
	else if( trilst.ny[i] > trilst.nx[i] && trilst.ny[i] > trilst.nz[i] )
		trilst.dax[i] = DAXY;
	else
		trilst.dax[i] = DAXZ;

	// If the vertex normal length for a vertex is tiny, substitute a unit facet normal.
	// Otherwise, ensure the vertex normals are unit vectors.
	fnx = trilst.nx[i];
	fny = trilst.ny[i];
	fnz = trilst.nz[i];
	nrmvec( fnx, fny, fnz );

	nvx = trilst.nvx1[i];
	nvy = trilst.nvy1[i];
	nvz = trilst.nvz1[i];
	len = nvx * nvx + nvy * nvy + nvz * nvz;
	// PRECISION!
	if( len < (MINEPS*MINEPS) ){
		trilst.nvx1[i] = fnx;
		trilst.nvy1[i] = fny;
		trilst.nvz1[i] = fnz;
	}
	else{
		len = sqrt( len );
		trilst.nvx1[i] /= len;
		trilst.nvy1[i] /= len;
		trilst.nvz1[i] /= len;
	}

	nvx = trilst.nvx2[i];
	nvy = trilst.nvy2[i];
	nvz = trilst.nvz2[i];
	len = nvx * nvx + nvy * nvy + nvz * nvz;
	// PRECISION!
	if( len < (MINEPS*MINEPS) ){
		trilst.nvx2[i] = fnx;
		trilst.nvy2[i] = fny;
		trilst.nvz2[i] = fnz;
	}
	else{
		len = sqrt( len );
		trilst.nvx2[i] /= len;
		trilst.nvy2[i] /= len;
		trilst.nvz2[i] /= len;
	}

	nvx = trilst.nvx3[i];
	nvy = trilst.nvy3[i];
	nvz = trilst.nvz3[i];
	len = nvx * nvx + nvy * nvy + nvz * nvz;
	// RECISION!
	if( len < (MINEPS*MINEPS) ){
		trilst.nvx3[i] = fnx;
		trilst.nvy3[i] = fny;
		trilst.nvz3[i] = fnz;
	}
	else{
		len = sqrt( len );
		trilst.nvx3[i] /= len;
		trilst.nvy3[i] /= len;
		trilst.nvz3[i] /= len;
	}
}

int readdb( char* scene )
//-----------------------
// Read the scene database.
// Return 0 if OK
// Format of database is (ugh!):
// accel shadow (If 1 use acceleration, otherwise don't. Shadow > 0 do shadows with [shadow] feelers.)
// nx ny os (number of pixels in raster in X and Y, oversample)
// aper fpd focal fstop (aperture, focal plane distance, focal length, f-stop)
// maxdep minwgt (max tree depth, min ray weight)
// bktype ibackr ibackg ibackb fbackr fbackg fbackb (background type, initial colour, final colour)
// <number of materials>
// car cag cab
// cdr cdg cdb
// csr csg csb
// kr kt eta gloss roughness (* num mtls)
// car cag cab (ambient light colour)
// <number of lights>
// ox oy oz clr clg clb radius
// dtype dx dy dz ang1 ang2  (* num lights)
// <number of spheres>
// mtridx ox oy oz rad (* num spheres)
// <number of triangles>
// mtridx
// x1 y1 z1 nx1 ny1 nz1
// x2 y2 z2 nx2 ny2 nz2
// x3 y3 z3 nx3 ny3 nz3
{
	FILE* fin;
	int i;

	fin = fopen( scene, "r" );
	if( fin ){
		// Screen/camera specification.
		if( fscanf( fin, "%d %d", &screen.accel, &screen.shadow ) != 2 ){
			fprintf( stderr, "Error: Failed to read acceleration scheme number or shadow on/off\n" );
			goto error;
		}
		if( fscanf( fin, "%d %d %d", &screen.nx, &screen.ny, &screen.os ) != 3 ){
			fprintf( stderr, "Error: Failed to read camera/screen definition (part 1)\n" );
			goto error;
		}
		if( fscanf( fin, "%lf %lf %lf %lf", &screen.aper, &screen.fpd, &screen.focal, &screen.fstop ) != 4 ){
			fprintf( stderr, "Error: Failed to read camera/screen definition (part 2)\n" );
			goto error;
		}
		if( fscanf( fin, "%d %lf", &screen.maxdep, &screen.minwgt ) != 2 ){
			fprintf( stderr, "Error: Failed to read camera/screen definition (part 3)\n" );
			goto error;
		}
		if( screen.nx > MAXPXL ){
			fprintf( stderr, "Error: Horizontal resolution %d exceeds maximum %d\n", screen.nx, MAXPXL );
			goto error;
		}
		if( screen.os > 0 ){
			if( ( screen.os & 1 ) == 0 )
				screen.os++;  // Force oversample rate to be odd.
		}
		if( fscanf( fin, "%d %lf %lf %lf %lf %lf %lf", &screen.bkt, &screen.bkri, &screen.bkgi, &screen.bkbi, &screen.bkrf, &screen.bkgf, &screen.bkbf ) != 7 ){
			fprintf( stderr, "Error: Failed to read screen background colour\n" );
			goto error;
		}
		if( screen.bkt < BKCNST || screen.bkt > BKHORZ ){
			fprintf( stderr, "Error: Invalid background grade type\n" );
			goto error;
		}
		initbk();

		// Materials specification.
		if( fscanf( fin, "%d", &mtrlst.nmtr ) != 1 ){
			fprintf( stderr, "Error: Failed to read number of materials\n" );
			goto error;
		}
		if( mtrlst.nmtr > MAXMTR ){
			fprintf( stderr, "Error: Too many materials specified: %d, limit: %d\n", mtrlst.nmtr, MAXMTR );
			goto error;
		}
		for( i=1; i<=mtrlst.nmtr; i++ ){
			if( fscanf( fin, "%lf %lf %lf", &mtrlst.car[i], &mtrlst.cag[i], &mtrlst.cab[i] ) != 3 ){
				fprintf( stderr, "Error: Failed to read material number: %d (part 1: ambient)\n", i );
				goto error;
			}
			if( fscanf( fin, "%lf %lf %lf", &mtrlst.cdr[i], &mtrlst.cdg[i], &mtrlst.cdb[i] ) != 3 ){
				fprintf( stderr, "Error: Failed to read material number: %d (part 2: diffuse)\n", i );
				goto error;
			}
			if( fscanf( fin, "%lf %lf %lf", &mtrlst.csr[i], &mtrlst.csg[i], &mtrlst.csb[i] ) != 3 ){
				fprintf( stderr, "Error: Failed to read material number: %d (part 3: specular)\n", i );
				goto error;
			}
			if( fscanf( fin, "%lf %lf %lf %lf %lf", &mtrlst.kr[i], &mtrlst.kt[i], &mtrlst.eta[i], &mtrlst.gls[i], &mtrlst.rgh[i] ) != 5 ){
				fprintf( stderr, "Error: Failed to read material number: %d (part 4: other)\n", i );
				goto error;
			}
		}

		// Lights specification.
		if( fscanf( fin, "%lf %lf %lf", &lgtlst.car, &lgtlst.cag, &lgtlst.cab ) != 3 ){
			fprintf( stderr, "Error: Failed to read ambient light colour\n", i );
			goto error;
		}
		if( fscanf( fin, "%d", &lgtlst.nlgt ) != 1 ){
			fprintf( stderr, "Error: Failed to read number of lights\n" );
			goto error;
		}
		if( lgtlst.nlgt > MAXLGT ){
			fprintf( stderr, "Error: Too many lights specified: %d, limit: %d\n", lgtlst.nlgt, MAXLGT );
			goto error;
		}
		for( i=1; i<=lgtlst.nlgt; i++ ){
			if( fscanf( fin, "%lf %lf %lf %lf %lf %lf %lf", &lgtlst.ox[i], &lgtlst.oy[i], &lgtlst.oz[i], &lgtlst.clr[i], &lgtlst.clg[i], &lgtlst.clb[i], &lgtlst.rad[i] ) != 7 ){
				fprintf( stderr, "Error: Failed to read light number: %d (part 1: pos & col & rad)\n", i );
				goto error;
			}
			if( fscanf( fin, "%d %lf %lf %lf %lf %lf", &lgtlst.dir[i], &lgtlst.dx[i], &lgtlst.dy[i], &lgtlst.dz[i], &lgtlst.tgt[i], &lgtlst.tg2[i] ) != 6 ){
				fprintf( stderr, "Error: Failed to read light number: %d (part 2: directionality)\n", i );
				goto error;
			}
			if( lgtlst.tg2[i] <= lgtlst.tgt[i] )
				lgtlst.tg2[i] = lgtlst.tgt[i] + 0.0001;
			lgtlst.tgt[i] *= DEGRAD;
			lgtlst.tg2[i] *= DEGRAD;
		}
		initlg();

		// No geometric primitives yet
		prmlst.nprm = 0;

		// Spheres specification.
		if( fscanf( fin, "%d", &sphlst.nsph ) != 1 ){
			fprintf( stderr, "Error: Failed to read number of spheres\n" );
			goto error;
		}
		if( sphlst.nsph > MAXSPH ){
			fprintf( stderr, "Error: Too many spheres specified: %d, limit: %d\n", sphlst.nsph, MAXSPH );
			goto error;
		}
		for( i=1; i<=sphlst.nsph; i++ ){
			++prmlst.nprm;
			if( prmlst.nprm > MAXPRM ){
				fprintf( stderr, "Too many primitives required: %d, limit: %d, reading sphere: %d\n", prmlst.nprm, MAXPRM, i );
				goto error;
			}
			if( fscanf( fin, "%d %lf %lf %lf %lf", &prmlst.matter[prmlst.nprm], &sphlst.ox[i], &sphlst.oy[i], &sphlst.oz[i], &sphlst.rad[i] ) != 5 ){
				fprintf( stderr, "Error: Failed to read sphere number: %d\n", i );
				goto error;
			}
			if( prmlst.matter[prmlst.nprm] > mtrlst.nmtr || prmlst.matter[prmlst.nprm] < 1 ){
				fprintf( stderr, "Error: Material index for primitive number: %d (sphere: %d) is out of range: %d\n", prmlst.nprm, i, prmlst.matter[prmlst.nprm] );
				goto error;
			}
			prmlst.accel[prmlst.nprm] = 0;        // Not in acceleration structure.
			prmlst.rayid[prmlst.nprm] = 0;        // No ray has intersected it yet.
			prmlst.prmtyp[prmlst.nprm] = SPHPRM;  // Primitive type: Sphere.
			prmlst.prmidx[prmlst.nprm] = i;       // Index of definition in sphere table.
		}

		// Triangles specification.
		if( fscanf( fin, "%d", &trilst.ntri ) != 1 ){
			fprintf( stderr, "Error: Failed to read number of triangles\n" );
			goto error;
		}
		if( trilst.ntri > MAXTRI ){
			fprintf( stderr, "Error: Too many triangles specified: %d, limit: %d\n", trilst.ntri, MAXTRI );
			goto error;
		}
		for( i=1; i<=trilst.ntri; i++ ){
			++prmlst.nprm;
			if( prmlst.nprm > MAXPRM ){
				fprintf( stderr, "Too many primitives required: %d, limit: %d, reading triangle: %d\n", prmlst.nprm, MAXPRM, i );
				goto error;
			}
			if( fscanf( fin, "%d", &prmlst.matter[prmlst.nprm] ) != 1 ){
				fprintf( stderr, "Error: Failed to read triangle number: %d (part 1: material)\n", i );
				goto error;
			}
			if( prmlst.matter[prmlst.nprm] > mtrlst.nmtr || prmlst.matter[prmlst.nprm] < 1 ){
				fprintf( stderr, "Error: Material index for primitive number: %d (sphere: %d) is out of range: %d\n", prmlst.nprm, i, prmlst.matter[prmlst.nprm] );
				goto error;
			}
			if( fscanf( fin, "%lf %lf %lf %lf %lf %lf", &trilst.x1[i], &trilst.y1[i], &trilst.z1[i], &trilst.nvx1[i], &trilst.nvy1[i], &trilst.nvz1[i] ) != 6 ){
				fprintf( stderr, "Error: Failed to read triangle number: %d (part 2: vertex 1)\n", i );
				goto error;
			}
			if( fscanf( fin, "%lf %lf %lf %lf %lf %lf", &trilst.x2[i], &trilst.y2[i], &trilst.z2[i], &trilst.nvx2[i], &trilst.nvy2[i], &trilst.nvz2[i] ) != 6 ){
				fprintf( stderr, "Error: Failed to read triangle number: %d (part 3: vertex 2)\n", i );
				goto error;
			}
			if( fscanf( fin, "%lf %lf %lf %lf %lf %lf", &trilst.x3[i], &trilst.y3[i], &trilst.z3[i], &trilst.nvx3[i], &trilst.nvy3[i], &trilst.nvz3[i] ) != 6 ){
				fprintf( stderr, "Error: Failed to read triangle number: %d (part 4: vertex 3)\n", i );
				goto error;
			}
			initri( i );
			prmlst.accel[prmlst.nprm] = 0;        // Not in acceleration structure.
			prmlst.rayid[prmlst.nprm] = 0;        // No ray has intersected it yet.
			prmlst.prmtyp[prmlst.nprm] = TRIPRM;  // Primitive type: Triangle.
			prmlst.prmidx[prmlst.nprm] = i;       // Index of definition in triangle table.
		}

		// Finished successfully.
		return( 0 );
	}
	else
		return( 1 );
error:
	if( fin )
		fclose( fin );
	return( 1 );
}

//*************************************//
// GRID SPACE SUBDIVISION ACCELERATION //
//*************************************//

void stacel( void )
//-----------------
// Set up the acceleration structure. This is a simple rectangular space
// subdivision scheme. (But are these things ever really simple?)
{
	int i, j, k, p, pidx;
	REAL pxl, pxh, pyl, pyh, pzl, pzh;
	REAL dxc, dyc, dzc;
	REAL cxl, cxh, cyl, cyh, czl, czh;
	int ixl, ixh, iyl, iyh, izl, izh;
	REAL maxdir, mindir, xc, yc, zc;

	accel.anx = MXCELX;
	accel.any = MXCELY;
	accel.anz = MXCELZ;

	// Set all "pointers" to primitive lists in the cells to "NULL".
	// Primitive lists are empty;
	for( k=0; k<accel.anz; k++ ){
		for( j=0; j<accel.any; j++ ){
			for( i=0; i<accel.anx; i++ ){
				accel.cells[i+j*accel.anx+k*(accel.anx*accel.any)+1] = 0;
			}
		}
	}
	accel.nlist = 0;

	// Survey the scene to get its bounds.
	for( p=1; p<=prmlst.nprm; p++ ){
		pidx = prmlst.prmidx[p];
		if( prmlst.prmtyp[p] == SPHPRM )
			sphbnd( pidx, pxl, pxh, pyl, pyh, pzl, pzh );
		else if( prmlst.prmtyp[p] == TRIPRM )
			tribnd( pidx, pxl, pxh, pyl, pyh, pzl, pzh );
		else{
			fprintf( stderr, "INTERNAL ERROR: Unknown primitive type in stacel().\n" );
			exit( 1 );
		}
		if( p == 1 ){
			accel.axl = pxl;
			accel.axh = pxh;
			accel.ayl = pyl;
			accel.ayh = pyh;
			accel.azl = pzl;
			accel.azh = pzh;
		}
		else{
			if( pxl < accel.axl )
				accel.axl = pxl;
			if( pxh > accel.axh )
				accel.axh = pxh;
			if( pyl < accel.ayl )
				accel.ayl = pyl;
			if( pyh > accel.ayh )
				accel.ayh = pyh;
			if( pzl < accel.azl )
				accel.azl = pzl;
			if( pzh > accel.azh )
				accel.azh = pzh;
		}
	}
	accel.dxs = accel.axh - accel.axl;
	accel.dys = accel.ayh - accel.ayl;
	accel.dzs = accel.azh - accel.azl;

	printf( "\nAcceleration structure information:\n" );
	printf( "Grid dimensions         = %d x %d x %d\n", accel.anx, accel.any, accel.anz );
	printf( "Indexable size (cells)  = %d x %d x %d\n", MXPCVR, MXPCVR, MXPCVR );
	printf( "Adjust if ratio exceeds = %f\n", MXDMRT );
	printf( "Scene X range           = (%f:%f)\n", accel.axl, accel.axh );
	printf( "Scene Y range           = (%f:%f)\n", accel.ayl, accel.ayh );
	printf( "Scene Z range           = (%f:%f)\n", accel.azl, accel.azh );

	// Optionally, choose the biggest axis to scale the whole grid
	// If there is a big disparity in the dimensions of the axes, this can help
	// ensure that cells are made big enough to still index primitives. If the cells
	// are made very small in one direction, it may be that too many are covered
	// otherwise, and the acceleration scheme can't be used.
	// It seems like this is almost always a good idea anyway.
	maxdir = accel.dxs;
	if( accel.dys > maxdir )
		maxdir = accel.dys;
	if( accel.dzs > maxdir )
		maxdir = accel.dzs;

	mindir = accel.dxs;
	if( accel.dys < mindir )
		mindir = accel.dys;
	if( accel.dzs < mindir )
		mindir = accel.dzs;

	if( ( maxdir / mindir ) > MXDMRT ){ // Tune this for when to do it...
		xc = 0.5 * ( accel.axl + accel.axh );
		yc = 0.5 * ( accel.ayl + accel.ayh );
		zc = 0.5 * ( accel.azl + accel.azh );
		accel.axl = xc - 0.5 * maxdir;
		accel.axh = xc + 0.5 * maxdir;
		accel.ayl = yc - 0.5 * maxdir;
		accel.ayh = yc + 0.5 * maxdir;
		accel.azl = zc - 0.5 * maxdir;
		accel.azh = zc + 0.5 * maxdir;
		accel.dxs = maxdir;
		accel.dys = maxdir;
		accel.dzs = maxdir;
		printf( "Scene ranges were adjusted for dimensional uniformity.\n" );
		printf( "Scene X range (adj.)    = (%f:%f)\n", accel.axl, accel.axh );
		printf( "Scene Y range (adj.)    = (%f:%f)\n", accel.ayl, accel.ayh );
		printf( "Scene Z range (adj.)    = (%f:%f)\n", accel.azl, accel.azh );
	}

	// Calculate the size of each cell in eye space
	dxc = accel.dxs / accel.anx;
	dyc = accel.dys / accel.any;
	dzc = accel.dzs / accel.anz;

	printf( "Cell sizes              = %f x %f x %f\n", dxc, dyc, dzc );

	// Go over the primitives again, assigning them to cells.
	// If the primitive bounds cover more than MXPCVR cells in any dimension, we
	// don't use acceleration for that primitive. This is to limit the size of
	// the cell lists.

	for( p=1; p<=prmlst.nprm; p++ ){
		pidx = prmlst.prmidx[p];
		if( prmlst.prmtyp[p] == SPHPRM )
			sphbnd( pidx, pxl, pxh, pyl, pyh, pzl, pzh );
		else if( prmlst.prmtyp[p] == TRIPRM )
			tribnd( pidx, pxl, pxh, pyl, pyh, pzl, pzh );
		else{
			fprintf( stderr, "INTERNAL ERROR: Unknown primitive type in stacel().\n" );
			exit( 1 );
		}
		cxl = ( pxl - accel.axl ) / dxc;
		cxh = ( pxh - accel.axl ) / dxc;
		cyl = ( pyl - accel.ayl ) / dyc;
		cyh = ( pyh - accel.ayl ) / dyc;
		czl = ( pzl - accel.azl ) / dzc;
		czh = ( pzh - accel.azl ) / dzc;

		// Move to integer grid cell coordinates.
		ixl = (int)cxl;
		ixh = (int)cxh;
		iyl = (int)cyl;
		iyh = (int)cyh;
		izl = (int)czl;
		izh = (int)czh;

		// Add primitive to the list of primitives for each cell covered by the bounds of the primitive.
		// If the primitive is too big, don't use the acceleration structure at all.
		// This is to limit the size of the primitive lists in nasty cases.
		if( ixl == ixh && iyl == iyh && izl == izh ){
			prmlst.accel[p] = 1;
			++stats.nacc;
			addlst( ixl, iyl, izl, p );
		}
		else{
			if( ( ixh - ixl ) >= MXPCVR || ( iyh - iyl ) >= MXPCVR || ( izh - izl ) >= MXPCVR ){
				prmlst.accel[p] = 0;
				++stats.noacc;
			}
			else{
				prmlst.accel[p] = 1;
				++stats.nacc;
				for( k=izl; k<=izh; k++ ){
					for( j=iyl; j<=iyh; j++ ){
						for( i=ixl; i<=ixh; i++ ){
							addlst( i, j, k, p );
						}
					}
				}
			}
		}
	}

	printf( "Primitives indexed      = %d\n", stats.nacc );
	printf( "Primitives not indexed  = %d\n", stats.noacc );
}

void addlst( int i, int j, int k, int p )
//---------------------------------------
// Add primitive p to the list associated with cell (i,j,k) if it isn't already on the list.
// *** N.B.: (i,j,k) are ZERO based!
{
	int ci, lp, lc;

  ci = i + j * accel.anx + k * (accel.anx*accel.any) + 1;
	if( accel.cells[ci] == 0 ){
		++accel.nlist;
		if( accel.nlist > MAXLST ){
			printf( "Trying to add too many list elements.\n" );
		}
		else{
			accel.cells[ci] = accel.nlist;
			accel.lprim[accel.nlist] = p;
			accel.lnext[accel.nlist] = 0;
		}
	}
	else{
		lp = 0;
		lc = accel.cells[ci];
		while( lc != 0 ){
			if( accel.lprim[lc] == p ){
				return;
			}
			else{
				lp = lc;
				lc = accel.lnext[lc];
			}
		}
		++accel.nlist;
		if( accel.nlist > MAXLST ){
			printf( "Trying to add too many list elements.\n" );
		}
		else{
			accel.lnext[lp] = accel.nlist;
			accel.lprim[accel.nlist] = p;
			accel.lnext[accel.nlist] = 0;
		}
	}
}

void sphbnd( int i, REAL& pxl, REAL& pxh, REAL& pyl, REAL& pyh, REAL& pzl, REAL& pzh )
//------------------------------------------------------------------------------------
// Find a bounding box for a sphere primitive. Very easy!
{
	pxl = sphlst.ox[i] - sphlst.rad[i];
	pxh = sphlst.ox[i] + sphlst.rad[i];
	pyl = sphlst.oy[i] - sphlst.rad[i];
	pyh = sphlst.oy[i] + sphlst.rad[i];
	pzl = sphlst.oz[i] - sphlst.rad[i];
	pzh = sphlst.oz[i] + sphlst.rad[i];
}

void tribnd( int i, REAL& pxl, REAL& pxh, REAL& pyl, REAL& pyh, REAL& pzl, REAL& pzh )
//------------------------------------------------------------------------------------
// Find a bounding box for a triangle primitive. Very easy!
{
	pxl = trilst.x1[i];
	if( trilst.x2[i] < pxl )
		pxl = trilst.x2[i];
	if( trilst.x3[i] < pxl )
		pxl = trilst.x3[i];

	pyl = trilst.y1[i];
	if( trilst.y2[i] < pyl )
		pyl = trilst.y2[i];
	if( trilst.y3[i] < pyl )
		pyl = trilst.y3[i];
		
	pzl = trilst.z1[i];
	if( trilst.z2[i] < pzl )
		pzl = trilst.z2[i];
	if( trilst.z3[i] < pzl )
		pzl = trilst.z3[i];

	pxh = trilst.x1[i];
	if( trilst.x2[i] > pxh )
		pxh = trilst.x2[i];
	if( trilst.x3[i] > pxh )
		pxh = trilst.x3[i];

	pyh = trilst.y1[i];
	if( trilst.y2[i] > pyh )
		pyh = trilst.y2[i];
	if( trilst.y3[i] > pyh )
		pyh = trilst.y3[i];
		
	pzh = trilst.z1[i];
	if( trilst.z2[i] > pzh )
		pzh = trilst.z2[i];
	if( trilst.z3[i] > pzh )
		pzh = trilst.z3[i];
}

void grdtrv( REAL ux, REAL uy, REAL uz, REAL vx, REAL vy, REAL vz, int nx, int ny, int nz )
//-----------------------------------------------------------------------------------------
// Visit all cells in a 3D integer grid of nx x ny x nz cells that are pierced by a ray with origin
// (ux,uy,uz) - **positive and within the grid** - and direction vector (vx,vy,vz) - not necessarily normalized.
// *** N.B.: Cell coordinates passed to viscel() are ZERO based. ***
// This is Amantides & Woo's algorithm.
{
	int x, y, z, sx, sy, sz, jx, jy, jz, quit;
	REAL tx, ty, tz, dx, dy, dz;

	// Initial cell coordinates
	x = (int)ux;
	y = (int)uy;
	z = (int)uz;

	// Visit the initial cell
	viscel( x, y, z, quit );
	if( quit != 0 )
		return;

	// Integer cell steps for the vector
	sx = ( vx < 0 ) ? -1 : 1 ;
	sy = ( vy < 0 ) ? -1 : 1 ;
	sz = ( vz < 0 ) ? -1 : 1 ;

	// Find tx,ty,tz - the t values that move along the ray to the boundary of the initial cell.
	// Find dx,dy,dz - the t values that step the width, height and depth of a full cell.
	// PRECISION!
	if( fabs( vx ) < EPS ){
		tx = GTHUGE;
		dx = GTHUGE;
	}
	else if( vx > 0 ){
		tx = ( ( x + 1 ) - ux ) / vx;
		dx = 1.0f / vx;
	}
	else{
		tx = ( x  - ux ) / vx;
		dx = -1.0f / vx;
	}
	if( fabs( vy ) < EPS ){
		ty = GTHUGE;
		dy = GTHUGE;
	}
	else if( vy > 0 ){
		ty = ( ( y + 1 ) - uy ) / vy;
		dy = 1.0f / vy;
	}
	else{
		ty = ( y  - uy ) / vy;
		dy = -1.0f / vy;
	}
	if( fabs( vz ) < EPS ){
		tz = GTHUGE;
		dz = GTHUGE;
	}
	else if( vz > 0 ){
		tz = ( ( z + 1 ) - uz ) / vz;
		dz = 1.0f / vz;
	}
	else{
		tz = ( z  - uz ) / vz;
		dz = -1.0f / vz;
	}

	// If any of the t values to move to the boundary are exactly zero, bump them a bit.
	// PRECISION!
	if( tx == 0 )tx += EPS;
	if( ty == 0 )ty += EPS;
	if( tz == 0 )tz += EPS;

	// Find just out of bounds cell coordinates for the ray along each axis
	if( sx > 0 )
		jx = nx;
	else
		jx = -1;
	if( sy > 0 )
		jy = ny;
	else
		jy = -1;
	if( sz > 0 )
		jz = nz;
	else
		jz = -1;

	// Step through the cells until we exit the array
	for( ;; ){
		if( tx < ty ){
			if( tx < tz ){
				x += sx;
				if( x == jx )
					return;
				tx += dx;
			}
			else{
				z += sz;
				if( z == jz )
					return;
				tz += dz;
			}
		}
		else{
			if( ty < tz ){
				y += sy;
				if( y == jy )
					return;
				ty += dy;
			}
			else{
				z += sz;
				if( z == jz )
					return;
				tz += dz;
			}
		}
		viscel( x, y, z, quit );
		if( quit != 0 )
			return;
	}
}

void boxins( REAL ox, REAL oy, REAL oz, REAL dx, REAL dy, REAL dz, int& sect, REAL& xi, REAL& yi, REAL& zi )
//----------------------------------------------------------------------------------------------------------
// See if the ray with origin (ox,oy,oz) direction (dx, dy, dz) (not necessarily normalised) intersects the bounding
// box for the scene geometric objects found by stacel(). If so, return sect set to 1 and the intersection point in
// (xi, yi, zi). Otherwise, return sect 0.
{
	REAL t1, t2, tnear, tfar, t;

	xi = yi = zi = 0.0f;
	sect = 0;
	tnear = MINFLT;
	tfar = MAXFLT;

	// PRECISION!
	if( fabs( dx ) < EPS ){
		if( ox < accel.axl || ox > accel.axh ){
			sect = 0;
			return;
		}
	}
	else{
		tnear = ( accel.axl - ox ) / dx;
		tfar = ( accel.axh - ox ) / dx;
		if( tnear > tfar ){ t = tnear; tnear = tfar; tfar = t; }
		if( tfar < 0.0 ){
			sect = 0;
			return;
		}
	}

	// PRECISION!
	if( fabs( dy ) < EPS ){
		if( oy < accel.ayl || oy > accel.ayh ){
			sect = 0;
			return;
		}
	}
	else{
		t1 = ( accel.ayl - oy ) / dy;
		t2 = ( accel.ayh - oy ) / dy;
		if( t1 > t2 ){ t = t1; t1 = t2; t2 = t; }
		if( t1 > tnear )tnear = t1;
		if( t2 < tfar )tfar = t2;
		if( tnear > tfar ){
			sect = 0;
			return;
		}
		if( tfar < 0.0 ){
			sect = 0;
			return;
		}
	}

	// PRECISION!
	if( fabs( dz ) < EPS ){
		if( oz < accel.azl || oz > accel.azh ){
			sect = 0;
			return;
		}
	}
	else{
		t1 = ( accel.azl - oz ) / dz;
		t2 = ( accel.azh - oz ) / dz;
		if( t1 > t2 ){ t = t1; t1 = t2; t2 = t; }
		if( t1 > tnear )tnear = t1;
		if( t2 < tfar )tfar = t2;
		if( tnear > tfar ){
			sect = 0;
			return;
		}
		if( tfar < 0.0 ){
			sect = 0;
			return;
		}
	}

	// There is an intersection. Find where.
	xi = ox + tnear * dx;
	yi = oy + tnear * dy;
	zi = oz + tnear * dz;
	sect = 1;
}

//******************//
// UTILITY ROUTINES //
//******************//

REAL ranhlf( void )
//------------------
// Return a random number in the range +/- 0.5
{
	return( ( (REAL)rand() / RAND_MAX ) - 0.5f );
}

void nrmvec( REAL& dx, REAL& dy, REAL& dz )
//--------------------------------------------
// Make (dx,dy,dz) a unit vector
{
	REAL len;

	len = sqrt( dx*dx + dy*dy + dz*dz );
	// PRECISION!
	if( len > EPS ){
		dx /= len;
		dy /= len;
		dz /= len;
	}
}

//**********************//
// TIFF FORMAT ROUTINES //
//**********************//

int image_sizeof( int w, int h, int c, int t )
//--------------------------------------------
// Calculate number of bytes in an image of supplied Width, Height, Channels and data Type.
// Return -1 if type of image data not recognised.
{
	int t_size = -1;
	
	t_size = w * h * c;
	
	switch( t ){
	case IMG_FMT_BYTE:  break;
	case IMG_FMT_WORD:  t_size <<= 1; break;
	case IMG_FMT_WORD32: t_size <<= 2; break;
	case IMG_FMT_FLOAT: t_size *= sizeof(float); break;
	default: t_size = -1; break;
	}
    return( t_size );
}

int image_stride( int w, int h, int c, int t )
//--------------------------------------------
// Calculate line stride : Number of bytes to move to next line of image.
{
	int t_stride = -1;
	
	t_stride = w * c;
	
	switch( t ){
	case IMG_FMT_BYTE:  break;
	case IMG_FMT_WORD:  t_stride <<= 1; break;
	case IMG_FMT_WORD32: t_stride <<= 2; break;
	case IMG_FMT_FLOAT: t_stride *= sizeof(float); break;
	default: t_stride = -1; break;
	}
	
	return( t_stride );
}

int image_stride( image_data* img )
//---------------------------------
// Calculate line stride. 
{
    return( image_stride( img->w, img->h, img->c, img->t ) );
}

image_data* image_data_create( int w, int h, int c, int t, int owns_data )
//------------------------------------------------------------------------
// Create an image data structure for given Width, Height, Channels, Type.
// If [owns_data] allocate a buffer for image data. Otherwise don't.
// Return NULL on error.
{
  image_data* img = NULL;
  int dsize;
  void* data = NULL;
  
  if( w <= 0 || w > IMG_MAX_WIDTH || h <= 0 || h > IMG_MAX_HEIGHT ){
    return( NULL );
  }
  
  dsize = image_sizeof( w, h, c, t );
  
  if( dsize > 0 ){
    if( owns_data )
      data = malloc( dsize );
    
    if( data || ( owns_data == FALSE ) ){
      img = (image_data *)malloc( sizeof(image_data) );
      if( img ){
        img->data.v = data;
        img->w = w;
        img->h = h;
        img->c = c;
        img->t = t;
        img->top_down = FALSE;
        img->owns_data = owns_data;
        img->vmin = 0.0;
        img->vmax = 0.0;
        img->vminmaxset = FALSE; // Well, we could say TRUE, but more useful not to.
      }
      else if( data ) // This is CORRECT for owns_data FALSE too, because then data == NULL. Phew!
        free( data );
    }
  } 
  
  return( img );
}

int image_data_delete( image_data* img )
//--------------------------------------
// Delete an image, Return -1 on error else 0.
{
   if( img ){
       if( img->data.v && img->owns_data )
		   free( img->data.v );
       free( img );
       return( 0 );
   }
   return( -1 );
}

static int filename_to_img_fmt( char* filename )
//----------------------------------------------
// Return an image format code based on the extension part of filename.
{
	char* extp = strrchr( filename, '.' );
	if( extp == NULL )
		return( IMG_FMT_UNKNOWN );
	else{
		if( ! STRICMP( extp, ".tif" ) )
			return( IMG_FMT_TIFF );
		else
			return( IMG_FMT_UNKNOWN );
	}
}

int image_write_file( char* filename, image_data* image )
//-------------------------------------------------------
// Write an image file from an image.
// Image file format is determined solely from the supplied filename extension.
{
	FILE* fout;
	int status = FALSE;
	void* info;

	int img_fmt = filename_to_img_fmt( filename );
	switch( img_fmt ){
	case IMG_FMT_UNKNOWN:
		return( status );

	case IMG_FMT_TIFF:
		fout = fopen( filename, "wb+" );
		if( fout ){
			if( image_tiff_write_header( fout, image->w, image->h, image->c, image->t, &info ) ){
				if( image_tiff_write_data( image, &info ) )
					status = TRUE;
				image_tiff_free_info( &info );
			}
			fclose( fout );
			return( status );
		}
		else
			return( status );
		break;

	default:
		return( status );
	}

	return( status );
}

#define SF(f,t,d,c,v) \
	f.tag = t; \
	f.field = d; \
	f.count = c; \
	f.offval = v;

int image_tiff_write_header( FILE* fin, int w, int h, int c, int t, void** info )
//-------------------------------------------------------------------------------
// Write the vital things to a TIFF file header.
// Keep it as simple as possible:
// - Uncompressed
// - RGB or RGBA or Monochrome (Photometric Interpretation = 2 or 1)
// - Single plane (Planar Configuration = 1 )
// - Entire image in a single strip
// - Only one image in the file (single IFD).
// This is a really *horrible* file format ...
// Return FALSE on error.
{
#define N_TIFF_FIELDS 14
#define TIFF_SHORT    3
#define TIFF_LONG     4
#define TIFF_RATIONAL 5
	TiffHead head;
	TiffField field[N_TIFF_FIELDS];  // Various bits of **** we must fill in...

	unsigned short bits_per_sample[4];
	unsigned int xres[2], yres[2];
	unsigned int nullifd=0;
	int i, shs;

	unsigned int bits_per_sample_offset = sizeof(head)+sizeof(short)+N_TIFF_FIELDS*sizeof(TiffField)+sizeof(int);
	unsigned int xres_offset = bits_per_sample_offset + c*sizeof(short);
	unsigned int yres_offset = xres_offset + 2 * sizeof(int);
	unsigned int data_offset = yres_offset + 2 * sizeof(int);
	unsigned int total_bytes = image_sizeof( w, h, c, t );

	bits_per_sample[0] = 8 * ( total_bytes / ( w * h * c ) );

	xres[0] = 300; xres[1] = 1;
	yres[0] = 300; yres[1] = 1;

	if( c > 4 )
		c = 4;

	head.endian = LOCAL_TIFF_ENDIAN;
	shs = ( head.endian == 0x4d4d ) ? 16 : 0 ;
	head.magic = 42;
	head.ifd_offset = sizeof(head);

#ifdef DEBUG_PRINT
	printf( "w:%d h:%d c:%d t:%d sizeof(TiffField)=%d\n", w, h, c, t, sizeof(TiffField) );
#endif

	// Fill in the field structures to reflect our image.
	// Note - these buggers must be in ascending tag order.
	SF( field[0], 256, TIFF_SHORT, 1, (w<<shs) ); //--------------------- Width
#ifdef DEBUG_PRINT
	printf( "field[0].offval=%d\n", field[0].offval );
#endif
	SF( field[1], 257, TIFF_SHORT, 1, (h<<shs) ); //--------------------- Height
	if( c == 1 ){
	  SF( field[2], 258, TIFF_SHORT, c, (bits_per_sample[0]<<shs) ); //-- Bits per channel - 1 channel, so scalar
	}
	else{
		SF( field[2], 258, TIFF_SHORT, c, bits_per_sample_offset );  //-- Bits per channel - Multiple channels, so vector (sigh)
	}
	SF( field[3], 259, TIFF_SHORT, 1, (1<<shs) ); //--------------------- Compression = NONE
	if( c == 1 ){
	  SF( field[4], 262, TIFF_SHORT, 1, (1<<shs) ); //------------------- Mono: 0 = black, bigger => brighter (sigh)
	}
	else{
	  SF( field[4], 262, TIFF_SHORT, 1, (2<<shs) ); //------------------- RGB
	}
	SF( field[5],  273, TIFF_LONG,  1, data_offset ); //----------------- Where the image data starts
	SF( field[6],  274, TIFF_SHORT, 1, (4<<shs) ); //-------------------- First pixel is at BOTTOM LEFT!!!!!!
	SF( field[7],  277, TIFF_SHORT, 1, (c<<shs) ); //-------------------- Number of channels
	SF( field[8],  278, TIFF_SHORT, 1, (h<<shs) ); //-------------------- All lines ("rows") in one "strip" (sigh)
	SF( field[9],  279, TIFF_LONG,  1, image_sizeof( w, h, c, t ) ); //-- Bytes in the image data
	SF( field[10], 282, TIFF_RATIONAL, 1, xres_offset ); //-------------- X resolution in pixels per resolution unit.
	SF( field[11], 283, TIFF_RATIONAL, 1, yres_offset ); //-------------- Y resolution in pixels per resolution unit.
	SF( field[12], 284, TIFF_SHORT, 1, (1<<shs) ); //-------------------- Channels are interleaved
	SF( field[13], 296, TIFF_SHORT, 1, (1<<shs) ); //-------------------- Resolution unit is the inch.

	struct TiffInfo* ti = (TiffInfo*) malloc( sizeof( struct TiffInfo ) );
	if( ti ){
		memset( ti, 0, sizeof( struct TiffInfo ) );
		ti->fp = fin;
		*info = (void*)ti;

		if( fwrite( &head, sizeof(head), 1, ti->fp ) != 1 )
			goto error;
		ti->entry_count = N_TIFF_FIELDS;
		if( fwrite( &ti->entry_count, sizeof(unsigned short), 1, ti->fp ) != 1 )
			goto error;
		if( fwrite( &field, sizeof(TiffField), N_TIFF_FIELDS, ti->fp ) != N_TIFF_FIELDS )
			goto error;
		if( fwrite( &nullifd, sizeof(unsigned int), 1, ti->fp ) != 1 )
			goto error;
		ti->bytes = total_bytes;
		for( i=1; i<c; i++ )
			bits_per_sample[i] = bits_per_sample[0];
		if( (int)fwrite( bits_per_sample, sizeof(unsigned short), c, ti->fp ) != c )
			goto error;
		if( (int)fwrite( xres, sizeof(unsigned int), 2, ti->fp ) != 2 )
			goto error;
		if( (int)fwrite( yres, sizeof(unsigned int), 2, ti->fp ) != 2 )
			goto error;
		ti->dataoffset = data_offset;
		return( TRUE );
	}

error:
	return FALSE;
}

int image_tiff_write_data( image_data* data, void** info )
//-------------------------------------------------------
// Write data to an open TIFF format file.
{
	struct TiffInfo* ti = (TiffInfo*)(*info);
	
	if( fseek( ti->fp, ti->dataoffset, SEEK_SET ) != 0 )
		return( FALSE );
	if( fwrite( data->data.uc, sizeof(unsigned char), ti->bytes, ti->fp ) != ti->bytes )
		return( FALSE );
	return( TRUE );
}

void image_tiff_free_info( void** info )
//--------------------------------------
// Free up TIFF specific data
{
	struct TiffInfo* ti = (TiffInfo*)(*info);
	if( ti ){
		if( ti->ifd )
			free( ti->ifd );
		if( ti->strip_offsets )
			free( ti->strip_offsets );
		if( ti->strip_counts )
			free( ti->strip_counts );
		free( ti );
	}
}

//*****//
// END //
//*****//
