      program trt2
c**********************************************************************
c        1         2         3         4         5         6         7 v
c23456789012345678901234567890123456789012345678901234567890123456789012
c
c ********************************
c toy ray tracer mark 2
c trt2 - fortran-77 implementation
c ********************************
c VMS version
c ***********
c
c a simple ray tracer renderer.
c the main "purpose" of this code is to serve as a non-trivial and
c arguably "fun" application to run in "retro-computing" environments.
c ok - so it really doesn't have much of a purpose!
c
c a prototype was actually written and debugged in c++ (written so it
c could be translated pretty much directly to fortran-77). it should
c be straightforward (if time consuming) to move it to other languages
c - although that wouldn't exactly show the best features of those
c lanugages.
c
c it is also an experiment in how simple and small a "useful"
c renderer can be - in terms of lines of code, anyway.
c
c the goals were:
c  1) portability
c  2) ease of recoding in different languages
c  3) small memory footprint
c  4) still capable of making pretty pictures
c  5) easy extensibility:
c     a) new geometric primitive types should be straightforward
c        to add
c     b) texture mapping, etc., would be straightforward to
c        add on machines with enough memory to make it feasible.
c
c current features:
c  1) sphere primitives (proving again that ray tracing is mostly
c     a load of balls).
c  2) triangle primitives with optional vertex normals
c     (these will be correctly interpolated)
c  3) stochastic anti-aliasing (basic distributed ray tracing)
c  4) optional shadows, optionally appropriately "soft edged"
c  5) specularly reflective materials (correct specular reflection)
c  6) transparent materials (correct refraction)
c  7) simple surface roiughness factor for materials.
c  8) accelerated ray intersection tests using a space subdivision
c     (voxel) scheme
c  9) images output as text files (well - it is portable!).
c 10) directional lights (raised cosine and barn door types)
c
c having space subdivision acceleration means it isn't all
c that simple! however, this can make a really big difference
c with lots of primitives. well - assuming they can be
c "indexed" (see code). to save memory, primitives that
c are too big cannot be "accelerated".
c
c because this code must work with standard fortran-77,
c recursion is explicitly coded using a ray stack without making
c any actual recursive calls. also, all storage must be
c statically allocated at compile time.
c
c as always with floating point based code, the precision with which
c calculations are done is important. tests for approximate equivalence
c and "nudging" things to avoid degenerate cases is unavoidable. parts o
c the code which may suffer from "precision" issues are highlighted with
c the comment: precision! the parameters eps and mineps can tune some of
c this. note that 32bit floating point is *not* good enough for the grid
c traversal code. it works perfectly with 64bit (well - pretty exhaustiv
c testing has failed to make it go wrong) but *not* with 32 bit. what
c will happen with cdc format 60 bit? so far, so good.
c
c the primary goal of the fortran version of this code is to run on
c emulated cdc mainframes. the total size of the program must be kept
c (well) below 131000 60 bit words.
c
c author: nick glazzard
c         c++ experiments and prototype: november 2003
c         f-77 implementation: july/august 2005
c         trt (mark 1): summer 1984 (lost long ago)
c
c----------------------------------------------------------------------
c main
c----------------------------------------------------------------------
c
      implicit character*1 (a-z)
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c-------------------------
c-- screen common block --
c-------------------------
c size: 3 * dimpxl + 15 words = 12303
c
      integer maxpxl, dimpxl
      parameter( maxpxl=4096 )
      parameter( dimpxl=maxpxl)
c
      integer snx, sny, smaxdp, os, accel, shadow, bkt, curx, cury
      common /screei/ snx, sny, smaxdp, os, accel, shadow, bkt,
     +                curx, cury
c
      real fpd, aper, focal, fstop, minwgt
      real colr, colg, colb
      real backr, backg, backb
      real bkri, bkrf, bkrd
      real bkgi, bkgf, bkgd
      real bkbi, bkbf, bkbd
      real liner(dimpxl), lineg(dimpxl), lineb(dimpxl)
      common /screer/ fpd, aper, focal, fstop, minwgt,
     +                colr, colg, colb,
     +                backr, backg, backb,
     +                bkri, bkrf, bkrd,
     +                bkgi, bkgf, bkgd,
     +                bkbi, bkbf, bkbd,
     +                liner, lineg, lineb
c
c-----------------------------
c-- statistics common block --
c-----------------------------
c size: 8 words.
c
      integer maxdps, maxrys, maxiss, totray, totfel, nacc, noacc
      common /statsi/ maxdps, maxrys, maxiss, totray, totfel, nacc,
     +                noacc
c
      real nsects
      common /statsr/ nsects
c
c---------------------------------
c-- primitive list common block --
c---------------------------------
c size: 5 * dimprm + 1 words = 5121
c
      integer maxprm, dimprm
      parameter( maxprm=1024 )
      parameter( dimprm=maxprm )
c
      integer nprm, matter(dimprm), paccel(dimprm), prayid(dimprm)
      integer prmtyp(dimprm), prmidx(dimprm)
      common /prmlst/ nprm, matter, paccel, prayid, prmtyp, prmidx
c
c-----------------
c-- sphere list --
c-----------------
c size: 4 * dimsph + 1 words = 4097
c
      integer maxsph, dimsph
      parameter( maxsph=1024 )
      parameter( dimsph=maxsph )
c
      integer nsph
      common /sphlsi/ nsph
c
      real sox(dimsph), soy(dimsph), soz(dimsph), srad(dimsph)
      common /sphlsr/ sox, soy, soz, srad
c
c-------------------
c-- triangle list --
c-------------------
c 23 * dimtri + 1 words = 5889
c
      integer maxtri, dimtri
      parameter( maxtri=256 )
      parameter( dimtri=maxtri )
c
      integer ntri, dax(dimtri)
      common /trilsi/ ntri, dax
c
      real tx1(dimtri), ty1(dimtri), tz1(dimtri)
      real tx2(dimtri), ty2(dimtri), tz2(dimtri)
      real tx3(dimtri), ty3(dimtri), tz3(dimtri)
      real tnx(dimtri), tny(dimtri), tnz(dimtri), tnd(dimtri)
      real nvx1(dimtri), nvy1(dimtri), nvz1(dimtri)
      real nvx2(dimtri), nvy2(dimtri), nvz2(dimtri)
      real nvx3(dimtri), nvy3(dimtri), nvz3(dimtri)
      common /trilsr/ tx1, ty1, tz1, tx2, ty2, tz2, tx3, ty3, tz3,
     +                tnx, tny, tnz, nvx1, nvy1, nvz1, nvx2, nvy2,
     +                nvz2, nvx3, nvy3, nvz3, tnd
c
c--------------------------------
c-- material list common block --
c--------------------------------
c size: 14 * dimmtr + 1 words = 449
c
      integer maxmtr, dimmtr
      parameter( maxmtr=32 )
      parameter( dimmtr=maxmtr )
c
      integer nmtr
      common /mtrlsi/ nmtr
c
      real mkr(dimmtr), mkt(dimmtr)
      real mcar(dimmtr), mcag(dimmtr), mcab(dimmtr)
      real mcdr(dimmtr), mcdg(dimmtr), mcdb(dimmtr)
      real mcsr(dimmtr), mcsg(dimmtr), mcsb(dimmtr)
      real meta(dimmtr), mgls(dimmtr), mrgh(dimmtr)
      common /mtrlsr/ mkr, mkt, mcar, mcag, mcab, mcdr, mcdg, mcdb,
     +                mcsr, mcsg, mcsb, meta, mgls, mrgh
c
c-----------------------------
c-- light list common block --
c-----------------------------
c size: 13 * dimlgt + 4 words = 420
c
      integer maxlgt, dimlgt
      parameter( maxlgt=32 )
      parameter( dimlgt=maxlgt )
c
      integer nlgt, ldir(dimlgt)
      common /lgtlsi/ nlgt, ldir
c
      real lcar, lcag, lcab, lox(dimlgt), loy(dimlgt), loz(dimlgt)
      real lclr(dimlgt), lclg(dimlgt), lclb(dimlgt)
      real ldx(dimlgt), ldy(dimlgt), ldz(dimlgt)
      real ltgt(dimlgt), ltg2(dimlgt), lrad(dimlgt)
      common /lgtlsr/ lcar, lcag, lcab, lox, loy, loz, lclr, lclg,
     +                lclb, ldx, ldy, ldz, ltgt, ltg2, lrad
c
c---------------------------
c-- ray list common block --
c---------------------------
c size: 13 * dimray + 2 words = 6658
c
      integer maxray, dimray
      parameter( maxray=512 )
      parameter( dimray=maxray )
c
      integer nlray, raynum, risect(dimray), rtype(dimray)
      integer rdep(dimray), rrayid(dimray)
      common /raylsi/ nlray, raynum, risect, rtype, rdep, rrayid
c
      real rox(dimray), roy(dimray), roz(dimray)
      real rdx(dimray), rdy(dimray), rdz(dimray)
      real rwgtr(dimray), rwgtg(dimray), rwgtb(dimray)
      common /raylsr/ rox, roy, roz, rdx, rdy, rdz, rwgtr, rwgtg, rwgtb
c
c------------------------------------
c-- intersection list common block --
c------------------------------------
c size: 13 * dimins + 1 words = 6683
c
      integer maxins, dimins, newins, shdins
      parameter( maxins=512 )
      parameter( dimins=maxins+2 )
      parameter( newins=maxins+1 )
      parameter( shdins=maxins+2 )
c
      integer nins, iprim(dimins), irray(dimins), itray(dimins)
      integer ittyp(dimins)
      common /inslsi/ nins, iprim, irray, itray, ittyp
c
      real ipx(dimins), ipy(dimins), ipz(dimins)
      real inx(dimins), iny(dimins), inz(dimins)
      real icolr(dimins), icolg(dimins), icolb(dimins)
      common /inslsr/ ipx, ipy, ipz, inx, iny, inz, icolr, icolg, icolb
      integer y, rdok
c
      integer stsecs, ensecs
      call system_clock( stsecs )
c
      write(6,100)
 100  format(1h1,'toy ray tracer mark2 v0.1 fortran-77' )
      write(6,101)
 101  format(1x,'************************************' )
c
c-- initialize the scene database
c
      call initdb
c
c-- initialize statistics
c
      call initst
c
c-- read the scene from "input" and build scene database
c
      call readdb( rdok )
      if( rdok .eq. 0 )then
         write(6,102)
 102     format(1x,'error: failed to read scene database.')
         stop
      endif
c
c-- print information
c
      write(6,103)
 103  format(1x,'screen specification:')
      if( accel .eq. noacl )then
         write(6,104)
 104     format(6x,'            accel: none')
      else
         write(6,105)
 105     format(6x,'            accel: voxel grid subdivision')
      endif
      if( shadow .eq. 0 )then
         write(6,106)
 106     format(6x,'          shadows: no')
      else
         write(6,107)shadow
 107     format(6x,'          shadows: yes,',i5,' feelers')
      endif
      write(6,108)snx, maxpxl
 108  format(6x,'    output pixels:',i5,' max:',i5)
      write(6,109)sny, maxpxl
 109  format(6x,'     output lines:',i5,' max:',i5)
      write(6,110)os
 110  format(6x,'       oversample:',i3,' (squared)')
      write(6,111)aper
 111  format(6x,'         aperture:',f9.4)
      write(6,112)fpd
 112  format(6x,' focal plane dist:',f9.4)
      write(6,113)focal
 113  format(6x,'lens focal length:',f9.4)
      write(6,114)fstop
 114  format(6x,'      lens f-stop:',f9.4)
      if( bkt .eq.bkcnst )then
         write(6,115)bkri, bkgi, bkbi
 115     format(6x,'background: constant',f9.4,',',f9.4,',',f9.4)
      else if( bkt .eq. bkvert )then
         write(6,116)bkri, bkgi, bkbi, bkrf, bkgf, bkbf
 116     format(6x,'background: vertgrad',f9.4,',',f9.4,',',f9.4,
     +              ' to ',f9.4,',',f9.4,',',f9.4)
      else
         write(6,117)bkri, bkgi, bkbi, bkrf, bkgf, bkbf
 117     format(6x,'background: horzgrad',f9.4,',',f9.4,',',f9.4,
     +              ' to ',f9.4,',',f9.4,',',f9.4)
      endif
c
      write(6,118)
 118  format(1x,'constraints:')
      write(6,119)smaxdp
 119  format(6x,'max ray tree depth: ',i5)
      write(6,120)minwgt
 120  format(6x,'min sig ray weight: ',f9.4)
c
      write(6,121)
 121  format(1x,'database statistics:')
      write(6,122)nprm, maxprm
 122  format(6x,'primitives:',i5,' max:',i5)
      write(6,123)nsph, maxsph
 123  format(6x,'   spheres:',i5,' max:',i5)
      write(6,124)ntri, maxtri
 124  format(6x,' triangles:',i5,' max:',i5)
      write(6,125)nmtr, maxmtr
 125  format(6x,' materials:',i5,' max:',i5)
      write(6,126)nlgt, maxlgt
 126  format(6x,'    lights:',i5,' max:',i5)
c
c-- if accelerating, init the accel structures
c
      if( accel .eq. subacl )call stacel
c
c-- step over all the lines of the output image
c-- trace all rays needed to generate the line
c-- shade the line
c-- output the line
c
      write(7,200)snx, sny
 200  format(1x,2(i5))
      do 1 y=1,sny
         if( ( ( y / 10 ) * 10 ) .eq. y )then
            write(6,*)'... line: ', y
         endif
         cury = y
         call initln
         call trcln( y )
         call outln( y )
 1    continue
c
c-- print statistics
c
      write(6,127)
 127  format(1x,'execution statistics:')
      write(6,128)maxrys, maxray
 128  format(6x,'      max rays in tree:',i5,' max:',i5)
      write(6,129)maxiss, maxins
 129  format(6x,'max intersects per ray:',i5,' max:',i5)
      write(6,130)nsects
 130  format(6x,'     total isect tests:',f11.0)
      write(6,131)totray
 131  format(6x,'     total rays traced:',i12)
      write(6,132)totfel
 132  format(6x,'   total light feelers:',i12)
c
      call system_clock( ensecs )
      write(6,133)real(ensecs-stsecs)/10000.0
 133  format(6x, '  Elapsed time (secs):', f11.3)
      stop
      end
c
c
      subroutine outln( y )
      implicit character*1 (a-z)
      integer y
c**********************************************************************
c output a line of the image as formatted text
c**********************************************************************
c
c-------------------------
c-- screen common block --
c-------------------------
c size: 3 * dimpxl + 15 words = 12303
c
      integer maxpxl, dimpxl
      parameter( maxpxl=4096 )
      parameter( dimpxl=maxpxl)
c
      integer snx, sny, smaxdp, os, accel, shadow, bkt, curx, cury
      common /screei/ snx, sny, smaxdp, os, accel, shadow, bkt,
     +                curx, cury
c
      real fpd, aper, focal, fstop, minwgt
      real colr, colg, colb
      real backr, backg, backb
      real bkri, bkrf, bkrd
      real bkgi, bkgf, bkgd
      real bkbi, bkbf, bkbd
      real liner(dimpxl), lineg(dimpxl), lineb(dimpxl)
      common /screer/ fpd, aper, focal, fstop, minwgt,
     +                colr, colg, colb,
     +                backr, backg, backb,
     +                bkri, bkrf, bkrd,
     +                bkgi, bkgf, bkgd,
     +                bkbi, bkbf, bkbd,
     +                liner, lineg, lineb
      integer trec, trem, irec, s, e, i, j, k
      integer pixi(24)
c
c-- enforce legal colour range now.
c
      do 5 i=1,snx
         if( liner(i) .lt. 0.0 )liner(i) = 0.0
         if( liner(i) .gt. 1.0 )liner(i) = 1.0
         if( lineg(i) .lt. 0.0 )lineg(i) = 0.0
         if( lineg(i) .gt. 1.0 )lineg(i) = 1.0
         if( lineb(i) .lt. 0.0 )lineb(i) = 0.0
         if( lineb(i) .gt. 1.0 )lineb(i) = 1.0
 5    continue
c
c-- output 8 rgb values per full text line (72 chars/line)
c-- order rgbrgb...
c
      trec = snx / 8
      trem = snx - ( trec * 8 )
      s = 1
      e = 8
c
c-- output any full text lines
c
      do 1 irec=1,trec
         k = 1
         do 2 j=s,e
            pixi(k) = int( liner(j) * 255.9 )
            k = k + 1
            pixi(k) = int( lineg(j) * 255.9 )
            k = k + 1
            pixi(k) = int( lineb(j) * 255.9 )
            k = k + 1
2        continue
         write(7,100)(pixi(i),i=1,24)
 100     format(1x,24(i3))
         s = s + 8
         e = e + 8
 1    continue
c
c-- output any partial text line
c
      if( trem .ne. 0 )then
         e = s + trem - 1
         k = 1
         do 3 j=s,e
            pixi(k) = int( liner(j) * 255.9 )
            k = k + 1
            pixi(k) = int( lineg(j) * 255.9 )
            k = k + 1
            pixi(k) = int( lineb(j) * 255.9 )
            k = k + 1
3        continue
         write(7,100)(pixi(i),i=1,(3*trem))
      endif
c
      return
      end
c
c----------------------------------------------------------------------
c ray tracing routines
c----------------------------------------------------------------------
c
      subroutine initln
      implicit character*1 (a-z)
c**********************************************************************
c initialize the output line
c**********************************************************************
c
c-------------------------
c-- screen common block --
c-------------------------
c size: 3 * dimpxl + 15 words = 12303
c
      integer maxpxl, dimpxl
      parameter( maxpxl=4096 )
      parameter( dimpxl=maxpxl)
c
      integer snx, sny, smaxdp, os, accel, shadow, bkt, curx, cury
      common /screei/ snx, sny, smaxdp, os, accel, shadow, bkt,
     +                curx, cury
c
      real fpd, aper, focal, fstop, minwgt
      real colr, colg, colb
      real backr, backg, backb
      real bkri, bkrf, bkrd
      real bkgi, bkgf, bkgd
      real bkbi, bkbf, bkbd
      real liner(dimpxl), lineg(dimpxl), lineb(dimpxl)
      common /screer/ fpd, aper, focal, fstop, minwgt,
     +                colr, colg, colb,
     +                backr, backg, backb,
     +                bkri, bkrf, bkrd,
     +                bkgi, bkgf, bkgd,
     +                bkbi, bkbf, bkbd,
     +                liner, lineg, lineb
c
      integer i
      do 1 i=1,snx,1
         liner(i) = 0.0
         lineg(i) = 0.0
         lineb(i) = 0.0
 1    continue
      return
      end
c
c
      subroutine initst
      implicit character*1 (a-z)
c**********************************************************************
c initialize the statistics
c**********************************************************************
c
c-----------------------------
c-- statistics common block --
c-----------------------------
c size: 8 words.
c
      integer maxdps, maxrys, maxiss, totray, totfel, nacc, noacc
      common /statsi/ maxdps, maxrys, maxiss, totray, totfel, nacc,
     +                noacc
c
      real nsects
      common /statsr/ nsects
c
      maxdps = 0
      maxrys = 0
      maxiss = 0
      totray = 0
      totfel = 0
      nacc = 0
      noacc = 0
      nsects = 0.0
      return
      end
c
c
      subroutine trcln( y )
      implicit character*1 (a-z)
      integer y
c**********************************************************************
c trace all rays needed to create line y
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c-------------------------
c-- screen common block --
c-------------------------
c size: 3 * dimpxl + 15 words = 12303
c
      integer maxpxl, dimpxl
      parameter( maxpxl=4096 )
      parameter( dimpxl=maxpxl)
c
      integer snx, sny, smaxdp, os, accel, shadow, bkt, curx, cury
      common /screei/ snx, sny, smaxdp, os, accel, shadow, bkt,
     +                curx, cury
c
      real fpd, aper, focal, fstop, minwgt
      real colr, colg, colb
      real backr, backg, backb
      real bkri, bkrf, bkrd
      real bkgi, bkgf, bkgd
      real bkbi, bkbf, bkbd
      real liner(dimpxl), lineg(dimpxl), lineb(dimpxl)
      common /screer/ fpd, aper, focal, fstop, minwgt,
     +                colr, colg, colb,
     +                backr, backg, backb,
     +                bkri, bkrf, bkrd,
     +                bkgi, bkgf, bkgd,
     +                bkbi, bkbf, bkbd,
     +                liner, lineg, lineb
c
c---------------------------
c-- ray list common block --
c---------------------------
c size: 13 * dimray + 2 words = 6658
c
      integer maxray, dimray
      parameter( maxray=512 )
      parameter( dimray=maxray )
c
      integer nlray, raynum, risect(dimray), rtype(dimray)
      integer rdep(dimray), rrayid(dimray)
      common /raylsi/ nlray, raynum, risect, rtype, rdep, rrayid
c
      real rox(dimray), roy(dimray), roz(dimray)
      real rdx(dimray), rdy(dimray), rdz(dimray)
      real rwgtr(dimray), rwgtg(dimray), rwgtb(dimray)
      common /raylsr/ rox, roy, roz, rdx, rdy, rdz, rwgtr, rwgtg, rwgtb
c
c---- in-memory and globally accessible data structures
c
c----------------------------
c-- ray stack common block --
c----------------------------
c size: dimstk + 2 words = 514
c
      integer maxstk, dimstk
      parameter( maxstk=512 )
      parameter( dimstk=maxstk )
c
      integer nsray, cray
      integer ray(dimstk)
      common /raystk/ nsray, cray, ray
c
c------------------------------------
c-- intersection list common block --
c------------------------------------
c size: 13 * dimins + 1 words = 6683
c
      integer maxins, dimins, newins, shdins
      parameter( maxins=512 )
      parameter( dimins=maxins+2 )
      parameter( newins=maxins+1 )
      parameter( shdins=maxins+2 )
c
      integer nins, iprim(dimins), irray(dimins), itray(dimins)
      integer ittyp(dimins)
      common /inslsi/ nins, iprim, irray, itray, ittyp
c
      real ipx(dimins), ipy(dimins), ipz(dimins)
      real inx(dimins), iny(dimins), inz(dimins)
      real icolr(dimins), icolg(dimins), icolb(dimins)
      common /inslsr/ ipx, ipy, ipz, inx, iny, inz, icolr, icolg, icolb
c
c-----------------------------
c-- statistics common block --
c-----------------------------
c size: 8 words.
c
      integer maxdps, maxrys, maxiss, totray, totfel, nacc, noacc
      common /statsi/ maxdps, maxrys, maxiss, totray, totfel, nacc,
     +                noacc
c
      real nsects
      common /statsr/ nsects
c
c-- function "prototypes"
c
      real ranhlf
c
c-- local variables
c
      real left, right, bottom, top, dp, dpx, dpy, rns, rmag
      real fy, fx, ds, fys, fysj, fxs, fxsj, dsx, dsy, lrad
      real etheta, erad, ex, ey
      integer x, subx, suby, ns, os2, uos, idum
c
c-- calculate the pixel center coordinates at the edges of the aperture
c
      dp = aper / real(snx)
      left = ( -aper + dp ) * 0.5
      right = ( aper - dp ) * 0.5
      bottom = left * real(sny) / real(snx)
      top = right * real(sny) / real(snx)
c
c-- calculate the reciprocal of the magnification of the lens
c
      if( fpd .lt. ( focal + 0.5 ) )fpd = focal + 0.5
      rmag = ( fpd - focal ) / focal
c
c-- scale aperture edge coordinates to focal plane edge coordinates
c
      left = left * rmag
      right = right * rmag
      bottom = bottom * rmag
      top = top * rmag
c
c-- calculate the effective lens radius
c
      lrad = 0.5 * focal / fstop
c
c-- calculate constants for subsampling
c
      if( os .eq. 0 )then
         ds = 0.0
         ns = 1
         rns = 1.0
         os2 = 0
         uos = 1
      else
         ds = 1.0 / os
         ns = os * os
         rns = 1.0 / ns
         os2 = os / 2
         uos = os
      endif
c
c-- calculate the x and y steps on the focal plane between pixel centers
c
      dpx = ( right - left ) / real( snx - 1 )
      dpy = ( top - bottom ) / real( sny - 1 )
      dsx = ds * dpx
      dsy = ds * dpy
c
c-- calculate y coordinate of current line center
c
      fy = bottom + ( y - 1 ) * dpy
c
c-- step over the pixels in the line.
c
      do 1 x=1,snx
         curx = x
         fx = left + ( x - 1 ) * dpx
c
c-- step over y subsamples - and jitter them
c
         do 2 suby=1,uos
            fys = ( suby - os2 - 1 ) * dsy + fy
            fysj = fys + ranhlf(idum) * dsy

c
c-- step over x subsamples - and jitter them.
c
            do 3 subx=1,uos
               fxs = ( subx - os2 - 1 ) * dsx + fx
               fxsj = fxs + ranhlf(idum) * dsx
c
c-- begin tracing rays. there are initially no rays. and no intersection
c
               nlray = 0
               nsray = 0
               nins = 0
               colr = 0.0
               colg = 0.0
               colb = 0.0
c
c-- generate a screen ray from a randomly chosen point on the lens (ex,e
c-- through the focal plane position (fx,fy,screen.fpd)
c
               nlray = nlray + 1
               raynum = raynum + 1
               etheta = ranhlf(idum) * trtpi
               erad = ( ranhlf(idum) + 0.5 ) * lrad
               ex = erad * cos( etheta )
               ey = erad * sin( etheta )
               rrayid(nlray) = raynum
               rtype(nlray) = eyeray
               risect(nlray) = 0
               rwgtr(nlray) = 1.0
               rwgtg(nlray) = 1.0
               rwgtb(nlray) = 1.0
               rdep(nlray) = 1
               rox(nlray) = ex
               roy(nlray) = ey
               roz(nlray) = 0.0
               rdx(nlray) = fxsj - ex
               rdy(nlray) = fysj - ey
               rdz(nlray) = fpd
               call nrmvec( rdx(nlray), rdy(nlray), rdz(nlray) )
               totray = totray + 1
c
c-- push the ray onto the ray stack.
c
               nsray = nsray + 1
               ray(nsray) = nlray
               if( nlray .gt. maxrys )maxrys = nlray
c
c-- while there are rays on the ray stack, pop a ray off
c
 4             continue
               if( nsray .eq. 0 )goto 5
                  cray = ray(nsray)
                  nsray = nsray - 1
c
c-- find intersections of this current ray with primitives.
c-- when there is an intersection, calculate a shade at the intersection
c-- and add this contribution to the pixel colour.
c-- add new reflected and transmitted rays to the ray stack if appropria
c
                  call trace
               goto 4
 5             continue
c
c--- add in the new sample colour to the pixel
c
               liner(x) = liner(x) + colr
               lineg(x) = lineg(x) + colg
               lineb(x) = lineb(x) + colb
c
c-- end loop over subsamples in x
c
 3          continue
c
c-- end loop over subsamples in y
c
 2       continue
c
c-- normalize the result colour for the output pixel
c
         liner(x) = liner(x) * rns
         lineg(x) = lineg(x) * rns
         lineb(x) = lineb(x) * rns
c
c-- end loop over pixels of output line
c
 1    continue
      return
      end
c
c
      subroutine trace
      implicit character*1 (a-z)
c**********************************************************************
c intersect the "current ray" (raystk.cray) with all the primitives in t
c if there is an intersection record the intersection in the intersectio
c set the current ray intersection index to the new intersection.
c find a shade at the intersection and add that contribution to the samp
c spawn new reflected and transmitted rays if appropriate. push them on
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c---------------------------
c-- ray list common block --
c---------------------------
c size: 13 * dimray + 2 words = 6658
c
      integer maxray, dimray
      parameter( maxray=512 )
      parameter( dimray=maxray )
c
      integer nlray, raynum, risect(dimray), rtype(dimray)
      integer rdep(dimray), rrayid(dimray)
      common /raylsi/ nlray, raynum, risect, rtype, rdep, rrayid
c
      real rox(dimray), roy(dimray), roz(dimray)
      real rdx(dimray), rdy(dimray), rdz(dimray)
      real rwgtr(dimray), rwgtg(dimray), rwgtb(dimray)
      common /raylsr/ rox, roy, roz, rdx, rdy, rdz, rwgtr, rwgtg, rwgtb
c
c---- in-memory and globally accessible data structures
c
c----------------------------
c-- ray stack common block --
c----------------------------
c size: dimstk + 2 words = 514
c
      integer maxstk, dimstk
      parameter( maxstk=512 )
      parameter( dimstk=maxstk )
c
      integer nsray, cray
      integer ray(dimstk)
      common /raystk/ nsray, cray, ray
c
c------------------------------------
c-- intersection list common block --
c------------------------------------
c size: 13 * dimins + 1 words = 6683
c
      integer maxins, dimins, newins, shdins
      parameter( maxins=512 )
      parameter( dimins=maxins+2 )
      parameter( newins=maxins+1 )
      parameter( shdins=maxins+2 )
c
      integer nins, iprim(dimins), irray(dimins), itray(dimins)
      integer ittyp(dimins)
      common /inslsi/ nins, iprim, irray, itray, ittyp
c
      real ipx(dimins), ipy(dimins), ipz(dimins)
      real inx(dimins), iny(dimins), inz(dimins)
      real icolr(dimins), icolg(dimins), icolb(dimins)
      common /inslsr/ ipx, ipy, ipz, inx, iny, inz, icolr, icolg, icolb
c
c---------------------------------
c-- primitive list common block --
c---------------------------------
c size: 5 * dimprm + 1 words = 5121
c
      integer maxprm, dimprm
      parameter( maxprm=1024 )
      parameter( dimprm=maxprm )
c
      integer nprm, matter(dimprm), paccel(dimprm), prayid(dimprm)
      integer prmtyp(dimprm), prmidx(dimprm)
      common /prmlst/ nprm, matter, paccel, prayid, prmtyp, prmidx
c
c--------------------------------
c-- material list common block --
c--------------------------------
c size: 14 * dimmtr + 1 words = 449
c
      integer maxmtr, dimmtr
      parameter( maxmtr=32 )
      parameter( dimmtr=maxmtr )
c
      integer nmtr
      common /mtrlsi/ nmtr
c
      real mkr(dimmtr), mkt(dimmtr)
      real mcar(dimmtr), mcag(dimmtr), mcab(dimmtr)
      real mcdr(dimmtr), mcdg(dimmtr), mcdb(dimmtr)
      real mcsr(dimmtr), mcsg(dimmtr), mcsb(dimmtr)
      real meta(dimmtr), mgls(dimmtr), mrgh(dimmtr)
      common /mtrlsr/ mkr, mkt, mcar, mcag, mcab, mcdr, mcdg, mcdb,
     +                mcsr, mcsg, mcsb, meta, mgls, mrgh
c
c-------------------------
c-- screen common block --
c-------------------------
c size: 3 * dimpxl + 15 words = 12303
c
      integer maxpxl, dimpxl
      parameter( maxpxl=4096 )
      parameter( dimpxl=maxpxl)
c
      integer snx, sny, smaxdp, os, accel, shadow, bkt, curx, cury
      common /screei/ snx, sny, smaxdp, os, accel, shadow, bkt,
     +                curx, cury
c
      real fpd, aper, focal, fstop, minwgt
      real colr, colg, colb
      real backr, backg, backb
      real bkri, bkrf, bkrd
      real bkgi, bkgf, bkgd
      real bkbi, bkbf, bkbd
      real liner(dimpxl), lineg(dimpxl), lineb(dimpxl)
      common /screer/ fpd, aper, focal, fstop, minwgt,
     +                colr, colg, colb,
     +                backr, backg, backb,
     +                bkri, bkrf, bkrd,
     +                bkgi, bkgf, bkgd,
     +                bkbi, bkbf, bkbd,
     +                liner, lineg, lineb
c
c-----------------------------
c-- statistics common block --
c-----------------------------
c size: 8 words.
c
      integer maxdps, maxrys, maxiss, totray, totfel, nacc, noacc
      common /statsi/ maxdps, maxrys, maxiss, totray, totfel, nacc,
     +                noacc
c
      real nsects
      common /statsr/ nsects
c
c-- local variables
c
      real ox, oy, oz
      real dx, dy, dz
      real ndoti, eta, reta, cos1, cos22, k2, swgt, tval
      real rwgt, lrwgtr, lrwgtg, lrwgtb
      real twgt, twgtr, twgtg, twgtb
      real cwgtr, cwgtg, cwgtb
      integer ninst, prim, matl, asect, dep, ttyp, idum
c
c-- get useful things about the current ray.
c
      ox = rox(cray)
      oy = roy(cray)
      oz = roz(cray)
      dx = rdx(cray)
      dy = rdy(cray)
      dz = rdz(cray)
      cwgtr = rwgtr(cray)
      cwgtg = rwgtg(cray)
      cwgtb = rwgtb(cray)
      dep = rdep(cray)
      if( dep .gt. maxdps )maxdps = dep
c
c-- test for intersections. if there are any, we need the nearest.
c-- if there was an intersection, its position and surface normal will b
c-- inslst.xxx(inslst.nins+1) after this.
c
      ninst = nins + 1
      if( accel .eq. noacl )then
         call gensct( ox, oy, oz, dx, dy, dz, ninst, asect, tval, 1 )
      else
         call celsct( ox, oy, oz, dx, dy, dz, ninst, asect, tval, 1 )
      endif
c
c-- if there actually was an intersection, update the intersection count
c-- create new reflected and transmitted rays if appropriate.
c
      if( asect .ne. 0 )then
         nins = ninst
         if( ninst .gt. maxiss )maxiss = ninst
         prim = iprim(ninst)
         matl = matter(prim)
         lrwgtr = mkr(matl) * mcdr(matl) * cwgtr
         lrwgtg = mkr(matl) * mcdg(matl) * cwgtg
         lrwgtb = mkr(matl) * mcdb(matl) * cwgtb
         twgtr = mkt(matl) * mcdr(matl) * cwgtr
         twgtg = mkt(matl) * mcdg(matl) * cwgtg
         twgtb = mkt(matl) * mcdb(matl) * cwgtb
         ttyp = ittyp(ninst)
         eta = meta(matl)
c
c-- use maximum of colour component weights to decide if to terminate
c
         rwgt = lrwgtr
         if( lrwgtg .gt. rwgt )rwgt = lrwgtg
         if( lrwgtb .gt. rwgt )rwgt = lrwgtb
         twgt = twgtr
         if( twgtg .gt. twgt )twgt = twgtg
         if( twgtb .gt. twgt )twgt = twgtb
c
c-- flip the normal?
c
         if( ( dx * inx(ninst) + dy * iny(ninst) + dz * inz(ninst) )
     +       .gt. 0.0 )then
            inx(ninst) = -inx(ninst)
            iny(ninst) = -iny(ninst)
            inz(ninst) = -inz(ninst)
         endif
c
c-- calculate the colour at the intersection.
c
         call shade( ninst, matl )
         swgt = 1.0 - mkr(matl) - mkt(matl)
         if( swgt .lt. 0.0 )swgt = 0.0
         colr = colr + ( icolr(ninst) * swgt * cwgtr )
         colg = colg + ( icolg(ninst) * swgt * cwgtg )
         colb = colb + ( icolb(ninst) * swgt * cwgtb )
c
c-- spawn new rays (recurse, effectively) if we are not too deep.
c
         if( dep .lt. smaxdp )then
            dep = dep + 1
c
c-- new reflected ray
c
            if( rwgt .gt. minwgt )then
               ndoti = 2 * ( inx(ninst) * dx + iny(ninst) * dy +
     +                       inz(ninst) * dz )
               nlray = nlray + 1
               raynum = raynum + 1
               rrayid(nlray) = raynum
               rtype(nlray) = rflray
               risect(nlray) = 0
               rwgtr(nlray) = lrwgtr
               rwgtg(nlray) = lrwgtg
               rwgtb(nlray) = lrwgtb
               rdep(nlray) = dep
               rox(nlray) = ipx(ninst)
               roy(nlray) = ipy(ninst)
               roz(nlray) = ipz(ninst)
               rdx(nlray) = dx - ndoti * inx(ninst)
               rdy(nlray) = dy - ndoti * iny(ninst)
               rdz(nlray) = dz - ndoti * inz(ninst)
               call nrmvec( rdx(nlray), rdy(nlray), rdz(nlray) )
               irray(ninst) = nlray
               nsray = nsray + 1
               ray(nsray) = nlray
               totray = totray + 1
               if( nlray .gt. maxrys )maxrys = nlray
            endif
c
c-- new transmitted ray
c
            if( twgt .gt. minwgt )then
               if( ttyp .eq. enter )then
                  reta = eta
               else
                  reta = 1.0 / eta
               endif
               cos1 = -( inx(ninst) * dx + iny(ninst) * dy +
     +                   inz(ninst) * dz )
               cos22 = 1.0 - reta * reta * ( 1.0 - cos1 * cos1 )
               if( cos22 .gt. 0.0 )then
                  k2 = reta * cos1 - sqrt( cos22 )
                  nlray = nlray + 1
                  raynum = raynum + 1
                  rrayid(nlray) = raynum
                  rtype(nlray) = trnray
                  risect(nlray) = 0
                  rwgtr(nlray) = twgtr
                  rwgtg(nlray) = twgtg
                  rwgtb(nlray) = twgtb
                  rdep(nlray) = dep
                  rox(nlray) = ipx(ninst)
                  roy(nlray) = ipy(ninst)
                  roz(nlray) = ipz(ninst)
                  rdx(nlray) = reta * dx + k2 * inx(ninst)
                  rdy(nlray) = reta * dy + k2 * iny(ninst)
                  rdz(nlray) = reta * dz + k2 * inz(ninst)
                  call nrmvec( rdx(nlray), rdy(nlray), rdz(nlray) )
                  itray(ninst) = nlray
                  nsray = nsray + 1
                  ray(nsray) = nlray
                  totray = totray + 1
                  if( nlray .gt. maxrys )maxrys = nlray
               endif
            endif
c
c-- end of less than maximum ray stack depth
c
         endif
c
c-- if this ray doesn't hit anything, add in the background colour.
c
      else
         call calcbk
         colr = colr + backr * cwgtr
         colg = colg + backg * cwgtg
         colb = colb + backb * cwgtb
      endif
c
      return
      end
c
c
      subroutine calcbk
      implicit character*1 (a-z)
c**********************************************************************
c calculate the background colour for graded backgrounds
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c-------------------------
c-- screen common block --
c-------------------------
c size: 3 * dimpxl + 15 words = 12303
c
      integer maxpxl, dimpxl
      parameter( maxpxl=4096 )
      parameter( dimpxl=maxpxl)
c
      integer snx, sny, smaxdp, os, accel, shadow, bkt, curx, cury
      common /screei/ snx, sny, smaxdp, os, accel, shadow, bkt,
     +                curx, cury
c
      real fpd, aper, focal, fstop, minwgt
      real colr, colg, colb
      real backr, backg, backb
      real bkri, bkrf, bkrd
      real bkgi, bkgf, bkgd
      real bkbi, bkbf, bkbd
      real liner(dimpxl), lineg(dimpxl), lineb(dimpxl)
      common /screer/ fpd, aper, focal, fstop, minwgt,
     +                colr, colg, colb,
     +                backr, backg, backb,
     +                bkri, bkrf, bkrd,
     +                bkgi, bkgf, bkgd,
     +                bkbi, bkbf, bkbd,
     +                liner, lineg, lineb
c
      integer cx, cy

      cx = curx - 1
      cy = cury - 1
      if( bkt .eq. bkcnst )then
         backr = bkri
         backg = bkgi
         backb = bkbi
      else if( bkt .eq. bkvert )then
         backr = bkri + bkrd * cy
         backg = bkgi + bkgd * cy
         backb = bkbi + bkbd * cy
      else if( bkt .eq. bkhorz )then
         backr = bkri + bkrd * cx
         backg = bkgi + bkgd * cx
         backb = bkbi + bkbd * cx
      endif
      if( backr .lt. 0.0 )then
         backr = 0.0
      else if( backr .gt. 1.0 )then
         backr = 1.0
      endif
      if( backg .lt. 0.0 )then
         backg = 0.0
      else if( backg .gt. 1.0 )then
         backg = 1.0
      endif
      if( backb .lt. 0.0 )then
         backb = 0.0
      else if( backb .gt. 1.0 )then
         backb = 1.0
      endif
c
      return
      end
c
c----------------------------------------------------------------------
c shading routines
c----------------------------------------------------------------------
c
      subroutine shade( i, matl )
      implicit character*1 (a-z)
      integer i, matl
c**********************************************************************
c find the colour at intersection i where material matl is to be found.
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c------------------------------------
c-- intersection list common block --
c------------------------------------
c size: 13 * dimins + 1 words = 6683
c
      integer maxins, dimins, newins, shdins
      parameter( maxins=512 )
      parameter( dimins=maxins+2 )
      parameter( newins=maxins+1 )
      parameter( shdins=maxins+2 )
c
      integer nins, iprim(dimins), irray(dimins), itray(dimins)
      integer ittyp(dimins)
      common /inslsi/ nins, iprim, irray, itray, ittyp
c
      real ipx(dimins), ipy(dimins), ipz(dimins)
      real inx(dimins), iny(dimins), inz(dimins)
      real icolr(dimins), icolg(dimins), icolb(dimins)
      common /inslsr/ ipx, ipy, ipz, inx, iny, inz, icolr, icolg, icolb
c
c-------------------------
c-- screen common block --
c-------------------------
c size: 3 * dimpxl + 15 words = 12303
c
      integer maxpxl, dimpxl
      parameter( maxpxl=4096 )
      parameter( dimpxl=maxpxl)
c
      integer snx, sny, smaxdp, os, accel, shadow, bkt, curx, cury
      common /screei/ snx, sny, smaxdp, os, accel, shadow, bkt,
     +                curx, cury
c
      real fpd, aper, focal, fstop, minwgt
      real colr, colg, colb
      real backr, backg, backb
      real bkri, bkrf, bkrd
      real bkgi, bkgf, bkgd
      real bkbi, bkbf, bkbd
      real liner(dimpxl), lineg(dimpxl), lineb(dimpxl)
      common /screer/ fpd, aper, focal, fstop, minwgt,
     +                colr, colg, colb,
     +                backr, backg, backb,
     +                bkri, bkrf, bkrd,
     +                bkgi, bkgf, bkgd,
     +                bkbi, bkbf, bkbd,
     +                liner, lineg, lineb
c
c-----------------------------
c-- light list common block --
c-----------------------------
c size: 13 * dimlgt + 4 words = 420
c
      integer maxlgt, dimlgt
      parameter( maxlgt=32 )
      parameter( dimlgt=maxlgt )
c
      integer nlgt, ldir(dimlgt)
      common /lgtlsi/ nlgt, ldir
c
      real lcar, lcag, lcab, lox(dimlgt), loy(dimlgt), loz(dimlgt)
      real lclr(dimlgt), lclg(dimlgt), lclb(dimlgt)
      real ldx(dimlgt), ldy(dimlgt), ldz(dimlgt)
      real ltgt(dimlgt), ltg2(dimlgt), lrad(dimlgt)
      common /lgtlsr/ lcar, lcag, lcab, lox, loy, loz, lclr, lclg,
     +                lclb, ldx, ldy, ldz, ltgt, ltg2, lrad
c
c--------------------------------
c-- material list common block --
c--------------------------------
c size: 14 * dimmtr + 1 words = 449
c
      integer maxmtr, dimmtr
      parameter( maxmtr=32 )
      parameter( dimmtr=maxmtr )
c
      integer nmtr
      common /mtrlsi/ nmtr
c
      real mkr(dimmtr), mkt(dimmtr)
      real mcar(dimmtr), mcag(dimmtr), mcab(dimmtr)
      real mcdr(dimmtr), mcdg(dimmtr), mcdb(dimmtr)
      real mcsr(dimmtr), mcsg(dimmtr), mcsb(dimmtr)
      real meta(dimmtr), mgls(dimmtr), mrgh(dimmtr)
      common /mtrlsr/ mkr, mkt, mcar, mcag, mcab, mcdr, mcdg, mcdb,
     +                mcsr, mcsg, mcsb, meta, mgls, mrgh
c
      real ranhlf
c
      integer l, f
      real dx, dy, dz, d, s, vx, vy, vz, hx, hy, hz, cr, cg, cb
      real lattr, lattg, lattb, p
      real nx, ny, nz, a, lxf, lyf, lzf, lattrf, lattgf, lattbf
c
      integer idum
      data idum/12345/
c
c-- initialize intersection colour to ambient term.
c
      icolr(i) = mcar(matl) * lcar
      icolg(i) = mcag(matl) * lcar
      icolb(i) = mcab(matl) * lcar
c
c-- step over lights. find contribution to intersection colour for each
c
      do 1 l=1,nlgt
c
c-- see if this light is partially shadowed.
c
         if( shadow .eq. 0 )then
            lattr = 1.0
            lattg = 1.0
            lattb = 1.0
          else if( ( shadow .eq. 1 ) .or. ( lrad(l) .lt. eps ) )then
c-- 1 feeler ray = point light source.
            call ltfeel( iprim(i), ipx(i), ipy(i), ipz(i),
     +                   lox(l), loy(l), loz(l), lattr, lattg, lattb )
         else
c-- multiple feeler rays = uniform sphere of light.
            lattr = 0.0
            lattg = 0.0
            lattb = 0.0
            do 2 f=1,shadow
              lxf = lox(l) + ranhlf(idum) * lrad(l)
              lyf = loy(l) + ranhlf(idum) * lrad(l)
              lzf = loz(l) + ranhlf(idum) * lrad(l)
              call ltfeel( iprim(i), ipx(i), ipy(i), ipz(i),
     +                     lxf, lyf, lzf, lattrf, lattgf, lattbf )
              lattr = lattr + lattrf
              lattg = lattg + lattgf
              lattb = lattb + lattbf
 2          continue
            lattr = lattr / shadow
            lattg = lattg / shadow
            lattb = lattb / shadow
         endif
c
c-- find unit vector from surface intersection to light.
c
         dx = lox(l) - ipx(i)
         dy = loy(l) - ipy(i)
         dz = loz(l) - ipz(i)
         call nrmvec( dx, dy, dz )
c
c-- if the light is directional, modify light attenuation based on
c-- angle btw light-surface vector and light direction.
c
         if( ldir(l) .ne. lgtcon )then
            p =  dx * ldx(l) + dy * ldy(l) + dz * ldz(l)
            if( p .lt. 0.0 )then
               p = 0.0
            else
               if( ldir(l) .eq. lgtcos )then
c-- raised cosine sort ...
                  p = p ** ltgt(l)
               else if( ldir(l) .eq. lgtbrn )then
c-- barn door sort ...
                  a = acos( p )
                  if( a .lt. ltgt(l) )then
                     p = 1.0
                  else if( a .gt. ltg2(l) )then
                     p = 0.0
                  else
                     p = ( ( a - ltgt(l) ) / ( ltg2(l) - ltgt(l) ) )
c-- nice smooth hermite ...
                     p = 1.0 - ( p * p * ( 3 - 2 * p ) )
                  endif
               endif
            endif
            lattr = lattr * p
            lattg = lattg * p
            lattb = lattb * p
         endif
c
c-- get the normal and perturb it if there is some roughness.
c
         nx = inx(i)
         ny = iny(i)
         nz = inz(i)
         if( mrgh(matl) .gt. 0.0 )then
            nx = nx + ranhlf(idum) * mrgh(matl)
            ny = ny + ranhlf(idum) * mrgh(matl)
            nz = nz + ranhlf(idum) * mrgh(matl)
            call nrmvec( nx, ny, nz )
         endif
c
c-- diffuse term.
c
         d = dx * nx + dy * ny + dz * nz
         if( d .lt. 0.0 )d = 0.0
c
c-- specular term.
c
         vx = - ipx(i)
         vy = - ipy(i)
         vz = - ipz(i)
         call nrmvec( vx, vy, vz )
c
         hx = 0.5 * ( dx + vx )
         hy = 0.5 * ( dy + vy )
         hz = 0.5 * ( dz + vz )
c
         s = hx * nx + hy * ny + hz * nz
         if( s .lt. 0.0 )then
            s = 0.0
         else
            s = s ** mgls(matl)
         endif
c
c-- find colour due to this light.
c
         cr = lclr(l) * lattr * ( d * mcdr(matl) + s * mcsr(matl) )
         cg = lclg(l) * lattg * ( d * mcdg(matl) + s * mcsg(matl) )
         cb = lclb(l) * lattb * ( d * mcdb(matl) + s * mcsb(matl) )
c
c-- add to intersection colour.
c
         icolr(i) = icolr(i) + cr
         icolg(i) = icolg(i) + cg
         icolb(i) = icolb(i) + cb
 1    continue
c
      return
      end
c
c
      subroutine ltfeel( srfidx, px, py, pz, lx, ly, lz,
     +                   lattr, lattg, lattb )
      implicit character*1 (a-z)
      integer srfidx
      real px, py, pz, lx, ly, lz, lattr, lattg, lattb
c**********************************************************************
c generate a light feeler ray from surface point (px,py,pz) to light (lx
c see what this intersects and generate an appropriate (ish) light atten
c this tries to deal with transparent objects as well as opaque ones. it
c incorrectly, as refraction cannot be taken into account.
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c-----------------------------
c-- statistics common block --
c-----------------------------
c size: 8 words.
c
      integer maxdps, maxrys, maxiss, totray, totfel, nacc, noacc
      common /statsi/ maxdps, maxrys, maxiss, totray, totfel, nacc,
     +                noacc
c
      real nsects
      common /statsr/ nsects
c
c-------------------------
c-- screen common block --
c-------------------------
c size: 3 * dimpxl + 15 words = 12303
c
      integer maxpxl, dimpxl
      parameter( maxpxl=4096 )
      parameter( dimpxl=maxpxl)
c
      integer snx, sny, smaxdp, os, accel, shadow, bkt, curx, cury
      common /screei/ snx, sny, smaxdp, os, accel, shadow, bkt,
     +                curx, cury
c
      real fpd, aper, focal, fstop, minwgt
      real colr, colg, colb
      real backr, backg, backb
      real bkri, bkrf, bkrd
      real bkgi, bkgf, bkgd
      real bkbi, bkbf, bkbd
      real liner(dimpxl), lineg(dimpxl), lineb(dimpxl)
      common /screer/ fpd, aper, focal, fstop, minwgt,
     +                colr, colg, colb,
     +                backr, backg, backb,
     +                bkri, bkrf, bkrd,
     +                bkgi, bkgf, bkgd,
     +                bkbi, bkbf, bkbd,
     +                liner, lineg, lineb
c
c------------------------------------
c-- intersection list common block --
c------------------------------------
c size: 13 * dimins + 1 words = 6683
c
      integer maxins, dimins, newins, shdins
      parameter( maxins=512 )
      parameter( dimins=maxins+2 )
      parameter( newins=maxins+1 )
      parameter( shdins=maxins+2 )
c
      integer nins, iprim(dimins), irray(dimins), itray(dimins)
      integer ittyp(dimins)
      common /inslsi/ nins, iprim, irray, itray, ittyp
c
      real ipx(dimins), ipy(dimins), ipz(dimins)
      real inx(dimins), iny(dimins), inz(dimins)
      real icolr(dimins), icolg(dimins), icolb(dimins)
      common /inslsr/ ipx, ipy, ipz, inx, iny, inz, icolr, icolg, icolb
c
c---------------------------------
c-- primitive list common block --
c---------------------------------
c size: 5 * dimprm + 1 words = 5121
c
      integer maxprm, dimprm
      parameter( maxprm=1024 )
      parameter( dimprm=maxprm )
c
      integer nprm, matter(dimprm), paccel(dimprm), prayid(dimprm)
      integer prmtyp(dimprm), prmidx(dimprm)
      common /prmlst/ nprm, matter, paccel, prayid, prmtyp, prmidx
c
c--------------------------------
c-- material list common block --
c--------------------------------
c size: 14 * dimmtr + 1 words = 449
c
      integer maxmtr, dimmtr
      parameter( maxmtr=32 )
      parameter( dimmtr=maxmtr )
c
      integer nmtr
      common /mtrlsi/ nmtr
c
      real mkr(dimmtr), mkt(dimmtr)
      real mcar(dimmtr), mcag(dimmtr), mcab(dimmtr)
      real mcdr(dimmtr), mcdg(dimmtr), mcdb(dimmtr)
      real mcsr(dimmtr), mcsg(dimmtr), mcsb(dimmtr)
      real meta(dimmtr), mgls(dimmtr), mrgh(dimmtr)
      common /mtrlsr/ mkr, mkt, mcar, mcag, mcab, mcdr, mcdg, mcdb,
     +                mcsr, mcsg, mcsb, meta, mgls, mrgh
c
      real dx, dy, dz, ds, acatr, acatg, acatb, mincat
      integer prim, asect, matl
      real tval, pox, poy, poz
      real obdx, obdy, obdz,obds
c
c-- find distance between surface point & light and a normalized vector
c-- pointing from the surface to the light. if point is very close to li
c-- no attenuation.
c
      lattr = 1.0
      lattg = 1.0
      lattb = 1.0
      totfel = totfel + 1
c
      dx = lx - px
      dy = ly - py
      dz = lz - pz
      ds = dx * dx + dy * dy + dz * dz
      if( ds .lt. mineps )then
         lattr = 1.0
         lattg = 1.0
         lattb = 1.0
         return
      endif
c
c-- attempt to get around dx,dy,dz appearing in gensct/celsct as
c-- indefinites when compiled opt=2 on cdc ftn 5
c-- note that just *outputting* dx,dy,dz in a write statement *did*
c-- fix the problem. it looks like the optimizer had decided that
c-- ds or dx,dy,dz are superfluous by this point. (wrongly, obviously).
c-- *eventually* ;-) introducing obds finally fixed this acceptably.
c
      obds = sqrt( ds )
      obdx = dx / obds
      obdy = dy / obds
      obdz = dz / obds
c
c-- find all intersecting primitives. if a primitive is made of fully
c-- opaque material, or the accumulated attenuation is very great,
c-- abandon search and return latt=0.
c
      acatr = 1.0
      acatg = 1.0
      acatb = 1.0
c
      pox = px
      poy = py
      poz = pz
c
 1    continue
         if( accel .eq. noacl )then
            call gensct( pox, poy, poz, obdx, obdy, obdz, shdins,
     +                   asect, tval, 0 )
         else
            call celsct( pox, poy, poz, obdx, obdy, obdz, shdins,
     +                   asect, tval, 0 )
         endif
c
c-- see if there is an intersection between the surface point being shad
c-- and the light (not beyond it).
c
         if( asect .eq. 0 .or. tval .lt. 0.0 .or. tval .gt. ds )goto 2
         if( asect .ne. 0 )then
c
c-- check intersection is not with primitive we are finding shadows for.
c-- (primitives cannot shade themselves - difficult numerical issues if
c
            if( asect .ne. srfidx )then
c-- primitive index we have intersected with.
               prim = iprim(shdins)
c-- material index primitive is made from.
               matl = matter(prim)
               if( mkt(matl) .lt. minwgt )then
                  lattr = 0.0
                  lattg = 0.0
                  lattb = 0.0
                  return
               endif
c
               acatr = acatr * ( mkt(matl) * mcdr(matl) )
               acatg = acatg * ( mkt(matl) * mcdg(matl) )
               acatb = acatb * ( mkt(matl) * mcdb(matl) )
c
               mincat = acatr
               if( acatg .lt. mincat )mincat = acatg
               if( acatb .lt. mincat )mincat = acatb
               if( mincat .lt. minwgt )then
                  lattr = 0.0
                  lattg = 0.0
                  lattb = 0.0
                  return
               endif
            endif
c
c-- found an intersection (possibly with self).
c-- move a tad further along the feeler and see if we intersect anything
c-- precision!
c
            pox = ipx(shdins) + eps * obdx
            poy = ipy(shdins) + eps * obdy
            poz = ipz(shdins) + eps * obdz
         endif
      goto 1
 2    continue
c
      lattr = acatr
      lattg = acatg
      lattb = acatb
      return
      end
c
c----------------------------------------------------------------------
c intersection testing routines
c----------------------------------------------------------------------
c
      subroutine gensct( ox, oy, oz, dx, dy, dz, ninst, asect, tval,
     +                   urayid )
      implicit character*1 (a-z)
      real ox, oy, oz, dx, dy, dz, tval
      integer ninst, asect, urayid
c**********************************************************************
c general intersection finder. given ray with origin (ox,oy,oz) and unit
c direction vector (dx,dy,dz), the index of the next free element in the
c intersection list (ninst), find the intersection closest to (ox,oy,oz)
c for all primitives. if there is an intersection, return
c (asect) set to the index of the primitive we hit else set it to 0.
c in the case of an intersection, its position and the surface
c normal at that position will be in inslst.xxx(ninst).
c if there is an intersection, return the ray parameter of the
c intersection in (tval).
c if (urayid), use ray id based intersection acceleration.
c this can't be done for shadow feelers.
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c---- in-memory and globally accessible data structures
c
c----------------------------
c-- ray stack common block --
c----------------------------
c size: dimstk + 2 words = 514
c
      integer maxstk, dimstk
      parameter( maxstk=512 )
      parameter( dimstk=maxstk )
c
      integer nsray, cray
      integer ray(dimstk)
      common /raystk/ nsray, cray, ray
c
c------------------------------------
c-- intersection list common block --
c------------------------------------
c size: 13 * dimins + 1 words = 6683
c
      integer maxins, dimins, newins, shdins
      parameter( maxins=512 )
      parameter( dimins=maxins+2 )
      parameter( newins=maxins+1 )
      parameter( shdins=maxins+2 )
c
      integer nins, iprim(dimins), irray(dimins), itray(dimins)
      integer ittyp(dimins)
      common /inslsi/ nins, iprim, irray, itray, ittyp
c
      real ipx(dimins), ipy(dimins), ipz(dimins)
      real inx(dimins), iny(dimins), inz(dimins)
      real icolr(dimins), icolg(dimins), icolb(dimins)
      common /inslsr/ ipx, ipy, ipz, inx, iny, inz, icolr, icolg, icolb
c
c---------------------------
c-- ray list common block --
c---------------------------
c size: 13 * dimray + 2 words = 6658
c
      integer maxray, dimray
      parameter( maxray=512 )
      parameter( dimray=maxray )
c
      integer nlray, raynum, risect(dimray), rtype(dimray)
      integer rdep(dimray), rrayid(dimray)
      common /raylsi/ nlray, raynum, risect, rtype, rdep, rrayid
c
      real rox(dimray), roy(dimray), roz(dimray)
      real rdx(dimray), rdy(dimray), rdz(dimray)
      real rwgtr(dimray), rwgtg(dimray), rwgtb(dimray)
      common /raylsr/ rox, roy, roz, rdx, rdy, rdz, rwgtr, rwgtg, rwgtb
c
c---------------------------------
c-- primitive list common block --
c---------------------------------
c size: 5 * dimprm + 1 words = 5121
c
      integer maxprm, dimprm
      parameter( maxprm=1024 )
      parameter( dimprm=maxprm )
c
      integer nprm, matter(dimprm), paccel(dimprm), prayid(dimprm)
      integer prmtyp(dimprm), prmidx(dimprm)
      common /prmlst/ nprm, matter, paccel, prayid, prmtyp, prmidx
      integer i, sect, pidx
      real t, tmin
c
      asect = 0
      risect(cray) = 0
      do 1 i=1, nprm
         if( (prayid(i) .ne. rrayid(cray)) .or. (urayid .eq. 0) )then
            pidx = prmidx(i)
            if( prmtyp(i) .eq. sphprm )then
               call sphins( ox, oy, oz, dx, dy, dz, pidx, sect, t, i )
            else if( prmtyp(i) .eq. triprm )then
               call triins( ox, oy, oz, dx, dy, dz, pidx, sect, t, i )
            else
               write(6,100)
 100           format( 1x,'internal error: gensct.' )
               write(6,101)
 101           format( 6x,'unknown primitive type.' )
               stop
            endif
            if( sect .ne. 0 )then
               if( asect .eq. 0 )then
                    tmin = t
                    call cpyins( ninst, newins )
                    risect(cray) = ninst
                    asect = i
                    prayid(i) = rrayid(cray)
               else
                  if( t .lt. tmin )then
                     tmin = t
                     call cpyins( ninst, newins )
                     risect(cray) = ninst
                     asect = i
                     rrayid(i) = rrayid(cray)
                  endif
               endif
            endif
         endif
 1    continue
      if( asect .ne. 0 )then
         tval = tmin
      else
         tval = 0.0
      endif
c
      return
      end
c
c
      subroutine celsct( ox, oy, oz, dx, dy, dz, ninst, asect, tval,
     +                   urayid )
      implicit character*1 (a-z)
      real ox, oy, oz, dx, dy, dz, tval
      integer ninst, asect, urayid
c**********************************************************************
c find an intersection using the acceleration structure.
c the arguments are the same as for gensct().
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c---- in-memory and globally accessible data structures
c
c----------------------------
c-- ray stack common block --
c----------------------------
c size: dimstk + 2 words = 514
c
      integer maxstk, dimstk
      parameter( maxstk=512 )
      parameter( dimstk=maxstk )
c
      integer nsray, cray
      integer ray(dimstk)
      common /raystk/ nsray, cray, ray
c
c------------------------------------
c-- intersection list common block --
c------------------------------------
c size: 13 * dimins + 1 words = 6683
c
      integer maxins, dimins, newins, shdins
      parameter( maxins=512 )
      parameter( dimins=maxins+2 )
      parameter( newins=maxins+1 )
      parameter( shdins=maxins+2 )
c
      integer nins, iprim(dimins), irray(dimins), itray(dimins)
      integer ittyp(dimins)
      common /inslsi/ nins, iprim, irray, itray, ittyp
c
      real ipx(dimins), ipy(dimins), ipz(dimins)
      real inx(dimins), iny(dimins), inz(dimins)
      real icolr(dimins), icolg(dimins), icolb(dimins)
      common /inslsr/ ipx, ipy, ipz, inx, iny, inz, icolr, icolg, icolb
c
c---------------------------
c-- ray list common block --
c---------------------------
c size: 13 * dimray + 2 words = 6658
c
      integer maxray, dimray
      parameter( maxray=512 )
      parameter( dimray=maxray )
c
      integer nlray, raynum, risect(dimray), rtype(dimray)
      integer rdep(dimray), rrayid(dimray)
      common /raylsi/ nlray, raynum, risect, rtype, rdep, rrayid
c
      real rox(dimray), roy(dimray), roz(dimray)
      real rdx(dimray), rdy(dimray), rdz(dimray)
      real rwgtr(dimray), rwgtg(dimray), rwgtb(dimray)
      common /raylsr/ rox, roy, roz, rdx, rdy, rdz, rwgtr, rwgtg, rwgtb
c
c---------------------------------
c-- primitive list common block --
c---------------------------------
c size: 5 * dimprm + 1 words = 5121
c
      integer maxprm, dimprm
      parameter( maxprm=1024 )
      parameter( dimprm=maxprm )
c
      integer nprm, matter(dimprm), paccel(dimprm), prayid(dimprm)
      integer prmtyp(dimprm), prmidx(dimprm)
      common /prmlst/ nprm, matter, paccel, prayid, prmtyp, prmidx
c
c------------------
c-- acceleration --
c------------------
c for typical scenes, the "max cells" would be best around 20 to 40.
c (maybe - my tests say 10 is good). but the memory cost is quite high.
c likewise, it would be good to allow largish primitives to be indexed,
c but the maximum possible list size goes up as the cube of the largest
c number of cells allowed in any direction for the primitive to be accep
c for indexing. so the memory cost is (potentially) very high. it would
c possible to use the same list for more than one grid cell if it is ide
c to that for another cell. we don't try to detect that, though.
c if the ratio of the biggest dimension range to the smallest exceeds
c mxdmrt, adjust the ranges for calculating the grid parameters. this se
c to be always a good idea (hence set to 1.0).
c
c size: celdim + 2 * lstdim + 13 words = 1000 + 2 * 8000 + 13 = 17013 wo
c
      integer mxcelx, mxcely, mxcelz, dmcelx, dmcely, dmcelz
      integer celdim, mxcels, mxpcvr, maxlst, lstdim
      real mxdmrt
      parameter( mxcelx=10 )
      parameter( mxcely=10 )
      parameter( mxcelz=10 )
      parameter( dmcelx=mxcelx )
      parameter( dmcely=mxcely )
      parameter( dmcelz=mxcelz )
      parameter( celdim=dmcelx*dmcely*dmcelz )
      parameter( mxcels=mxcelx*mxcely*mxcelz )
      parameter( mxpcvr=2 )
      parameter( maxlst=2*2*2*mxcels )
      parameter( lstdim=maxlst )
      parameter( mxdmrt=1.0 )
c
      integer nlist, anx, any, anz, cells(celdim)
      integer lprim(lstdim), lnext(lstdim)
      common /acceli/ nlist, anx, any, anz, cells, lprim, lnext
c
      real axl, axh, ayl, ayh, azl, azh, adxs, adys, adzs
      common /accelr/ axl, axh, ayl, ayh, azl, azh, adxs, adys, adzs
c
c--------------------
c-- grid traversal --
c--------------------
c size: 10 words.
c
      integer vninst, vasect, vurayi
      common /vscaci/ vninst, vasect, vurayi
c
      real vox, voy, voz, vdx, vdy, vdz, vtmin
      common /vscacr/ vox, voy, voz, vdx, vdy, vdz, vtmin
      real xcell, ycell, zcell
      integer bsect
      real xi, yi, zi
      integer i, sect, pidx
      real t, tmin
c
c-- calculate the origin coordinates in the cell coordinate system.
c
      xcell = ( ( ox - axl ) / adxs ) * anx
      ycell = ( ( oy - ayl ) / adys ) * any
      zcell = ( ( oz - azl ) / adzs ) * anz
c
c-- check we are inside the grid. except for eye rays, we should always
c-- if not, then see if the ray intersects the bounding box of the scene
c
      if( xcell .lt. 0 .or. xcell .ge. mxcelx .or.
     +    ycell .lt. 0 .or. ycell .ge. mxcely .or.
     +    zcell .lt. 0 .or. zcell .ge. mxcelz )then
         call boxins( ox, oy, oz, dx, dy, dz, bsect, xi, yi, zi )
c
c-- the ray doesn't hit the bounding box. no intersection.
c
         if( bsect .eq. 0 )then
            asect = 0
            return
         endif
c
c-- use the intersection point to find the initial grid cell to traverse
c
         xcell = ( ( xi - axl ) / adxs ) * anx
         ycell = ( ( yi - ayl ) / adys ) * any
         zcell = ( ( zi - azl ) / adzs ) * anz
      endif
c
c-- save ray data for use by viscel()
c
      vasect = 0
      vdx = dx
      vdy = dy
      vdz = dz
      vox = ox
      voy = oy
      voz = oz
      vninst = ninst
      vurayi = urayid
c
c-- traverse the cell grid. each cell visited is handled by viscel()
c-- note correction to the direction vector for grid space!
c
      call grdtrv( xcell, ycell, zcell, dx/adxs, dy/adys, dz/adzs,
     +             anx, any, anz )
c
c-- grid traversal ends either when we have fallen off the sides of the
c-- or viscel() has found an intersection. return (asect) in either case
c
      asect = vasect
c
c-- now check all the primitives that were too large to be indexed in th
c-- acceleration structure. one of these may have a closer intersection.
c
      if( asect .ne. 0 )tmin = vtmin
c
      do 1 i=1,nprm
         if( paccel(i) .eq. 0 )then
            if( prayid(i) .ne. rrayid(cray) .or. urayid .eq. 0 )then
               pidx = prmidx(i)
               if( prmtyp(i) .eq. sphprm )then
                  call sphins( ox, oy, oz, dx, dy, dz, pidx, sect,
     +                         t, i )
               else if( prmtyp(i) .eq. triprm )then
                  call triins( ox, oy, oz, dx, dy, dz, pidx, sect,
     +                         t, i )
               else
                  write(6,100)
 100              format( 1x,'internal error: celsct.' )
                  write(6,101)
 101              format( 6x,'unknown primitive type.' )
                  stop
               endif
               if( sect .ne. 0 )then
                  if( asect .eq. 0 )then
                     tmin = t
                     call cpyins( ninst, newins )
                     risect(cray) = ninst
                     asect = i
                     prayid(i) = rrayid(cray)
                  else
                     if( t .lt. tmin )then
                        tmin = t
                        call cpyins( ninst, newins )
                        risect(cray) = ninst
                        asect = i
                        prayid(i) = rrayid(cray)
                     endif
                  endif
               endif
            endif
         endif
 1    continue
c
      if( asect .ne. 0 )then
         tval = tmin
      else
         tval = 0.0
      endif
c
      return
      end
c
c
      subroutine viscel( x, y, z, quit )
      implicit character*1 (a-z)
      integer x, y, z, quit
c**********************************************************************
c process grid cell (x,y,z). if there are any primitives in this cell,
c check them for intersection with the current ray (in vscacc.xxx).
c if an intersection is found, leave the primitive index in vscacc.asect
c with the position and normal stored in inslst.xxx(vscacc.ninst) in
c the usual way. and terminate grid traversal.
c (if we want to stop grid traversal at this cell, return (quit)
c non-zero, else 0.)
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c---- in-memory and globally accessible data structures
c
c----------------------------
c-- ray stack common block --
c----------------------------
c size: dimstk + 2 words = 514
c
      integer maxstk, dimstk
      parameter( maxstk=512 )
      parameter( dimstk=maxstk )
c
      integer nsray, cray
      integer ray(dimstk)
      common /raystk/ nsray, cray, ray
c
c---------------------------
c-- ray list common block --
c---------------------------
c size: 13 * dimray + 2 words = 6658
c
      integer maxray, dimray
      parameter( maxray=512 )
      parameter( dimray=maxray )
c
      integer nlray, raynum, risect(dimray), rtype(dimray)
      integer rdep(dimray), rrayid(dimray)
      common /raylsi/ nlray, raynum, risect, rtype, rdep, rrayid
c
      real rox(dimray), roy(dimray), roz(dimray)
      real rdx(dimray), rdy(dimray), rdz(dimray)
      real rwgtr(dimray), rwgtg(dimray), rwgtb(dimray)
      common /raylsr/ rox, roy, roz, rdx, rdy, rdz, rwgtr, rwgtg, rwgtb
c
c---------------------------------
c-- primitive list common block --
c---------------------------------
c size: 5 * dimprm + 1 words = 5121
c
      integer maxprm, dimprm
      parameter( maxprm=1024 )
      parameter( dimprm=maxprm )
c
      integer nprm, matter(dimprm), paccel(dimprm), prayid(dimprm)
      integer prmtyp(dimprm), prmidx(dimprm)
      common /prmlst/ nprm, matter, paccel, prayid, prmtyp, prmidx
c
c------------------
c-- acceleration --
c------------------
c for typical scenes, the "max cells" would be best around 20 to 40.
c (maybe - my tests say 10 is good). but the memory cost is quite high.
c likewise, it would be good to allow largish primitives to be indexed,
c but the maximum possible list size goes up as the cube of the largest
c number of cells allowed in any direction for the primitive to be accep
c for indexing. so the memory cost is (potentially) very high. it would
c possible to use the same list for more than one grid cell if it is ide
c to that for another cell. we don't try to detect that, though.
c if the ratio of the biggest dimension range to the smallest exceeds
c mxdmrt, adjust the ranges for calculating the grid parameters. this se
c to be always a good idea (hence set to 1.0).
c
c size: celdim + 2 * lstdim + 13 words = 1000 + 2 * 8000 + 13 = 17013 wo
c
      integer mxcelx, mxcely, mxcelz, dmcelx, dmcely, dmcelz
      integer celdim, mxcels, mxpcvr, maxlst, lstdim
      real mxdmrt
      parameter( mxcelx=10 )
      parameter( mxcely=10 )
      parameter( mxcelz=10 )
      parameter( dmcelx=mxcelx )
      parameter( dmcely=mxcely )
      parameter( dmcelz=mxcelz )
      parameter( celdim=dmcelx*dmcely*dmcelz )
      parameter( mxcels=mxcelx*mxcely*mxcelz )
      parameter( mxpcvr=2 )
      parameter( maxlst=2*2*2*mxcels )
      parameter( lstdim=maxlst )
      parameter( mxdmrt=1.0 )
c
      integer nlist, anx, any, anz, cells(celdim)
      integer lprim(lstdim), lnext(lstdim)
      common /acceli/ nlist, anx, any, anz, cells, lprim, lnext
c
      real axl, axh, ayl, ayh, azl, azh, adxs, adys, adzs
      common /accelr/ axl, axh, ayl, ayh, azl, azh, adxs, adys, adzs
c
c--------------------
c-- grid traversal --
c--------------------
c size: 10 words.
c
      integer vninst, vasect, vurayi
      common /vscaci/ vninst, vasect, vurayi
c
      real vox, voy, voz, vdx, vdy, vdz, vtmin
      common /vscacr/ vox, voy, voz, vdx, vdy, vdz, vtmin
c
c------------------------------------
c-- intersection list common block --
c------------------------------------
c size: 13 * dimins + 1 words = 6683
c
      integer maxins, dimins, newins, shdins
      parameter( maxins=512 )
      parameter( dimins=maxins+2 )
      parameter( newins=maxins+1 )
      parameter( shdins=maxins+2 )
c
      integer nins, iprim(dimins), irray(dimins), itray(dimins)
      integer ittyp(dimins)
      common /inslsi/ nins, iprim, irray, itray, ittyp
c
      real ipx(dimins), ipy(dimins), ipz(dimins)
      real inx(dimins), iny(dimins), inz(dimins)
      real icolr(dimins), icolg(dimins), icolb(dimins)
      common /inslsr/ ipx, ipy, ipz, inx, iny, inz, icolr, icolg, icolb
      integer ci, li, lp, pidx
      integer sect, lasect, cx, cy, cz
      real xcell, ycell, zcell, t, tmin
c
c-- see if there is a list of primitives for this cell.
c-- if not continue traversal.
c
      ci = x + y * anx + z * ( anx * any ) + 1
      li = cells(ci)
      if( li .eq. 0 )then
         vasect = 0
         quit = 0
         return
      endif
c
c-- there is a list of primitives starting at list index li.
c-- move over this list looking for intersections with each
c-- primitive on the list in turn.
c
      lasect = 0
 1    continue
         lp = lprim(li)
         if( prayid(lp) .ne. rrayid(cray) .or. vurayi .eq. 0 )then
            pidx = prmidx(lp)
            if( prmtyp(lp) .eq. sphprm )then
               call sphins( vox, voy, voz, vdx, vdy, vdz, pidx,
     +                      sect, t, lp )
            else if( prmtyp(lp) .eq. triprm )then
               call triins( vox, voy, voz, vdx, vdy, vdz, pidx,
     +                      sect, t, lp )
            else
               write(6,100)
 100           format( 1x,'internal error: viscel.' )
               write(6,101)
 101           format( 6x,'unknown primitive type.' )
               stop
            endif
            if( sect .ne. 0 )then
c
c-- check that this intersection lies inside this grid cell.
c
               xcell = ( ( ipx(newins) - axl ) / adxs ) * anx
               ycell = ( ( ipy(newins) - ayl ) / adys ) * any
               zcell = ( ( ipz(newins) - azl ) / adzs ) * anz
               cx = int(xcell)
               cy = int(ycell)
               cz = int(zcell)
c
               if( x .eq. cx .and. y .eq. cy .and. z .eq. cz )then
c
c--  yes - a valid intersection. is it the closest in this cell?
c
                  if( lasect .eq. 0 )then
                     tmin = t
                     call cpyins( vninst, newins )
                     risect(cray) = vninst
                     lasect = lp
                     prayid(lp) = rrayid(cray)
                  else
                     if( t .lt. tmin )then
                        tmin = t
                        call cpyins( vninst, newins )
                        risect(cray) = vninst
                        lasect = lp
                        prayid(lp) = rrayid(cray)
                     endif
                  endif
               endif
            endif
         endif
c
         li = lnext(li)
         if( li .eq. 0 )goto 2
      goto 1
 2    continue
c
c-- no intersection with primitives in this cell. continue traversal.
c
      if( lasect .eq. 0 )then
         vasect = 0
         quit = 0
      else
c
c-- intersection found. terminate traversal.
c
         vasect = lasect
         vtmin = tmin
         quit = 1
      endif
c
      return
      end
c
c
      subroutine sphins( ox, oy, oz, dx, dy, dz, i, sect, t, inprim )
      implicit character*1 (a-z)
      real ox, oy, oz, dx, dy, dz, t
      integer i, sect, inprim
c**********************************************************************
c see if the ray from (ox,oy,oz) direction (dx,dy,dz) intersects
c sphere i. the primitive index is inprim.
c if so, set sect non-zero and set t to the intersection parameter.
c otherwise set sect to zero.
c if there is an intersection save the position and normal in
c inslst.xxx(newins).
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c-----------------
c-- sphere list --
c-----------------
c size: 4 * dimsph + 1 words = 4097
c
      integer maxsph, dimsph
      parameter( maxsph=1024 )
      parameter( dimsph=maxsph )
c
      integer nsph
      common /sphlsi/ nsph
c
      real sox(dimsph), soy(dimsph), soz(dimsph), srad(dimsph)
      common /sphlsr/ sox, soy, soz, srad
c
c------------------------------------
c-- intersection list common block --
c------------------------------------
c size: 13 * dimins + 1 words = 6683
c
      integer maxins, dimins, newins, shdins
      parameter( maxins=512 )
      parameter( dimins=maxins+2 )
      parameter( newins=maxins+1 )
      parameter( shdins=maxins+2 )
c
      integer nins, iprim(dimins), irray(dimins), itray(dimins)
      integer ittyp(dimins)
      common /inslsi/ nins, iprim, irray, itray, ittyp
c
      real ipx(dimins), ipy(dimins), ipz(dimins)
      real inx(dimins), iny(dimins), inz(dimins)
      real icolr(dimins), icolg(dimins), icolb(dimins)
      common /inslsr/ ipx, ipy, ipz, inx, iny, inz, icolr, icolg, icolb
c
c-----------------------------
c-- statistics common block --
c-----------------------------
c size: 8 words.
c
      integer maxdps, maxrys, maxiss, totray, totfel, nacc, noacc
      common /statsi/ maxdps, maxrys, maxiss, totray, totfel, nacc,
     +                noacc
c
      real nsects
      common /statsr/ nsects
      real ocx, ocy, ocz, rs
      real b, c, d, t0, t1, tmin
c
c-- initialize.
c
c      call pmddump
      nsects = nsects + 1
      sect = 0
      t = 0.0
c
c-- solve implicit quadratic equation
c
      ocx = ox - sox(i)
      ocy = oy - soy(i)
      ocz = oz - soz(i)
      rs = srad(i)
      b = -2 * ( dx * ocx + dy * ocy + dz * ocz )
      c = ocx * ocx + ocy * ocy + ocz * ocz - rs * rs
      d = b * b - 4 * c
c
c-- if discriminant -ve, no intersection.
c
      if( d .ge. 0.0 )then
         d = sqrt( d )
c
c-- ok - two possible intersections to be found - t0 and t1.
c-- precision!
c
c-- t1 must be .gt. t0 ... t1 would be exiting the sphere if t0 also val
c
         t1 = 0.5 * ( b + d )
         if( t1 .gt. mineps )then
c
c-- t1 is ahead, so we have an intersection.
c-- so t0 might be ahead and closer ...
c
            tmin = t1
            sect = 1
c
c-- this would be leaving the sphere.
c
            ittyp(newins) = leave
c
            t0 = 0.5 * ( b - d )
            if( t0 .gt. mineps )then
c
c-- t0 is ahead, we have a closer intersection ...
c
               tmin = t0
c
c-- this would be entering the sphere.
c
               ittyp(newins) = enter
            endif
c
c-- fill in all the required info for the new intersection
c
            iprim(newins) = inprim
            irray(newins) = 0
            itray(newins) = 0
            icolr(newins) = 0.0
            icolg(newins) = 0.0
            icolb(newins) = 0.0
            ipx(newins) = ox + tmin * dx
            ipy(newins) = oy + tmin * dy
            ipz(newins) = oz + tmin * dz
            inx(newins) = ( ipx(newins) - sox(i) ) / rs
            iny(newins) = ( ipy(newins) - soy(i) ) / rs
            inz(newins) = ( ipz(newins) - soz(i) ) / rs
c
c--- hand back the intersection parameter
c
            t = tmin
         endif
      endif
c
      return
      end
c
c
      subroutine triins( ox, oy, oz, dx, dy, dz, i, sect, t, inprim )
      implicit character*1 (a-z)
      real ox, oy, oz, dx, dy, dz, t
      integer i, sect, inprim
c**********************************************************************
c see if the ray from (ox,oy,oz) direction (dx,dy,dz) intersects
c triangle i. the primitive index is inprim.
c if so, set sect non-zero and set t to the intersection parameter.
c otherwise set sect to zero.
c if there is an intersection save the position and normal
c in inslst.xxx(newins).
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c-------------------
c-- triangle list --
c-------------------
c 23 * dimtri + 1 words = 5889
c
      integer maxtri, dimtri
      parameter( maxtri=256 )
      parameter( dimtri=maxtri )
c
      integer ntri, dax(dimtri)
      common /trilsi/ ntri, dax
c
      real tx1(dimtri), ty1(dimtri), tz1(dimtri)
      real tx2(dimtri), ty2(dimtri), tz2(dimtri)
      real tx3(dimtri), ty3(dimtri), tz3(dimtri)
      real tnx(dimtri), tny(dimtri), tnz(dimtri), tnd(dimtri)
      real nvx1(dimtri), nvy1(dimtri), nvz1(dimtri)
      real nvx2(dimtri), nvy2(dimtri), nvz2(dimtri)
      real nvx3(dimtri), nvy3(dimtri), nvz3(dimtri)
      common /trilsr/ tx1, ty1, tz1, tx2, ty2, tz2, tx3, ty3, tz3,
     +                tnx, tny, tnz, nvx1, nvy1, nvz1, nvx2, nvy2,
     +                nvz2, nvx3, nvy3, nvz3, tnd
c
c------------------------------------
c-- intersection list common block --
c------------------------------------
c size: 13 * dimins + 1 words = 6683
c
      integer maxins, dimins, newins, shdins
      parameter( maxins=512 )
      parameter( dimins=maxins+2 )
      parameter( newins=maxins+1 )
      parameter( shdins=maxins+2 )
c
      integer nins, iprim(dimins), irray(dimins), itray(dimins)
      integer ittyp(dimins)
      common /inslsi/ nins, iprim, irray, itray, ittyp
c
      real ipx(dimins), ipy(dimins), ipz(dimins)
      real inx(dimins), iny(dimins), inz(dimins)
      real icolr(dimins), icolg(dimins), icolb(dimins)
      common /inslsr/ ipx, ipy, ipz, inx, iny, inz, icolr, icolg, icolb
c
c-----------------------------
c-- statistics common block --
c-----------------------------
c size: 8 words.
c
      integer maxdps, maxrys, maxiss, totray, totfel, nacc, noacc
      common /statsi/ maxdps, maxrys, maxiss, totray, totfel, nacc,
     +                noacc
c
      real nsects
      common /statsr/ nsects
      real ndotd, ndoto, px, py, pz, u0, u1, u2, v0, v1, v2
      real alpha, beta, gamma, npx, npy, npz
c
c-- initialize. bump intersection test count.
c
      nsects = nsects + 1
c
c-- check ray is not parallel to plane embedding triangle.
c-- precision!
c
      ndotd = tnx(i) * dx + tny(i) * dy + tnz(i) * dz
      if( abs( ndotd ) .gt. eps )then
c
c-- find intersection parameter with plane.
c
         ndoto = tnx(i) * ox + tny(i) * oy + tnz(i) * oz
         t = -( ndoto + tnd(i) ) / ndotd
c
c-- find intersection point with embedding plane.
c
         px = ox + t * dx
         py = oy + t * dy
         pz = oz + t * dz
c
c-- find constants for locating intersection point in terms of parametri
c-- coordinates in a coordinate system using two edges of the triangle a
c-- its basis vectors. use the "biggest normal component" precomputed as
c-- to select the plane normal to the least degenerate possible parametr
c
         if( dax(i) .eq. daxx )then
            u0 = py - ty1(i)
            u1 = ty2(i) - ty1(i)
            u2 = ty3(i) - ty1(i)
            v0 = pz - tz1(i)
            v1 = tz2(i) - tz1(i)
            v2 = tz3(i) - tz1(i)
         else if( dax(i) .eq. daxy )then
            u0 = px - tx1(i)
            u1 = tx2(i) - tx1(i)
            u2 = tx3(i) - tx1(i)
            v0 = pz - tz1(i)
            v1 = tz2(i) - tz1(i)
            v2 = tz3(i) - tz1(i)
         else if( dax(i) .eq. daxz )then
            u0 = px - tx1(i)
            u1 = tx2(i) - tx1(i)
            u2 = tx3(i) - tx1(i)
            v0 = py - ty1(i)
            v1 = ty2(i) - ty1(i)
            v2 = ty3(i) - ty1(i)
         else
            write(6,100)
 100        format( 1x,'internal error: triins.' )
            write(6,101)
 101        format( 6x,'unknown primitive type.' )
            stop
         endif
c
c-- find the (alpha,beta) coordinates of the intersection point.
c-- precision!
c
         if( abs( u1 ) .lt. mineps )then
            beta = u0 / u2
            if( beta .ge. 0.0 .and. beta .le. 1.0 )then
               alpha = ( v0 - beta * v2 ) / v1
            else
               t = 0.0
               sect = 0
               return
            endif
         else
            beta = ( v0 * u1 - u0 * v1 ) / ( v2 * u1 - u2 * v1 )
            if( beta .ge. 0.0 .and. beta .le. 1.0 )then
               alpha = ( u0 - beta * u2 ) / u1
            else
               t = 0.0
               sect = 0
               return
            endif
         endif
c
c-- find the gamma coordinate.
c
         gamma = alpha + beta
         if( alpha .lt. 0.0 .or. gamma .gt. 1.0 )then
            t = 0.0
            sect = 0
            return
         endif
c
c-- we are inside the triangle.
c-- use (alpha,beta,gamma) to interpolate the vertex properties (current
c
         npx = gamma * nvx1(i) + alpha * nvx2(i) + beta * nvx3(i)
         npy = gamma * nvy1(i) + alpha * nvy2(i) + beta * nvy3(i)
         npz = gamma * nvz1(i) + alpha * nvz2(i) + beta * nvz3(i)
         call nrmvec( npx, npy, npz )
c
c-- set the intersection data.
c
         iprim(newins) = inprim
         irray(newins) = 0
         itray(newins) = 0
         icolr(newins) = 0.0
         icolg(newins) = 0.0
         icolb(newins) = 0.0
         ipx(newins) = px
         ipy(newins) = py
         ipz(newins) = pz
         inx(newins) = npx
         iny(newins) = npy
         inz(newins) = npz
c
c-- guess whether we are entering or leaving.
c
         ndotd = npx * dx + npy * dy + npz * dz
         if( ndotd .lt. 0.0 )then
            ittyp(newins) = enter
         else
            ittyp(newins) = leave
         endif
         sect = 1
c
c-- ray is parallel to embedding plane. no intersection possible.
c
      else
         t = 0.0
         sect = 0
      endif
c
      return
      end
c
c
      subroutine cpyins( i, j )
      implicit character*1 (a-z)
      integer i, j
c**********************************************************************
c move a newly found intersection from inslst.xxx(j) to inslst.xxx(i)
c**********************************************************************
c
c------------------------------------
c-- intersection list common block --
c------------------------------------
c size: 13 * dimins + 1 words = 6683
c
      integer maxins, dimins, newins, shdins
      parameter( maxins=512 )
      parameter( dimins=maxins+2 )
      parameter( newins=maxins+1 )
      parameter( shdins=maxins+2 )
c
      integer nins, iprim(dimins), irray(dimins), itray(dimins)
      integer ittyp(dimins)
      common /inslsi/ nins, iprim, irray, itray, ittyp
c
      real ipx(dimins), ipy(dimins), ipz(dimins)
      real inx(dimins), iny(dimins), inz(dimins)
      real icolr(dimins), icolg(dimins), icolb(dimins)
      common /inslsr/ ipx, ipy, ipz, inx, iny, inz, icolr, icolg, icolb
      iprim(i) = iprim(j)
      irray(i) = irray(j)
      itray(i) = itray(j)
      icolr(i) = icolr(j)
      icolg(i) = icolg(j)
      icolb(i) = icolb(j)
      ipx(i)   = ipx(j)
      ipy(i)   = ipy(j)
      ipz(i)   = ipz(j)
      inx(i)   = inx(j)
      iny(i)   = iny(j)
      inz(i)   = inz(j)
      ittyp(i) = ittyp(j)
c
      return
      end
c
c----------------------------------------------------------------------
c database routines
c----------------------------------------------------------------------
c
      subroutine initdb
      implicit character*1 (a-z)
c**********************************************************************
c initialize the scene database
c**********************************************************************
c
c---------------------------------
c-- primitive list common block --
c---------------------------------
c size: 5 * dimprm + 1 words = 5121
c
      integer maxprm, dimprm
      parameter( maxprm=1024 )
      parameter( dimprm=maxprm )
c
      integer nprm, matter(dimprm), paccel(dimprm), prayid(dimprm)
      integer prmtyp(dimprm), prmidx(dimprm)
      common /prmlst/ nprm, matter, paccel, prayid, prmtyp, prmidx
c
c-----------------
c-- sphere list --
c-----------------
c size: 4 * dimsph + 1 words = 4097
c
      integer maxsph, dimsph
      parameter( maxsph=1024 )
      parameter( dimsph=maxsph )
c
      integer nsph
      common /sphlsi/ nsph
c
      real sox(dimsph), soy(dimsph), soz(dimsph), srad(dimsph)
      common /sphlsr/ sox, soy, soz, srad
c
c--------------------------------
c-- material list common block --
c--------------------------------
c size: 14 * dimmtr + 1 words = 449
c
      integer maxmtr, dimmtr
      parameter( maxmtr=32 )
      parameter( dimmtr=maxmtr )
c
      integer nmtr
      common /mtrlsi/ nmtr
c
      real mkr(dimmtr), mkt(dimmtr)
      real mcar(dimmtr), mcag(dimmtr), mcab(dimmtr)
      real mcdr(dimmtr), mcdg(dimmtr), mcdb(dimmtr)
      real mcsr(dimmtr), mcsg(dimmtr), mcsb(dimmtr)
      real meta(dimmtr), mgls(dimmtr), mrgh(dimmtr)
      common /mtrlsr/ mkr, mkt, mcar, mcag, mcab, mcdr, mcdg, mcdb,
     +                mcsr, mcsg, mcsb, meta, mgls, mrgh
c
c-----------------------------
c-- light list common block --
c-----------------------------
c size: 13 * dimlgt + 4 words = 420
c
      integer maxlgt, dimlgt
      parameter( maxlgt=32 )
      parameter( dimlgt=maxlgt )
c
      integer nlgt, ldir(dimlgt)
      common /lgtlsi/ nlgt, ldir
c
      real lcar, lcag, lcab, lox(dimlgt), loy(dimlgt), loz(dimlgt)
      real lclr(dimlgt), lclg(dimlgt), lclb(dimlgt)
      real ldx(dimlgt), ldy(dimlgt), ldz(dimlgt)
      real ltgt(dimlgt), ltg2(dimlgt), lrad(dimlgt)
      common /lgtlsr/ lcar, lcag, lcab, lox, loy, loz, lclr, lclg,
     +                lclb, ldx, ldy, ldz, ltgt, ltg2, lrad
c
c---------------------------
c-- ray list common block --
c---------------------------
c size: 13 * dimray + 2 words = 6658
c
      integer maxray, dimray
      parameter( maxray=512 )
      parameter( dimray=maxray )
c
      integer nlray, raynum, risect(dimray), rtype(dimray)
      integer rdep(dimray), rrayid(dimray)
      common /raylsi/ nlray, raynum, risect, rtype, rdep, rrayid
c
      real rox(dimray), roy(dimray), roz(dimray)
      real rdx(dimray), rdy(dimray), rdz(dimray)
      real rwgtr(dimray), rwgtg(dimray), rwgtb(dimray)
      common /raylsr/ rox, roy, roz, rdx, rdy, rdz, rwgtr, rwgtg, rwgtb
c
c-------------------
c-- triangle list --
c-------------------
c 23 * dimtri + 1 words = 5889
c
      integer maxtri, dimtri
      parameter( maxtri=256 )
      parameter( dimtri=maxtri )
c
      integer ntri, dax(dimtri)
      common /trilsi/ ntri, dax
c
      real tx1(dimtri), ty1(dimtri), tz1(dimtri)
      real tx2(dimtri), ty2(dimtri), tz2(dimtri)
      real tx3(dimtri), ty3(dimtri), tz3(dimtri)
      real tnx(dimtri), tny(dimtri), tnz(dimtri), tnd(dimtri)
      real nvx1(dimtri), nvy1(dimtri), nvz1(dimtri)
      real nvx2(dimtri), nvy2(dimtri), nvz2(dimtri)
      real nvx3(dimtri), nvy3(dimtri), nvz3(dimtri)
      common /trilsr/ tx1, ty1, tz1, tx2, ty2, tz2, tx3, ty3, tz3,
     +                tnx, tny, tnz, nvx1, nvy1, nvz1, nvx2, nvy2,
     +                nvz2, nvx3, nvy3, nvz3, tnd
      nprm = 0
      nsph = 0
      ntri = 0
      nmtr = 0
      nlgt = 0
      raynum = 0
c
      return
      end
c
c
      subroutine initbk
      implicit character*1 (a-z)
c**********************************************************************
c initialise things for graded backgrounds
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c-------------------------
c-- screen common block --
c-------------------------
c size: 3 * dimpxl + 15 words = 12303
c
      integer maxpxl, dimpxl
      parameter( maxpxl=4096 )
      parameter( dimpxl=maxpxl)
c
      integer snx, sny, smaxdp, os, accel, shadow, bkt, curx, cury
      common /screei/ snx, sny, smaxdp, os, accel, shadow, bkt,
     +                curx, cury
c
      real fpd, aper, focal, fstop, minwgt
      real colr, colg, colb
      real backr, backg, backb
      real bkri, bkrf, bkrd
      real bkgi, bkgf, bkgd
      real bkbi, bkbf, bkbd
      real liner(dimpxl), lineg(dimpxl), lineb(dimpxl)
      common /screer/ fpd, aper, focal, fstop, minwgt,
     +                colr, colg, colb,
     +                backr, backg, backb,
     +                bkri, bkrf, bkrd,
     +                bkgi, bkgf, bkgd,
     +                bkbi, bkbf, bkbd,
     +                liner, lineg, lineb
      real fn
c
      if( bkt .eq. bkvert )then
         fn = sny - 1
      else if( bkt .eq. bkhorz )then
         fn = snx - 1
      else
         fn = 1.0
      endif
      bkrd = ( bkrf - bkri ) / fn
      bkgd = ( bkgf - bkgi ) / fn
      bkbd = ( bkbf - bkbi ) / fn
c
      return
      end
c
c
      subroutine initlg
      implicit character*1 (a-z)
c**********************************************************************
c normalize and invert the light pointing vectors for directional lights
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c-----------------------------
c-- light list common block --
c-----------------------------
c size: 13 * dimlgt + 4 words = 420
c
      integer maxlgt, dimlgt
      parameter( maxlgt=32 )
      parameter( dimlgt=maxlgt )
c
      integer nlgt, ldir(dimlgt)
      common /lgtlsi/ nlgt, ldir
c
      real lcar, lcag, lcab, lox(dimlgt), loy(dimlgt), loz(dimlgt)
      real lclr(dimlgt), lclg(dimlgt), lclb(dimlgt)
      real ldx(dimlgt), ldy(dimlgt), ldz(dimlgt)
      real ltgt(dimlgt), ltg2(dimlgt), lrad(dimlgt)
      common /lgtlsr/ lcar, lcag, lcab, lox, loy, loz, lclr, lclg,
     +                lclb, ldx, ldy, ldz, ltgt, ltg2, lrad
      integer l
c
      do 1 l=1,nlgt
         if( ldir(l) .ne. lgtcon )then
            call nrmvec( ldx(l), ldy(l), ldz(l) )
            ldx(l) = -ldx(l)
            ldy(l) = -ldy(l)
            ldz(l) = -ldz(l)
         endif
 1    continue
c
      return
      end
c
c
      subroutine initri( i )
      implicit character*1 (a-z)
      integer i
c**********************************************************************
c precompute plane equation and dominant axis info for triangle (i)
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c-------------------
c-- triangle list --
c-------------------
c 23 * dimtri + 1 words = 5889
c
      integer maxtri, dimtri
      parameter( maxtri=256 )
      parameter( dimtri=maxtri )
c
      integer ntri, dax(dimtri)
      common /trilsi/ ntri, dax
c
      real tx1(dimtri), ty1(dimtri), tz1(dimtri)
      real tx2(dimtri), ty2(dimtri), tz2(dimtri)
      real tx3(dimtri), ty3(dimtri), tz3(dimtri)
      real tnx(dimtri), tny(dimtri), tnz(dimtri), tnd(dimtri)
      real nvx1(dimtri), nvy1(dimtri), nvz1(dimtri)
      real nvx2(dimtri), nvy2(dimtri), nvz2(dimtri)
      real nvx3(dimtri), nvy3(dimtri), nvz3(dimtri)
      common /trilsr/ tx1, ty1, tz1, tx2, ty2, tz2, tx3, ty3, tz3,
     +                tnx, tny, tnz, nvx1, nvy1, nvz1, nvx2, nvy2,
     +                nvz2, nvx3, nvy3, nvz3, tnd
      real dxa, dya, dza, dxb, dyb, dzb
      real fnx, fny, fnz, len, nvx, nvy, nvz
c
      dxa = tx2(i) - tx1(i)
      dya = ty2(i) - ty1(i)
      dza = tz2(i) - tz1(i)
      dxb = tx3(i) - tx1(i)
      dyb = ty3(i) - ty1(i)
      dzb = tz3(i) - tz1(i)
      tnx(i) = dya * dzb - dza * dyb
      tny(i) = dza * dxb - dxa * dzb
      tnz(i) = dxa * dyb - dya * dxb
      tnd(i) = - ( tx1(i) * tnx(i) + ty1(i) * tny(i) +
     +             tz1(i) * tnz(i) )
      if( tnx(i) .gt. tny(i) .and. tnx(i) .gt. tnz(i) )then
         dax(i) = daxx
      else if( tny(i) .gt. tnx(i) .and. tny(i) .gt. tnz(i) )then
         dax(i) = daxy
      else
         dax(i) = daxz
      endif
c
c-- if the vertex normal length for a vertex is tiny, substitute a unit
c-- otherwise, ensure the vertex normals are unit vectors.
c
      fnx = tnx(i)
      fny = tny(i)
      fnz = tnz(i)
      call nrmvec( fnx, fny, fnz )
c
      nvx = nvx1(i)
      nvy = nvy1(i)
      nvz = nvz1(i)
      len = nvx * nvx + nvy * nvy + nvz * nvz
c-- precision!
      if( len .lt. (mineps*mineps) )then
         nvx1(i) = fnx
         nvy1(i) = fny
         nvz1(i) = fnz
      else
         len = sqrt( len )
         nvx1(i) = nvx1(i) / len
         nvy1(i) = nvy1(i) / len
         nvz1(i) = nvz1(i) / len
      endif
c
      nvx = nvx2(i)
      nvy = nvy2(i)
      nvz = nvz2(i)
      len = nvx * nvx + nvy * nvy + nvz * nvz
c-- precision!
      if( len .lt. (mineps*mineps) )then
        nvx2(i) = fnx
        nvy2(i) = fny
        nvz2(i) = fnz
      else
        len = sqrt( len )
         nvx2(i) = nvx2(i) / len
         nvy2(i) = nvy2(i) / len
         nvz2(i) = nvz2(i) / len
      endif
c
      nvx = nvx3(i)
      nvy = nvy3(i)
      nvz = nvz3(i)
      len = nvx * nvx + nvy * nvy + nvz * nvz
c-- precision!
      if( len .lt. (mineps*mineps) )then
         nvx3(i) = fnx
         nvy3(i) = fny
         nvz3(i) = fnz
      else
         len = sqrt( len )
         nvx3(i) = nvx3(i) / len
         nvy3(i) = nvy3(i) / len
         nvz3(i) = nvz3(i) / len
      endif
c
      return
      end
c
c
      subroutine readdb( rdok )
      implicit character*1 (a-z)
      integer rdok
c**********************************************************************
c read the scene database.
c the format of the database is pretty horrible.
c return 1 in rdok if ok, 0 if something bad happened.
c
c we use list directed io. any input that supplies the required data
c as set out below and is compatible with the rules for list directed
c reads is acceptable. the input data is described in terms of "groups"
c below. some sets of groups must be repeated in the input to supply
c the specified number of items. hopefully what this means will be
c clear when reading the description that follows.
c
c group 1: tracing control parameters
c accel shadow
c (int)  (int)
c   accel = 1 for accelerated intersection testing (celsct)
c           0 for the simpler "try everything" (gensct)
c  shadow = 0 for no shadows
c           n (integer) for number of feeler rays to use for shadows.
c           n .gt. 1 can give soft shadows.
c
c group 2: output image parameters
c nx    ny    os
c (int) (int) (int)
c      nx = number of output pixels
c      ny = number of output lines
c      os = oversampling rate for jittered stochastic anti-aliasing
c
c group 3: camera parameters
c aper   fpd    focal  fstop
c (real) (real) (real) (real)
c    aper = aperture in film plane (e.g. in mm)
c     fpd = focal plane distance (e.g. in mm)
c           distance from lens center on which you are focussing.
c   focal = focal length of lens (e.g. in mm)
c   fstop = lens fstop (used for depth of field - os .gt. 1)
c
c group 4: tracing termination parameters
c maxdep minwgt
c (int)  (real)
c  maxdep = maximum tree depth over which tracing a ray will stop.
c  minwgt = minimum weight under which tracing a ray will stop.
c
c group 5: background colour parameters
c bktype ibackr ibackg ibackb fbackr fbackg fbackb
c (int)  (real) (real) (real) (real) (real) real)
c  bktype = 0:constant, 1:vertical grade, 2:horiz grade
c
c group 6: materials count
c nummtr
c (int)
c  nummtr = number of materials being specified
c    groups 7,8,9,10 must follow. there must be nummtr repetitions
c    of this set of groups
c
c group 7: ambient material colour
c car    cag    cab
c (real) (real) (real)
c     car = ambient light reflected (red component)
c     cag = ambient light reflected (green component)
c     cab = ambient light reflected (blue component)
c     all usually in the range 0 to 1
c
c group 8: diffuse material colour
c cdr    cdg    cdb
c (real) (real) (real)
c     cdr = diffuse light reflected (red component)
c     cdg = diffuse light reflected (green component)
c     cdb = diffuse light reflected (blue component)
c     all usually in the range 0 to 1
c
c group 9: specular material colour
c csr    csg    csb
c (real) (real) (real)
c     csr = specular light reflected (red component)
c     csg = specular light reflected (green component)
c     csb = specular light reflected (blue component)
c     all usually in the range 0 to 1
c
c group 10: various material parameters
c kr     kt     eta    gloss  rough
c (real) (real) (real) (real) (real)
c      kr = coefficient of reflection
c      kt = coefficient of transmission
c        kr + kt .le. 1.0 usually
c     eta = refractive index (for transmission)
c        eta = 1.0 for "air", "glass" around 1.45, etc.
c   gloss = sharpness of specular highlight
c        gloss = 1.0 to 250.0 (raised cosine profile)
c   rough = roughness of surface (random normal perturbation)
c        rough = 0.0 to 1.0
c
c group 11: ambient light
c lar    lag    lab
c (real) (real) (real)
c     lar = ambient light (red component)
c     lag = ambient light (green component)
c     lab = ambient light (blue component)
c     all usually in the range 0 to 1
c
c group 12: lights count
c numlgt
c (int)
c  numlgt = number of lights to use.
c    groups 13,14 must follow. there must be numlgt repetitions
c    of this pair of groups.
c
c group 13: light properties 1
c ox     oy     oz     clr    clg    clb    radius
c (real) (real) (real) (real) (real) (real) (real)
c      ox = x coordinate of light (camera space) (e.g. in mm)
c      oy = y coordinate of light (camera space) (e.g. in mm)
c      oz = z coordinate of light (camera space) (e.g. in mm)
c     clr = red component of light colour (0 to 1 usually)
c     clg = green component of light colour (0 to 1 usually)
c     clb = blue component of light colour (0 to 1 usually)
c  radius = size of light source (radius of sphere, e.g. in mm)
c
c group 14: light properties 2
c dtype dx     dy     dz     ang1   ang2
c (int) (real) (real) (real) (real) (real)
c   dtype = directional type of light.
c           0 = not directional, 1 = raised cosine, 2 = "barn door"
c      dx = x component of direction vector (not normalised)
c      dy = y component of direction vector (not normalised)
c      dz = z component of direction vector (not normalised)
c    ang1 = power for raised cosine (dtype = 1)
c         = full power half angle (degrees) (dtype = 2)
c    ang2 = zero power half angle (degrees) (dtype = 2)
c
c group 15: sphere object count
c numsph
c (int)
c  numsph = number of spheres in scene.
c     group 16 must follow. there must be numsph repetitions
c     of this group.
c
c group 16: sphere specification
c mtridx ox     oy     oz     rad
c (int)  (real) (real) (real) (real)
c  mtridx = index number of material from which sphere is made.
c           material index numbers start at 1 for the first of
c           the nummtr materials defined (see above).
c      ox = x center position in camera space (e.g. in mm)
c      oy = y center position in camera space (e.g. in mm)
c      oz = z center position in camera space (e.g. in mm)
c     rad = radius of sphere (e.g. in mm)
c
c group 17: triangle object count
c numtri
c (int)
c  numtri = number of triangles in the scene
c     groups 18,19,20,21 must follow. there must be numtri
c     repetitions of this set of groups to define the triangles.
c
c group 18: triangle material
c mtridx
c (int)
c  mtridx = index of material from which the triangle is made.
c           material index numbers start at 1 for the first of
c           the nummtr materials defined (see above).
c
c group 19: first triangle vertex specification
c x1     y1     z1     nx1    ny1    nz1
c (real) (real) (real) (real) (real) (real)
c      x1 = x coordinate in camera space (e.g. in mm)
c      y1 = y coordinate in camera space (e.g. in mm)
c      z1 = z coordinate in camera space (e.g. in mm)
c     nx1 = x component of normal vector at the vertex.
c     ny1 = y component of normal vector at the vertex.
c     nz1 = z component of normal vector at the vertex.
c       if nx1=ny1=nz1 = 0 then a facet normal will be calculated
c       and used at all vertices of the triangle.
c
c group 20: second triangle vertex specification
c x2     y2     z2     nx2    ny2    nz2
c (real) (real) (real) (real) (real) (real)
c see group 19.
c
c group 21: third triangle vertex specification
c x3     y3     z3     nx3    ny3    nz3
c (real) (real) (real) (real) (real) (real)
c see group 19.
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c-------------------------
c-- screen common block --
c-------------------------
c size: 3 * dimpxl + 15 words = 12303
c
      integer maxpxl, dimpxl
      parameter( maxpxl=4096 )
      parameter( dimpxl=maxpxl)
c
      integer snx, sny, smaxdp, os, accel, shadow, bkt, curx, cury
      common /screei/ snx, sny, smaxdp, os, accel, shadow, bkt,
     +                curx, cury
c
      real fpd, aper, focal, fstop, minwgt
      real colr, colg, colb
      real backr, backg, backb
      real bkri, bkrf, bkrd
      real bkgi, bkgf, bkgd
      real bkbi, bkbf, bkbd
      real liner(dimpxl), lineg(dimpxl), lineb(dimpxl)
      common /screer/ fpd, aper, focal, fstop, minwgt,
     +                colr, colg, colb,
     +                backr, backg, backb,
     +                bkri, bkrf, bkrd,
     +                bkgi, bkgf, bkgd,
     +                bkbi, bkbf, bkbd,
     +                liner, lineg, lineb
c
c--------------------------------
c-- material list common block --
c--------------------------------
c size: 14 * dimmtr + 1 words = 449
c
      integer maxmtr, dimmtr
      parameter( maxmtr=32 )
      parameter( dimmtr=maxmtr )
c
      integer nmtr
      common /mtrlsi/ nmtr
c
      real mkr(dimmtr), mkt(dimmtr)
      real mcar(dimmtr), mcag(dimmtr), mcab(dimmtr)
      real mcdr(dimmtr), mcdg(dimmtr), mcdb(dimmtr)
      real mcsr(dimmtr), mcsg(dimmtr), mcsb(dimmtr)
      real meta(dimmtr), mgls(dimmtr), mrgh(dimmtr)
      common /mtrlsr/ mkr, mkt, mcar, mcag, mcab, mcdr, mcdg, mcdb,
     +                mcsr, mcsg, mcsb, meta, mgls, mrgh
c
c-----------------------------
c-- light list common block --
c-----------------------------
c size: 13 * dimlgt + 4 words = 420
c
      integer maxlgt, dimlgt
      parameter( maxlgt=32 )
      parameter( dimlgt=maxlgt )
c
      integer nlgt, ldir(dimlgt)
      common /lgtlsi/ nlgt, ldir
c
      real lcar, lcag, lcab, lox(dimlgt), loy(dimlgt), loz(dimlgt)
      real lclr(dimlgt), lclg(dimlgt), lclb(dimlgt)
      real ldx(dimlgt), ldy(dimlgt), ldz(dimlgt)
      real ltgt(dimlgt), ltg2(dimlgt), lrad(dimlgt)
      common /lgtlsr/ lcar, lcag, lcab, lox, loy, loz, lclr, lclg,
     +                lclb, ldx, ldy, ldz, ltgt, ltg2, lrad
c
c---------------------------------
c-- primitive list common block --
c---------------------------------
c size: 5 * dimprm + 1 words = 5121
c
      integer maxprm, dimprm
      parameter( maxprm=1024 )
      parameter( dimprm=maxprm )
c
      integer nprm, matter(dimprm), paccel(dimprm), prayid(dimprm)
      integer prmtyp(dimprm), prmidx(dimprm)
      common /prmlst/ nprm, matter, paccel, prayid, prmtyp, prmidx
c
c-----------------
c-- sphere list --
c-----------------
c size: 4 * dimsph + 1 words = 4097
c
      integer maxsph, dimsph
      parameter( maxsph=1024 )
      parameter( dimsph=maxsph )
c
      integer nsph
      common /sphlsi/ nsph
c
      real sox(dimsph), soy(dimsph), soz(dimsph), srad(dimsph)
      common /sphlsr/ sox, soy, soz, srad
c
c-------------------
c-- triangle list --
c-------------------
c 23 * dimtri + 1 words = 5889
c
      integer maxtri, dimtri
      parameter( maxtri=256 )
      parameter( dimtri=maxtri )
c
      integer ntri, dax(dimtri)
      common /trilsi/ ntri, dax
c
      real tx1(dimtri), ty1(dimtri), tz1(dimtri)
      real tx2(dimtri), ty2(dimtri), tz2(dimtri)
      real tx3(dimtri), ty3(dimtri), tz3(dimtri)
      real tnx(dimtri), tny(dimtri), tnz(dimtri), tnd(dimtri)
      real nvx1(dimtri), nvy1(dimtri), nvz1(dimtri)
      real nvx2(dimtri), nvy2(dimtri), nvz2(dimtri)
      real nvx3(dimtri), nvy3(dimtri), nvz3(dimtri)
      common /trilsr/ tx1, ty1, tz1, tx2, ty2, tz2, tx3, ty3, tz3,
     +                tnx, tny, tnz, nvx1, nvy1, nvz1, nvx2, nvy2,
     +                nvz2, nvx3, nvy3, nvz3, tnd
      integer i, cline
      character*80 cdiag
c
c-- screen/camera specification.
c
      cline = 1
      cdiag = 'acceleration scheme or shadow feelers'
      read(5,*,end=998,err=999)accel, shadow
      cline = cline + 1
      cdiag = 'resolution or oversampling'
      read(5,*,end=998,err=999)snx, sny, os
      cline = cline + 1
      cdiag = 'camera parameters'
      read(5,*,end=998,err=999)aper, fpd, focal, fstop
      cline = cline + 1
      cdiag = 'trace termination depth or weight'
      read(5,*,end=998,err=999)smaxdp, minwgt
c
c-- check horizontal resolution is in limits
c
      if( snx .gt. maxpxl )then
         write(6,100)snx, maxpxl
 100     format(1x,'horizontal resolution ',i5,' exceeds maximum ',i5)
         goto 999
      endif
c
c-- force oversample rate to be odd.
c
      if( os .gt. 0 )then
         if( ((os/2)*2) .eq. os )os = os + 1
      endif
      cline = cline + 1
      cdiag = 'background specification'
      read(5,*,end=998,err=999)bkt, bkri, bkgi, bkbi, bkrf, bkgf, bkbf
c
c-- check grade type
c
      if( bkt .lt. bkcnst .or. bkt .gt. bkhorz )then
         write(6,101)
 101     format(1x,'invalid background grade type' )
         goto 999
      endif
c
c-- initialize background data
c
      call initbk
c
c-- materials specification.
c
      cline = cline + 1
      cdiag = 'material count'
      read(5,*,end=998,err=999)nmtr
c
c-- check count is in limits
c
      if( nmtr .gt. maxmtr )then
         write(6,102)nmtr, maxmtr
 102     format(1x,'too many materials ',i5,' limit ',i5)
         goto 999
      endif
c
c-- read all those materials
c
      do 1 i=1,nmtr
         cline = cline + 1
         cdiag = 'material ambient'
         read(5,*,end=998,err=999)mcar(i), mcag(i), mcab(i)
         cline = cline + 1
         cdiag = 'material diffuse'
         read(5,*,end=998,err=999)mcdr(i), mcdg(i), mcdb(i)
         cline = cline + 1
         cdiag = 'material specular'
         read(5,*,end=998,err=999)mcsr(i), mcsg(i), mcsb(i)
         cline = cline + 1
         cdiag = 'material other'
         read(5,*,end=998,err=999)mkr(i), mkt(i), meta(i),
     +         mgls(i), mrgh(i)
 1    continue
c
c-- read ambient light
c
      cline = cline + 1
      cdiag = 'light ambient'
      read(5,*,end=998,err=999)lcar, lcag, lcab
c
c-- light count
c
      cline = cline + 1
      cdiag = 'light count'
      read(5,*,end=998,err=999)nlgt
c
c-- check count is in limits
c
      if( nlgt .gt. maxlgt )then
         write(6,109)nlgt, maxlgt
 109     format(1x,'too many lights ',i5,' limit ',i5)
         goto 999
      endif
c
c-- read all those lights
c
      do 2 i=1,nlgt
         cline = cline + 1
         cdiag = 'light pos, col, rad'
         read(5,*,end=998,err=999)lox(i), loy(i), loz(i),
     +      lclr(i), lclg(i), lclb(i), lrad(i)
         cline = cline + 1
         cdiag = 'light dir, vec, ang'
         read(5,*,end=998,err=999)ldir(i), ldx(i), ldy(i), ldz(i),
     +      ltgt(i), ltg2(i)
         if( ltg2(i) .le. ltgt(i) )ltg2(i) = ltgt(i) + 0.0001
         ltgt(i) = ltgt(i) * degrad
         ltg2(i) = ltg2(i) * degrad
 2    continue
c
c-- initialize the lights database
c
      call initlg
c
c-- no geometric primitives yet
c
      nprm = 0
c
c-- sphere count
c
      cline = cline + 1
      cdiag = 'sphere count'
      read(5,*,end=998,err=999)nsph
c
c-- check count is in limits
c
      if( nsph .gt. maxsph )then
         write(6,103)nsph, maxsph
 103     format(1x,'too many spheres ',i5,' limit ',i5)
         goto 999
      endif
c
c-- read all those spheres
c
      do 3 i=1,nsph
         nprm = nprm + 1
         if( nprm .gt. maxprm )then
            write(6,104)nprm, maxprm
 104        format(1x,'too many primitives ',i5,' limit ',i5)
            goto 999
         endif
         cline = cline + 1
         cdiag = 'sphere'
         read(5,*,end=998,err=999)matter(nprm),
     +       sox(i), soy(i), soz(i), srad(i)
         if( matter(nprm) .gt. nmtr .or. matter(nprm) .lt. 1 )then
            write(6,105) matter(nprm), cline
 105        format(1x,'mtr idx ',i5,' invalid. line ',i5)
            goto 999
         endif
c-- not in acceleration structure.
         paccel(nprm) = 0
c-- no ray has intersected it yet.
         prayid(nprm) = 0
c-- primitive type: sphere.
         prmtyp(nprm) = sphprm
c-- index of definition in sphere table.
         prmidx(nprm) = i
 3    continue
c
c-- triangle count
c
      cline = cline + 1
      cdiag = 'triangle count'
      read(5,*,end=998,err=999)ntri
c
c-- check count is in limits
c
      if( ntri .gt. maxtri )then
         write(6,106)ntri, maxtri
 106     format(1x,'too many triangles ',i5,' limit ',i5)
         goto 999
      endif
c
c-- read all those triangles
c
      do 4 i=1,ntri
         nprm = nprm + 1
         if( nprm .gt. maxprm )then
            write(6,104)nprm, maxprm
            goto 999
         endif
         cline = cline + 1
         cdiag = 'triangle mtr'
         read(5,*,end=998,err=999)matter(nprm)
         if( matter(nprm) .gt. nmtr .or. matter(nprm) .lt. 1 )then
            write(6,105) matter(nprm), cline
            goto 999
         endif
         cline = cline + 1
         cdiag = 'triangle v1'
         read(5,*,end=998,err=999)tx1(i), ty1(i), tz1(i),
     +     nvx1(i), nvy1(i), nvz1(i)
         cline = cline + 1
         cdiag = 'triangle v2'
         read(5,*,end=998,err=999)tx2(i), ty2(i), tz2(i),
     +     nvx2(i), nvy2(i), nvz2(i)
         cline = cline + 1
         cdiag = 'triangle v3'
         read(5,*,end=998,err=999)tx3(i), ty3(i), tz3(i),
     +     nvx3(i), nvy3(i), nvz3(i)
c
c-- initialize the triangle database (for this triangle)
c
         call initri( i )
c-- not in acceleration structure.
         paccel(nprm) = 0
c-- no ray has intersected it yet.
         prayid(nprm) = 0
c-- primitive type: sphere.
         prmtyp(nprm) = triprm
c-- index of definition in sphere table.
         prmidx(nprm) = i
 4    continue
c
c-- finished successfully.
c
      rdok = 1
      return
c
c-- error on read
c
999   continue
      write(6,107)cdiag, cline
107   format(1x,'read error doing: ',a,' line ',i5)
      rdok = 0
      return
c
c-- end of file on read
c
998   continue
      write(6,108)cline
108   format(1x,'end of file on read at line ',i5)
      rdok = 0
      return
      end
c
c----------------------------------------------------------------------
c grid space subdivision acceleration routines
c----------------------------------------------------------------------
c
      subroutine stacel
      implicit character*1 (a-z)
c**********************************************************************
c set up the acceleration structure. this is a simple rectangular space
c subdivision scheme. (but are these things ever really simple?)
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c-----------------------------
c-- statistics common block --
c-----------------------------
c size: 8 words.
c
      integer maxdps, maxrys, maxiss, totray, totfel, nacc, noacc
      common /statsi/ maxdps, maxrys, maxiss, totray, totfel, nacc,
     +                noacc
c
      real nsects
      common /statsr/ nsects
c
c---------------------------------
c-- primitive list common block --
c---------------------------------
c size: 5 * dimprm + 1 words = 5121
c
      integer maxprm, dimprm
      parameter( maxprm=1024 )
      parameter( dimprm=maxprm )
c
      integer nprm, matter(dimprm), paccel(dimprm), prayid(dimprm)
      integer prmtyp(dimprm), prmidx(dimprm)
      common /prmlst/ nprm, matter, paccel, prayid, prmtyp, prmidx
c
c------------------
c-- acceleration --
c------------------
c for typical scenes, the "max cells" would be best around 20 to 40.
c (maybe - my tests say 10 is good). but the memory cost is quite high.
c likewise, it would be good to allow largish primitives to be indexed,
c but the maximum possible list size goes up as the cube of the largest
c number of cells allowed in any direction for the primitive to be accep
c for indexing. so the memory cost is (potentially) very high. it would
c possible to use the same list for more than one grid cell if it is ide
c to that for another cell. we don't try to detect that, though.
c if the ratio of the biggest dimension range to the smallest exceeds
c mxdmrt, adjust the ranges for calculating the grid parameters. this se
c to be always a good idea (hence set to 1.0).
c
c size: celdim + 2 * lstdim + 13 words = 1000 + 2 * 8000 + 13 = 17013 wo
c
      integer mxcelx, mxcely, mxcelz, dmcelx, dmcely, dmcelz
      integer celdim, mxcels, mxpcvr, maxlst, lstdim
      real mxdmrt
      parameter( mxcelx=10 )
      parameter( mxcely=10 )
      parameter( mxcelz=10 )
      parameter( dmcelx=mxcelx )
      parameter( dmcely=mxcely )
      parameter( dmcelz=mxcelz )
      parameter( celdim=dmcelx*dmcely*dmcelz )
      parameter( mxcels=mxcelx*mxcely*mxcelz )
      parameter( mxpcvr=2 )
      parameter( maxlst=2*2*2*mxcels )
      parameter( lstdim=maxlst )
      parameter( mxdmrt=1.0 )
c
      integer nlist, anx, any, anz, cells(celdim)
      integer lprim(lstdim), lnext(lstdim)
      common /acceli/ nlist, anx, any, anz, cells, lprim, lnext
c
      real axl, axh, ayl, ayh, azl, azh, adxs, adys, adzs
      common /accelr/ axl, axh, ayl, ayh, azl, azh, adxs, adys, adzs
      integer i, j, k, p, pidx
      real pxl, pxh, pyl, pyh, pzl, pzh
      real dxc, dyc, dzc
      real cxl, cxh, cyl, cyh, czl, czh
      integer ixl, ixh, iyl, iyh, izl, izh
      real maxdir, mindir, xc, yc, zc
c
      anx = mxcelx
      any = mxcely
      anz = mxcelz
c
c-- set all "pointers" to primitive lists in the cells to "null".
c-- primitive lists are empty
c
      do 1 k=0,anz-1
         do 2 j=0,any-1
            do 3 i=0,anx-1
               cells(i+j*anx+k*(anx*any)+1) = 0
 3          continue
 2       continue
 1    continue
      nlist = 0
c
c-- survey the scene to get its bounds.
c
      do 4 p=1,nprm
         pidx = prmidx(p)
         if( prmtyp(p) .eq. sphprm )then
            call sphbnd( pidx, pxl, pxh, pyl, pyh, pzl, pzh )
         else if( prmtyp(p) .eq. triprm )then
            call tribnd( pidx, pxl, pxh, pyl, pyh, pzl, pzh )
         else
            write(6,100)
 100        format( 1x,'internal error: stacel.' )
            write(6,101)
 101        format( 6x,'unknown primitive type.' )
            stop
         endif
         if( p .eq. 1 )then
            axl = pxl
            axh = pxh
            ayl = pyl
            ayh = pyh
            azl = pzl
            azh = pzh
         else
            if( pxl .lt. axl )axl = pxl
            if( pxh .gt. axh )axh = pxh
            if( pyl .lt. ayl )ayl = pyl
            if( pyh .gt. ayh )ayh = pyh
            if( pzl .lt. azl )azl = pzl
            if( pzh .gt. azh )azh = pzh
         endif
 4    continue
      adxs = axh - axl
      adys = ayh - ayl
      adzs = azh - azl
c
c-- output useful information
c
      write(6,102)
 102  format(1x,'acceleration structure information:')
      write(6,103)anx, any, anz
 103  format(1x,'grid dimensions         = ',i4,' x ',i4,' x ',i4)
      write(6,104)mxpcvr, mxpcvr, mxpcvr
 104  format(1x,'indexable size (cells)  = ',i4,' x ',i4,' x ',i4)
      write(6,105)mxdmrt
 105  format(1x,'adjust if ratio exceeds = ',f8.3)
      write(6,106)axl, axh
 106  format(1x,'scene x range           = (',f8.3,':',f8.3,')')
      write(6,107)ayl, ayh
 107  format(1x,'scene y range           = (',f8.3,':',f8.3,')')
      write(6,108)azl, azh
 108  format(1x,'scene z range           = (',f8.3,':',f8.3,')')
c
c-- optionally, choose the biggest axis to scale the whole grid
c-- if there is a big disparity in the dimensions of the axes, this can
c-- ensure that cells are made big enough to still index primitives. if
c-- are made very small in one direction, it may be that too many are co
c-- otherwise, and the acceleration scheme can't be used.
c-- it seems like this is almost always a good idea anyway.
c
      maxdir = adxs
      if( adys .gt. maxdir )maxdir = adys
      if( adzs .gt. maxdir )maxdir = adzs
c
      mindir = adxs
      if( adys .lt. mindir )mindir = adys
      if( adzs .lt. mindir )mindir = adzs
c
c--tune this for when to do it...
c
      if( ( maxdir / mindir ) .gt. mxdmrt )then
         xc = 0.5 * ( axl + axh )
         yc = 0.5 * ( ayl + ayh )
         zc = 0.5 * ( azl + azh )
         axl = xc - 0.5 * maxdir
         axh = xc + 0.5 * maxdir
         ayl = yc - 0.5 * maxdir
         ayh = yc + 0.5 * maxdir
         azl = zc - 0.5 * maxdir
         azh = zc + 0.5 * maxdir
         adxs = maxdir
         adys = maxdir
         adzs = maxdir
         write(6,109)
 109     format(1x,'scene ranges adjusted for dimensional uniformity')
         write(6,110)axl, axh
 110     format(1x,'scene x range (adj)     = (',f8.3,':',f8.3,')')
         write(6,111)ayl, ayh
 111     format(1x,'scene y range (adj)     = (',f8.3,':',f8.3,')')
         write(6,112)azl, azh
 112     format(1x,'scene z range (adj)     = (',f8.3,':',f8.3,')')
      endif
c
c-- calculate the size of each cell in eye space
c
      dxc = adxs / anx
      dyc = adys / any
      dzc = adzs / anz
      write(6,113)dxc, dyc, dzc
 113  format(1x,'cell sizes              = ',f8.3,' x',f8.3,' x',f8.3)
c
c-- go over the primitives again, assigning them to cells.
c-- if the primitive bounds cover more than mxpcvr cells in any dimensio
c-- don't use acceleration for that primitive. this is to limit the size
c-- the cell lists.
c
      do 5 p=1,nprm
         pidx = prmidx(p)
         if( prmtyp(p) .eq. sphprm )then
            call sphbnd( pidx, pxl, pxh, pyl, pyh, pzl, pzh )
         else if( prmtyp(p) .eq. triprm )then
            call tribnd( pidx, pxl, pxh, pyl, pyh, pzl, pzh )
         else
            write(6,100)
            write(6,101)
            stop
         endif
         cxl = ( pxl - axl ) / dxc
         cxh = ( pxh - axl ) / dxc
         cyl = ( pyl - ayl ) / dyc
         cyh = ( pyh - ayl ) / dyc
         czl = ( pzl - azl ) / dzc
         czh = ( pzh - azl ) / dzc
c
c-- move to integer grid cell coordinates.
c
         ixl = int(cxl)
         ixh = int(cxh)
         iyl = int(cyl)
         iyh = int(cyh)
         izl = int(czl)
         izh = int(czh)
c
c-- add primitive to the list of primitives for each cell covered by the
c-- if the primitive is too big, don't use the acceleration structure at
c-- this is to limit the size of the primitive lists in nasty cases.
c
         if( ixl .eq. ixh .and. iyl .eq. iyh .and. izl .eq. izh )then
            paccel(p) = 1
            nacc = nacc + 1
            call addlst( ixl, iyl, izl, p )
         else
            if( ( ixh - ixl ) .ge. mxpcvr .or.
     +          ( iyh - iyl ) .ge. mxpcvr .or.
     +          ( izh - izl ) .ge. mxpcvr )then
               paccel(p) = 0
               noacc = noacc + 1
            else
               paccel(p) = 1
               nacc = nacc + 1
               do 6 k=izl,izh
                  do 7 j=iyl,iyh
                     do 8 i=ixl,ixh
                        call addlst( i, j, k, p )
 8                   continue
 7                continue
 6             continue
            endif
         endif
 5    continue
c
      write(6,114)nacc
 114  format(1x,'primitives indexed      = ',i5)
      write(6,115)noacc
 115  format(1x,'primitives not indexed  = ',i5)
c
      return
      end
c
c
      subroutine addlst( i, j, k, p )
      implicit character*1 (a-z)
      integer i, j, k, p
c**********************************************************************
c add primitive p to the list associated with cell (i,j,k) if it isn't
c already on the list. *** n.b.: (i,j,k) are zero based! ***
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c------------------
c-- acceleration --
c------------------
c for typical scenes, the "max cells" would be best around 20 to 40.
c (maybe - my tests say 10 is good). but the memory cost is quite high.
c likewise, it would be good to allow largish primitives to be indexed,
c but the maximum possible list size goes up as the cube of the largest
c number of cells allowed in any direction for the primitive to be accep
c for indexing. so the memory cost is (potentially) very high. it would
c possible to use the same list for more than one grid cell if it is ide
c to that for another cell. we don't try to detect that, though.
c if the ratio of the biggest dimension range to the smallest exceeds
c mxdmrt, adjust the ranges for calculating the grid parameters. this se
c to be always a good idea (hence set to 1.0).
c
c size: celdim + 2 * lstdim + 13 words = 1000 + 2 * 8000 + 13 = 17013 wo
c
      integer mxcelx, mxcely, mxcelz, dmcelx, dmcely, dmcelz
      integer celdim, mxcels, mxpcvr, maxlst, lstdim
      real mxdmrt
      parameter( mxcelx=10 )
      parameter( mxcely=10 )
      parameter( mxcelz=10 )
      parameter( dmcelx=mxcelx )
      parameter( dmcely=mxcely )
      parameter( dmcelz=mxcelz )
      parameter( celdim=dmcelx*dmcely*dmcelz )
      parameter( mxcels=mxcelx*mxcely*mxcelz )
      parameter( mxpcvr=2 )
      parameter( maxlst=2*2*2*mxcels )
      parameter( lstdim=maxlst )
      parameter( mxdmrt=1.0 )
c
      integer nlist, anx, any, anz, cells(celdim)
      integer lprim(lstdim), lnext(lstdim)
      common /acceli/ nlist, anx, any, anz, cells, lprim, lnext
c
      real axl, axh, ayl, ayh, azl, azh, adxs, adys, adzs
      common /accelr/ axl, axh, ayl, ayh, azl, azh, adxs, adys, adzs
      integer ci, lp, lc
c
      ci = i + j * anx + k * ( anx * any ) + 1
      if( cells(ci) .eq. 0 )then
         nlist = nlist + 1
         if( nlist .gt. maxlst )then
            write(6,100)
 100        format(1x,'addlst: trying to add too many list elements.')
         else
            cells(ci) = nlist
            lprim(nlist) = p
            lnext(nlist) = 0
         endif
      else
         lp = 0
         lc = cells(ci)
 1       continue
         if( lc .eq. 0 )goto 2
            if( lprim(lc) .eq. p )then
               return
            else
               lp = lc
               lc = lnext(lc)
            endif
         goto 1
 2       continue
         nlist = nlist + 1
         if( nlist .gt. maxlst )then
            write(6,100)
         else
            lnext(lp) = nlist
            lprim(nlist) = p
            lnext(nlist) = 0
         endif
      endif
c
      return
      end
c
c
      subroutine sphbnd( i, pxl, pxh, pyl, pyh, pzl, pzh )
      implicit character*1 (a-z)
      integer i
      real pxl, pxh, pyl, pyh, pzl, pzh
c**********************************************************************
c find a bounding box for a sphere primitive. very easy!
c**********************************************************************
c
c-----------------
c-- sphere list --
c-----------------
c size: 4 * dimsph + 1 words = 4097
c
      integer maxsph, dimsph
      parameter( maxsph=1024 )
      parameter( dimsph=maxsph )
c
      integer nsph
      common /sphlsi/ nsph
c
      real sox(dimsph), soy(dimsph), soz(dimsph), srad(dimsph)
      common /sphlsr/ sox, soy, soz, srad
      pxl = sox(i) - srad(i)
      pxh = sox(i) + srad(i)
      pyl = soy(i) - srad(i)
      pyh = soy(i) + srad(i)
      pzl = soz(i) - srad(i)
      pzh = soz(i) + srad(i)
c
      return
      end
c
c
      subroutine tribnd( i, pxl, pxh, pyl, pyh, pzl, pzh )
      implicit character*1 (a-z)
      integer i
      real pxl, pxh, pyl, pyh, pzl, pzh
c**********************************************************************
c find a bounding box for a triangle primitive. very easy!
c**********************************************************************
c
c-------------------
c-- triangle list --
c-------------------
c 23 * dimtri + 1 words = 5889
c
      integer maxtri, dimtri
      parameter( maxtri=256 )
      parameter( dimtri=maxtri )
c
      integer ntri, dax(dimtri)
      common /trilsi/ ntri, dax
c
      real tx1(dimtri), ty1(dimtri), tz1(dimtri)
      real tx2(dimtri), ty2(dimtri), tz2(dimtri)
      real tx3(dimtri), ty3(dimtri), tz3(dimtri)
      real tnx(dimtri), tny(dimtri), tnz(dimtri), tnd(dimtri)
      real nvx1(dimtri), nvy1(dimtri), nvz1(dimtri)
      real nvx2(dimtri), nvy2(dimtri), nvz2(dimtri)
      real nvx3(dimtri), nvy3(dimtri), nvz3(dimtri)
      common /trilsr/ tx1, ty1, tz1, tx2, ty2, tz2, tx3, ty3, tz3,
     +                tnx, tny, tnz, nvx1, nvy1, nvz1, nvx2, nvy2,
     +                nvz2, nvx3, nvy3, nvz3, tnd
      pxl = tx1(i)
      if( tx2(i) .lt. pxl )pxl = tx2(i)
      if( tx3(i) .lt. pxl )pxl = tx3(i)
c
      pyl = ty1(i)
      if( ty2(i) .lt. pyl )pyl = ty2(i)
      if( ty3(i) .lt. pyl )pyl = ty3(i)
c
      pzl = tz1(i)
      if( tz2(i) .lt. pzl )pzl = tz2(i)
      if( tz3(i) .lt. pzl )pzl = tz3(i)
c
      pxh = tx1(i)
      if( tx2(i) .gt. pxh )pxh = tx2(i)
      if( tx3(i) .gt. pxh )pxh = tx3(i)
c
      pyh = ty1(i)
      if( ty2(i) .gt. pyh )pyh = ty2(i)
      if( ty3(i) .gt. pyh )pyh = ty3(i)
c
      pzh = tz1(i)
      if( tz2(i) .gt. pzh )pzh = tz2(i)
      if( tz3(i) .gt. pzh )pzh = tz3(i)
c
      return
      end
c
c
      subroutine grdtrv( ux, uy, uz, vx, vy, vz, nx, ny, nz )
      implicit character*1 (a-z)
      real ux, uy, uz, vx, vy, vz
      integer nx, ny, nz
c**********************************************************************
c visit all cells in a 3d integer grid of nx x ny x nz cells that are
c pierced by a ray with origin (ux,uy,uz) - **positive and within the
c grid** - and direction vector (vx,vy,vz) - not necessarily normalized.
c *** n.b.: cell coordinates passed to viscel() are zero based. ***
c this is amantides & woo's algorithm.
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c------------------
c-- acceleration --
c------------------
c for typical scenes, the "max cells" would be best around 20 to 40.
c (maybe - my tests say 10 is good). but the memory cost is quite high.
c likewise, it would be good to allow largish primitives to be indexed,
c but the maximum possible list size goes up as the cube of the largest
c number of cells allowed in any direction for the primitive to be accep
c for indexing. so the memory cost is (potentially) very high. it would
c possible to use the same list for more than one grid cell if it is ide
c to that for another cell. we don't try to detect that, though.
c if the ratio of the biggest dimension range to the smallest exceeds
c mxdmrt, adjust the ranges for calculating the grid parameters. this se
c to be always a good idea (hence set to 1.0).
c
c size: celdim + 2 * lstdim + 13 words = 1000 + 2 * 8000 + 13 = 17013 wo
c
      integer mxcelx, mxcely, mxcelz, dmcelx, dmcely, dmcelz
      integer celdim, mxcels, mxpcvr, maxlst, lstdim
      real mxdmrt
      parameter( mxcelx=10 )
      parameter( mxcely=10 )
      parameter( mxcelz=10 )
      parameter( dmcelx=mxcelx )
      parameter( dmcely=mxcely )
      parameter( dmcelz=mxcelz )
      parameter( celdim=dmcelx*dmcely*dmcelz )
      parameter( mxcels=mxcelx*mxcely*mxcelz )
      parameter( mxpcvr=2 )
      parameter( maxlst=2*2*2*mxcels )
      parameter( lstdim=maxlst )
      parameter( mxdmrt=1.0 )
c
      integer nlist, anx, any, anz, cells(celdim)
      integer lprim(lstdim), lnext(lstdim)
      common /acceli/ nlist, anx, any, anz, cells, lprim, lnext
c
      real axl, axh, ayl, ayh, azl, azh, adxs, adys, adzs
      common /accelr/ axl, axh, ayl, ayh, azl, azh, adxs, adys, adzs
      integer x, y, z, sx, sy, sz, jx, jy, jz, quit
      real tx, ty, tz, dx, dy, dz
c
c-- initial cell coordinates
c
      x = int(ux)
      y = int(uy)
      z = int(uz)
c
c-- visit the initial cell
c
      call viscel( x, y, z, quit )
      if( quit .ne. 0 )return
c
c-- integer cell steps for the vector
c
      if( vx .lt. 0.0 )then
         sx = -1
      else
         sx = 1
      endif
      if( vy .lt. 0.0 )then
         sy = -1
      else
         sy = 1
      endif
      if( vz .lt. 0.0 )then
         sz = -1
      else
         sz = 1
      endif
c
c-- find tx,ty,tz - the t values that move along the ray to the boundary
c-- of the initial cell.
c-- find dx,dy,dz - the t values that step the width, height and depth
c-- of a full cell.
c-- precision!
c
      if( abs( vx ) .lt. eps )then
         tx = gthuge
         dx = gthuge
      else if( vx .gt. 0 )then
         tx = ( ( x + 1 ) - ux ) / vx
         dx = 1.0 / vx
      else
         tx = ( x  - ux ) / vx
         dx = -1.0 / vx
      endif
      if( abs( vy ) .lt. eps )then
         ty = gthuge
         dy = gthuge
      else if( vy .gt. 0 )then
         ty = ( ( y + 1 ) - uy ) / vy
         dy = 1.0 / vy
      else
         ty = ( y  - uy ) / vy
         dy = -1.0 / vy
      endif
      if( abs( vz ) .lt. eps )then
         tz = gthuge
         dz = gthuge
      else if( vz .gt. 0 )then
         tz = ( ( z + 1 ) - uz ) / vz
         dz = 1.0 / vz
      else
         tz = ( z  - uz ) / vz
         dz = -1.0 / vz
      endif
c
c-- if any of the t values to move to the boundary are exactly zero,
c-- bump them a bit.
c-- precision!
c
      if( tx .eq. 0 )tx = tx + eps
      if( ty .eq. 0 )ty = ty + eps
      if( tz .eq. 0 )tz = tz + eps
c
c-- find just out of bounds cell coords for the ray along each axis
c
      if( sx .gt. 0 )then
         jx = nx
      else
         jx = -1
      endif
      if( sy .gt. 0 )then
         jy = ny
      else
         jy = -1
      endif
      if( sz .gt. 0 )then
         jz = nz
      else
         jz = -1
      endif
c
c-- step through the cells until we exit the array
c
 1    continue
         if( tx .lt. ty )then
            if( tx .lt. tz )then
               x = x + sx
               if( x .eq. jx )return
               tx = tx + dx
            else
               z = z + sz
               if( z .eq. jz )return
               tz = tz + dz
            endif
         else
            if( ty .lt. tz )then
               y = y + sy
               if( y .eq. jy )return
               ty = ty + dy
            else
               z = z + sz
               if( z .eq. jz )return
               tz = tz + dz
            endif
         endif
c
c-- visit the current cell
c
         call viscel( x, y, z, quit )
         if( quit .ne. 0 )return
      goto 1
c
      end
c
c
      subroutine boxins( ox, oy, oz, dx, dy, dz, sect, xi, yi, zi )
      implicit character*1 (a-z)
      real ox, oy, oz, dx, dy, dz, xi, yi, zi
      integer sect
c**********************************************************************
c see if the ray with origin (ox,oy,oz) direction (dx, dy, dz)
c (not necessarily normalised) intersects the bounding
c box for the scene geometric objects found by stacel(). if so, return
c sect set to 1 and the intersection point in
c (xi, yi, zi). otherwise, return sect 0.
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
c
c------------------
c-- acceleration --
c------------------
c for typical scenes, the "max cells" would be best around 20 to 40.
c (maybe - my tests say 10 is good). but the memory cost is quite high.
c likewise, it would be good to allow largish primitives to be indexed,
c but the maximum possible list size goes up as the cube of the largest
c number of cells allowed in any direction for the primitive to be accep
c for indexing. so the memory cost is (potentially) very high. it would
c possible to use the same list for more than one grid cell if it is ide
c to that for another cell. we don't try to detect that, though.
c if the ratio of the biggest dimension range to the smallest exceeds
c mxdmrt, adjust the ranges for calculating the grid parameters. this se
c to be always a good idea (hence set to 1.0).
c
c size: celdim + 2 * lstdim + 13 words = 1000 + 2 * 8000 + 13 = 17013 wo
c
      integer mxcelx, mxcely, mxcelz, dmcelx, dmcely, dmcelz
      integer celdim, mxcels, mxpcvr, maxlst, lstdim
      real mxdmrt
      parameter( mxcelx=10 )
      parameter( mxcely=10 )
      parameter( mxcelz=10 )
      parameter( dmcelx=mxcelx )
      parameter( dmcely=mxcely )
      parameter( dmcelz=mxcelz )
      parameter( celdim=dmcelx*dmcely*dmcelz )
      parameter( mxcels=mxcelx*mxcely*mxcelz )
      parameter( mxpcvr=2 )
      parameter( maxlst=2*2*2*mxcels )
      parameter( lstdim=maxlst )
      parameter( mxdmrt=1.0 )
c
      integer nlist, anx, any, anz, cells(celdim)
      integer lprim(lstdim), lnext(lstdim)
      common /acceli/ nlist, anx, any, anz, cells, lprim, lnext
c
      real axl, axh, ayl, ayh, azl, azh, adxs, adys, adzs
      common /accelr/ axl, axh, ayl, ayh, azl, azh, adxs, adys, adzs
      real t1, t2, tnear, tfar, t
c
      xi = 0.0
      yi = 0.0
      zi = 0.0
      sect = 0
      tnear = minflt
      tfar = maxflt
c
c-- precision!
c
      if( abs( dx ) .lt. eps )then
         if( ox .lt. axl .or. ox .gt. axh )then
            sect = 0
            return
         endif
      else
         tnear = ( axl - ox ) / dx
         tfar  = ( axh - ox ) / dx
         if( tnear .gt. tfar )then
            t = tnear
            tnear = tfar
            tfar = t
         endif
         if( tfar .lt. 0.0 )then
            sect = 0
            return
         endif
      endif
c
c-- precision!
c
      if( abs( dy ) .lt. eps )then
         if( oy .lt. ayl .or. oy .gt. ayh )then
            sect = 0
            return
         endif
      else
         t1 = ( ayl - oy ) / dy
         t2 = ( ayh - oy ) / dy
         if( t1 .gt. t2 )then
            t = t1
            t1 = t2
            t2 = t
         endif
         if( t1 .gt. tnear )tnear = t1
         if( t2 .lt. tfar )tfar = t2
         if( tnear .gt. tfar )then
            sect = 0
            return
         endif
         if( tfar .lt. 0.0 )then
            sect = 0
            return
         endif
      endif
c
c-- precision!
c
      if( abs( dz ) .lt. eps )then
         if( oz .lt. azl .or. oz .gt. azh )then
            sect = 0
            return
         endif
      else
         t1 = ( azl - oz ) / dz
         t2 = ( azh - oz ) / dz
         if( t1 .gt. t2 )then
            t = t1
            t1 = t2
            t2 = t
         endif
         if( t1 .gt. tnear )tnear = t1
         if( t2 .lt. tfar )tfar = t2
         if( tnear .gt. tfar )then
            sect = 0
            return
         endif
         if( tfar .lt. 0.0 )then
            sect = 0
            return
         endif
      endif
c
c-- there is an intersection. find where.
c
      xi = ox + tnear * dx
      yi = oy + tnear * dy
      zi = oz + tnear * dz
      sect = 1
c
      return
      end
c
c----------------------------------------------------------------------
c utility routines
c----------------------------------------------------------------------
c
      real function ranhlf( idum )
      implicit character*1 (a-z)
      integer idum
c**********************************************************************
c return a random number in the range +/- 0.5
c**********************************************************************
      integer seed
      data seed/12345/
      ranhlf = ran(seed) - 0.5
      return
      end
c
c
      subroutine nrmvec( dx, dy, dz )
      implicit character*1 (a-z)
      real dx, dy, dz
c**********************************************************************
c make (dx,dy,dz) a unit vector
c**********************************************************************
c
c-----------------------
c-- sundry parameters --
c-----------------------
c
      real gthuge, mineps, eps, minflt, maxflt, trtpi, degrad
      parameter( gthuge=1e7 )
      parameter( mineps=1e-5 )
      parameter( eps=1e-7 )
      parameter( minflt=-1e20 )
      parameter( maxflt=1e20 )
      parameter( trtpi=3.1415926 )
      parameter( degrad=trtpi/180.0 )
c
      integer eyeray, rflray, trnray
      parameter( eyeray=1 )
      parameter( rflray=2 )
      parameter( trnray=3 )
c
      integer enter, leave
      parameter( enter=1 )
      parameter( leave=2 )
c
      integer noacl, subacl
      parameter( noacl=0 )
      parameter( subacl=1 )
c
      integer bkcnst, bkvert, bkhorz
      parameter( bkcnst=0 )
      parameter( bkvert=1 )
      parameter( bkhorz=2 )
c
      integer lgtcon, lgtcos, lgtbrn
      parameter( lgtcon=0 )
      parameter( lgtcos=1 )
      parameter( lgtbrn=2 )
c
      integer sphprm, triprm
      parameter( sphprm=1 )
      parameter( triprm=2 )
c
      integer daxx, daxy, daxz
      parameter( daxx=1 )
      parameter( daxy=2 )
      parameter( daxz=3 )
      real len
c
      len = sqrt( dx*dx + dy*dy + dz*dz )
c
c-- precision!
c
      if( len .gt. eps )then
         dx = dx / len
         dy = dy / len
         dz = dz / len
      endif
c
      return
      end
